/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include  <stdio.h>
#include  <stdlib.h>
#include  <fcntl.h>
#include  <sys/stat.h>
#include  <sys/socket.h>
#include  <unistd.h>
#include  <time.h>

#ifdef HAVE_SYS_MODE_H
#include <sys/mode.h>
#endif
 
#include  "config.h"
#include  "file.h"
#include  "http.h"
#include  "ftp.h"
#include  "gopher.h"
#include  "url.h"
#include  "mode.h"


/********************************************************/
/* parameter  -  URL dokumentu				*/
/* vracia deskriptor soketu alebo suboru pre dokument	*/
/* osetrenie vyskytu v lokalnom strome			*/
/********************************************************/
bufio *get_data_socket(urlr , pos , modtime)
url *urlr;
int pos;
struct tm *modtime;
{
	char *fn;
	bufio *sock;
	struct stat estat;

	cfg.errcode = ERR_NOERROR;

	urlr->status &= ~URL_REDIRECT;

	if (cfg.mode != MODE_SYNC)
	{
		fn = url_to_filename(urlr);

		if (!access(fn , R_OK) && urlr->type != FILET)
		{
			urlr->status |= URL_REDIRECT;
			if (stat(fn , &estat) == 0)
			{
        			if (!S_ISDIR(estat.st_mode))
				{
			        	if ((sock = bufio_open(fn , O_BINARY | O_RDONLY)) < 0)
		        		{
		                		xperror(fn);
						cfg.errcode = ERR_FILE_OPEN;
					}
					xprintf(1 , gettext("File redirect\n"));

					cfg.totsz = estat.st_size;
					free(fn);
					return sock;
				}
			}
		}
		_free(fn);
	}

	if (urlr->type == HTTP 
#ifdef USE_SSL
		|| urlr->type == HTTPS
#endif
		|| (urlr->type == FTP && cfg.ftp_proxy && cfg.ftp_via_http && !cfg.ftp_dirtyp)
		|| (urlr->type == GOPHER && cfg.gopher_proxy && cfg.gopher_via_http))
	{
		if (cfg.sleep) 
		{
			my_sleep(cfg.sleep);
		}
		urlr->status &= ~URL_REDIRECT;
		return get_http_data_socket(urlr , pos , modtime);
	}
	else if (urlr->type == FTP)
	{
		if (cfg.sleep) 
		{
			my_sleep(cfg.sleep);
		}
		urlr->status &= ~URL_REDIRECT;
		return ftp_get_data_socket(urlr , pos , modtime);
	}
	if (urlr->type == GOPHER)
	{
		if (cfg.sleep) 
		{
			my_sleep(cfg.sleep);
		}
		urlr->status &= ~URL_REDIRECT;
		return gopher_get_data_socket(urlr);
	}
	if (urlr->type == FILET)
	{
		urlr->status &= ~URL_REDIRECT;
		return get_file_data_socket(urlr);
	}
	xprintf(1 , gettext("Unsupported URL"));
	return NULL;
}

/********************************************************/
/* parameter URL dokumentu				*/
/* uzatvori deskirptor pre dany dokument		*/
/********************************************************/
void close_socket(sock , urlr , read_st)
bufio *sock;
url *urlr;
int read_st;
{
	switch (urlr->type)
	{
		case FILET:
			if (sock) bufio_close(sock);
			break;
		case HTTP:
			if (!sock) return;
			if (!(urlr->status & URL_REDIRECT)) shutdown(bufio_getfd(sock) , 2);
			bufio_close(sock);
			break;
#ifdef USE_SSL
		case HTTPS:
			if (!sock) return;
			if (!(urlr->status & URL_REDIRECT))
			{
				SSL_shutdown(cfg.ssl_con);
				SSL_free (cfg.ssl_con);
				SSL_CTX_free (cfg.ssl_ctx);
				cfg.ssl_con = NULL;
				cfg.ssl_ctx = NULL;
				shutdown(bufio_getfd(sock) , 2);
			}
			bufio_close(sock);
			break;
#endif
		case GOPHER:
			if (!sock) return;
			if (!(urlr->status & URL_REDIRECT)) shutdown(bufio_getfd(sock) , 2);
			bufio_close(sock);
			break;
		case FTP:
			if (!(urlr->status & URL_REDIRECT) && sock)
			{
			    shutdown(bufio_getfd(sock) , 2);
			}

			if (sock)
			{
				 bufio_close(sock);
				 cfg.ftp_datasock = NULL;
			}

			if (urlr->status & URL_REDIRECT)
				return;

			if (cfg.ftp_control)
			{
			    if (read_st && 
			        (ftp_get_response(cfg.ftp_control , NULL , FALSE) >= 400))
			    {
				xprintf(1 , gettext("Warning broken ftp transfer ..\n"));
				cfg.errcode = ERR_FTP_TRUNC;
			    }
			    else if (!cfg.ftp_fatal_err && cfg.urlstack[cfg.actual_url+1])
			    {
				if (cfg.urlstack[cfg.actual_url+1]->type == FTP)
				{
				    char *opass = url_get_pass(urlr, NULL);
				    char *npass = url_get_pass(cfg.urlstack[cfg.actual_url+1], NULL);
				    char *ouser = url_get_user(urlr, NULL);
				    char *nuser = url_get_user(cfg.urlstack[cfg.actual_url+1], NULL);

				    /**** ak sa bude v dalsom cykle a prenasat z toho isteho ****/
				    /**** servera nechaj otvorene riadiace spojenie          ****/
				    if ((!strcmp(
					cfg.urlstack[cfg.actual_url+1]->p.ftp.host ,
					urlr->p.ftp.host)) && (urlr->p.ftp.port == 
					cfg.urlstack[cfg.actual_url+1]->p.ftp.port) &&
					((opass && npass && !strcmp(opass , npass)) ||
				    	 (!ouser && !nuser)) &&
				    	 ((ouser && nuser && !strcmp(ouser , nuser)) ||
				    	 (!ouser && !nuser)))
				    {
					return;
				    }
				}
			    }
			    if (urlr->p.ftp.dir && ((cfg.actual_url + 1) == cfg.total_cnt))
				return;
			    /**** ukonci riadiace FTP spojenie ****/
			    bufio_write(cfg.ftp_control , "QUIT\r\n" , 6);
#ifdef DEBUG
			    if (cfg.debug)
			    {
			    	ftp_get_response(cfg.ftp_control , NULL , FALSE);
			    }
#endif
			    shutdown(bufio_getfd(cfg.ftp_control) , 2);
			    bufio_close(cfg.ftp_control);
			    cfg.ftp_control = NULL;
			}
			break;

		default:
			if (sock) bufio_close(sock);
	}
}
