/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>

#include "config.h"
#include "authinfo.h"
#include "tools.h"
#include "url.h"

#define AUTHI_UNKNOWN	0
#define AUTHI_PROTO	1
#define AUTHI_HOST	2
#define AUTHI_USER	3
#define	AUTHI_PASS	4
#define	AUTHI_BASE	5
#define AUTHI_TYPE	6
#define AUTHI_REALM	7

typedef struct {
	char	*id;
	int	type;
}authi_entry;

static authi_entry authidef[] = {
	{"Proto:" , AUTHI_PROTO} ,
	{"Host:" , AUTHI_HOST} ,
	{"User:" , AUTHI_USER} ,
	{"Pass:" , AUTHI_PASS} ,
	{"Base:" , AUTHI_BASE} ,
	{"Realm:" , AUTHI_REALM} ,
	{"Type:" , AUTHI_TYPE} ,
};

#define free_deep_authinfo(entry)\
	_free(entry->pass);\
	_free(entry->user);\
	_free(entry->host);\
	_free(entry->base);\
	_free(entry);

int authinfo_load(fn)
char *fn;
{
	bufio *fd;
	int lnsz;
	int i;
	char lnbuf[4096];
	char *lns;
	char *p;
	bool new_section = TRUE;
	bool found;
	authinfo *ap = NULL;

	for(i = 0 ; i < cfg.authdata_num ; i++)
	{
		free_deep_authinfo(cfg.authdata[i]);
	}
	_free(cfg.authdata);
	cfg.authdata_num = 0;

	if (!(fd = bufio_open(fn , O_BINARY | O_RDONLY)))
	{
		xperror(fn);
		return -1;
	}

	while((lnsz = bufio_readln(fd , lnbuf , sizeof(lnbuf))) > 0)
	{
		strtok(lnbuf , "\n\r");
		for(lns = lnbuf ; *lns && isspace(*lns) ; lns++);
		if (*lns == '#') continue;

		if (!*lns)
		{
			if (ap && ap->prot == UNKNOWN)
			{
				xprintf(0 , gettext("Bad section - specify protocol\n"));
				free_deep_authinfo(ap);
			}
			if (ap && !ap->host)
			{
				xprintf(0 , gettext("Bad section - specify hostname\n"));
				free_deep_authinfo(ap);
			}
			if (ap && ap->port == 0)
			{
				ap->port = prottable[ap->prot].default_port;
			} 
			if (ap)
			{
				cfg.authdata_num++;
				cfg.authdata = (authinfo **) _realloc(cfg.authdata , cfg.authdata_num * sizeof(authinfo *));
				cfg.authdata[cfg.authdata_num - 1] = ap;
				ap = NULL;
			}
			new_section = TRUE;
		}

		if (*lns == '#' || !*lns) continue;

		if (new_section)
		{
			new_section = FALSE;
			ap = (authinfo *) _malloc(sizeof(authinfo));
			ap->prot = UNKNOWN;
			ap->host = NULL;
			ap->port = 0;
			ap->user = NULL;
			ap->pass = NULL;
			ap->base = NULL;
			ap->type = 2;
		}

		found =  FALSE;
		for(i = 0 ; i < NUM_ELEM(authidef) ; i++)
		{
			if (!strncasecmp(lns , authidef[i].id , strlen(authidef[i].id)))
			{
				lns += strlen(authidef[i].id);
				for( ; *lns && isspace(*lns) ; lns++);
				for(p = lns + strlen(lns) ; *p && isspace(*p) ; p--) *p = '\0';
				found = TRUE;
				break;
			}
		}

		if (!found)
		{
			xprintf(0 , gettext("Unable to parse : \"%s\"\n") , lns);
			continue;
		}

		switch(authidef[i].type)
		{
			case AUTHI_PROTO:
				found = FALSE;
				for(i = 0 ; i < NUM_ELEM(prottable) ; i++)
				{
					if (prottable[i].urlid && !strcasecmp(lns , prottable[i].urlid))
					{
						found = TRUE;
						ap->prot = prottable[i].id;
						break;
					}
				}
				if (!found)
				{
					xprintf(0 , gettext("Bad protocol - %s\n") , lns);
				} 
			break;
			case AUTHI_HOST:
				strtok(lns , ":");
				ap->host = new_string(lns);
				p = strtok(NULL , "");
				if (p) 
				{
					ap->port = _atoi(p);
					if (errno == ERANGE)
					{
						xprintf(0 , gettext("Bad port number %s\n") , p);
					}
				}
			break;
			case AUTHI_USER:
				ap->user = new_string(lns);
			break;
			case AUTHI_PASS:
				ap->pass = new_string(lns);
			break;
			case AUTHI_BASE:
				ap->base = new_string(lns);
			break;
			case AUTHI_REALM:
				ap->realm = new_string(lns);
			break;
			case AUTHI_TYPE:
				ap->type = _atoi(lns);
				if (errno == ERANGE || !(ap->type == 1 || ap->type == 2 || ap->type == 3))
				{
					xprintf(0 , gettext("Unknown auth type - %s\n") , lns);
					ap->type = 2;
				}
			break;
		}
	}
	bufio_close(fd);
	return 0;
}

authinfo * authinfo_match_entry(prot , host , port , path , realm)
protocol prot;
char *host;
int port;
char *path;
char *realm;
{
	int mlen = -1 , len , i;
	authinfo *ret = NULL;

	for (i = 0; i < cfg.authdata_num ; i++)
	{
		if (cfg.authdata[i]->prot == prot &&
		    (!strcmp(cfg.authdata[i]->host , host)) &&
		    (port == cfg.authdata[i]->port))
		{
			if (realm && cfg.authdata[i]->realm &&
				!strcmp(realm , cfg.authdata[i]->realm))
			{
				ret = cfg.authdata[i];
				break;
			}
		 	if ((cfg.authdata[i]->base && 
				!strncmp(path , cfg.authdata[i]->base , 
				strlen(cfg.authdata[i]->base))) ||
				!cfg.authdata[i]->base)
			{
				if (cfg.authdata[i]->base) 
					len = strlen(cfg.authdata[i]->base);
				else
					len = 0;

				if (len > mlen)
				{
					ret = cfg.authdata[i];
					mlen = len;
				}
			}
		}
	}

	return ret;
}

