/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <stdlib.h>
#include <string.h>

#ifdef HAVE_FNMATCH
#include <fnmatch.h>
#else
#include "fnmatch.h"
#endif

#include "config.h"
#include "url.h"
#include "condition.h"
#include "tools.h"
#include "robots.h"
#include "uexit.h"

/********************************************************/
/* kontrola ci dany server je v zozname povolenych/	*/
/* zakazanych domen					*/
/********************************************************/
static bool domain_condition(site)
char *site;
{
	char **p = cfg.condition.domains;

	while(*p)
	{
		if (!(strlen(*p) > strlen(site)))
		{
			if (!strcasecmp(*p , site + strlen(site) - strlen(*p)))
			{
				return TRUE;
			}
		}
		p++;
	}
	return FALSE;
}

/********************************************************/
/* kontrola ci dokument ma dany sufix			*/
/********************************************************/
static bool cmp_sfx(urlr , sfx)
url *urlr;
char *sfx;
{
	int nlen,slen;
	char *p;

	switch(urlr->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			p = urlr->p.http.document;
			break;
		case FTP:
			p = urlr->p.ftp.path;
			break;
		case GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case FILET:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}
	
	slen = strlen(sfx);
	nlen = strlen(p);

	if (nlen < slen) return FALSE;

	return (!strcmp(sfx , p + nlen - slen));
}

/********************************************************/
/*  kontrola ci sufix URL je zo zoznamov sufixov	*/
/********************************************************/
static bool sfx_condition(urlr)
url *urlr;
{
	char **p=cfg.condition.sufix;

	while (*p)
	{
		if (cmp_sfx(urlr , *p))
		{
			return TRUE;
		}
		p++;
	}

	return FALSE;
}

/********************************************************/
/* kontrola ci URL ma dany prefix			*/
/********************************************************/
static bool cmp_prefix(urlr , prefix)
url *urlr;
char *prefix;
{
	char *p;

	switch(urlr->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			p = urlr->p.http.document;
			break;
		case FTP:
			p = urlr->p.ftp.path;
			break;
		case GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case FILET:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}
	

	return (!strncmp(prefix , p , strlen(prefix)));
}


/********************************************************/
/* kontrola ci URL ma prefix zo zoznamu 		*/
/* povolenych/zakaznych					*/
/********************************************************/
static bool prefix_condition(urlr)
url *urlr;
{
	char **p=cfg.condition.dir_prefix;

	while (*p)
	{
		if (cmp_prefix(urlr , *p))
		{
			return TRUE;
		}
		p++;
	}

	return FALSE;
}

/********************************************************/
/* kontola URL ci splna dany regularny vyraz		*/
/********************************************************/
static bool cmp_pattern(urlr, pattern)
url *urlr;
char **pattern;
{
	char *p;
	char **pp;

	switch(urlr->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			p = urlr->p.http.document;
			break;
		case FTP:
			p = urlr->p.ftp.path;
			break;
		case GOPHER:
			p = urlr->p.gopher.selector;
			break;
		case FILET:
			p = urlr->p.file.filename;
			break;
		default: return FALSE;
	}

	for (pp = pattern; pp && *pp ; pp++) 
	{
		if (!fnmatch(*pp , p , 0)) return FALSE;
	}

	return (pattern != NULL);
}

static bool cmp_url_pattern(urlr , url_pattern)
url *urlr;
char **url_pattern;
{
	char *p = url_to_urlstr_woauth(urlr);
	char **pp;

	for (pp = url_pattern; pp && *pp ; pp++) 
	{
		if (!fnmatch(*pp , p , 0))
		{
			_free(p);
			return FALSE;
		}
	}

	free(p);

	return (url_pattern != NULL);
}
/********************************************************/
/* kontrola ci URL splna vsetky obmedzujuce podmienky	*/
/********************************************************/
int url_append_condition(urlp , appnd)
url *urlp;
int appnd;
{
	char *p = url_to_filename(urlp);
	int ret_val;
	char *site;

	if (p == NULL) return 0;

	free(p);

	ret_val = ! prottable[urlp->type].supported;

	if (!cfg.condition.ftp)
		ret_val |= urlp->type == FTP;

	if (!cfg.condition.http)
		ret_val |= urlp->type == HTTP;

#ifdef USE_SSL
	if (!cfg.condition.https)
		ret_val |= urlp->type == HTTPS;
#endif

	if (!cfg.condition.gopher)
		ret_val |=  urlp->type == GOPHER;

	if (cfg.condition.max_levels)
	{
		ret_val |= urlp->level > (cfg.condition.max_levels -
			(urlp->status & URL_INLINE_OBJ) ? 1 : 0);
	}
	
	if (cfg.condition.max_documents)
	{
		ret_val |= (appnd && 
			    ((cfg.total_cnt + 1) > cfg.condition.max_documents)) || 
			   (!appnd && 
			    (cfg.actual_url > cfg.condition.max_documents));
	}

	if (urlp->type == GOPHER)
	{
		ret_val |= ! (urlp->p.gopher.type == '1' || urlp->p.gopher.type == '0' ||
			urlp->p.gopher.type == '2' || urlp->p.gopher.type == '4' ||
			urlp->p.gopher.type == '5' || urlp->p.gopher.type == '6' ||
			urlp->p.gopher.type == '9' || urlp->p.gopher.type == 'g' ||
			urlp->p.gopher.type == 'I');
	}

	if (!ret_val && (urlp->type == HTTP || urlp->type == HTTPS) && !cfg.condition.cgi)
	{
		ret_val |= urlp->p.http.searchstr != NULL;
	}

	site = url_get_site(urlp);
	if (!ret_val && cfg.condition.sites && cfg.condition.sites[0] && site)
	{
		if (cfg.condition.allow_site) 
		{
			ret_val |= !is_in_list(site , cfg.condition.sites);
		}
		else
		{
			ret_val |= is_in_list(site , cfg.condition.sites);
		}
	}

	if (!ret_val && cfg.condition.domains && cfg.condition.domains[0] && site)
	{
		if (cfg.condition.allow_domain) 
		{
			ret_val |= !domain_condition(site);
		}
		else
		{
			ret_val |= domain_condition(site);
		}
	}

	if (!ret_val && cfg.condition.sufix && cfg.condition.sufix[0] && urlp->type != FILET)
	{
		if (cfg.condition.allow_sufix)
		{
			ret_val |= !sfx_condition(urlp);
		}
		else
		{
			ret_val |= sfx_condition(urlp);
		}
	}

	if (!ret_val && cfg.condition.dir_prefix && cfg.condition.dir_prefix[0] && 
		urlp->type != FILET)
	{
		if (cfg.condition.allow_prefix)
		{
			ret_val |= !prefix_condition(urlp);
		}
		else
		{
			ret_val |= prefix_condition(urlp);
		}
	}

	if (!ret_val && cfg.condition.pattern && urlp->type != FILET)
	{
		ret_val |= cmp_pattern(urlp, cfg.condition.pattern);
	}

	if (!ret_val && cfg.condition.skip_pattern && urlp->type != FILET)
	{
		ret_val |= !cmp_pattern(urlp, cfg.condition.skip_pattern);
	}

	if (!ret_val && cfg.condition.url_pattern && urlp->type != FILET)
	{
		ret_val |= cmp_url_pattern(urlp , cfg.condition.url_pattern);
	}

	if (!ret_val && cfg.condition.skip_url_pattern && urlp->type != FILET)
	{
		ret_val |= !cmp_url_pattern(urlp , cfg.condition.skip_url_pattern);
	}

	if (!ret_val && urlp->type != FILET && cfg.condition.dont_leave_site)
	{
		url *gparent = urlp;

		while (gparent->parent_url[0]) gparent = gparent->parent_url[0];

		ret_val |= (urlp->type != gparent->type) || 
			(url_get_port(urlp) != url_get_port(gparent)) ||
			strcmp(url_get_site(urlp) , url_get_site(gparent));
	}
	
	if (!ret_val && urlp->type != FILET && cfg.condition.dont_leave_dir)
	{
		url *gparent = urlp;
		char *p1,*p2;

		while (gparent->parent_url[0]) gparent = gparent->parent_url[0];

		p1 = url_to_filename(urlp);
		p2 = url_to_filename(gparent);
		
		p = strrchr(p2 , '/');
		if (p) *p = '\0';

		ret_val |= (urlp->type != gparent->type) || 
			(url_get_port(urlp) != url_get_port(gparent)) ||
			strcmp(url_get_site(urlp) , url_get_site(gparent)) ||
			strncmp(p1 , p2 , strlen(p2));

		_free(p1);
		_free(p2);
	}

	if (!ret_val && urlp->type != FILET && cfg.condition.leave_level)
	{
		url *gparent = urlp;
		url *pomurl = urlp;
		int level = 0;

		while (gparent->parent_url[0]) gparent = gparent->parent_url[0];

		while (pomurl)
		{
			if ((pomurl->type == gparent->type) && 
				(url_get_port(pomurl) == url_get_port(gparent)) &&
				!strcmp(url_get_site(pomurl) , url_get_site(gparent)))
			{
				break;
			}

			if ((level - ((urlp->status & URL_INLINE_OBJ) ? 1 : 0)) >=
				cfg.condition.leave_level)
			{
				ret_val = TRUE; 
				break;
			}
			level ++;
			pomurl = pomurl->parent_url[0];
		}
	}

	if (!ret_val && cfg.condition.uexit)
	{
		ret_val = !uexit_condition(urlp , NULL , NULL);
	}

	return !ret_val;
}
