/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>

#include "config.h"
#include "tools.h"
#include "ftp.h"
#include "http.h"
#include "gopher.h"
#include "times.h"
#include "html.h"

static cfg_param params[] = {
#ifdef I_FACE
	{
	 "-X" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.xi_face , 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-X                 - start Xwindows interface\n") ,
	} ,
	{
	 "-runX" , 
	 "RunX:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.run_iface , 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-runX              - after start of GUI interface, immediately\n"
		     "\t                     start processing of enterd URLs\n") ,
	} ,
#endif
	{
	 "-progress" , 
	 "ShowProgress:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.progres , 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-progress/-noprogress\n"
		     "\t                   - show retrieving progress while running on terminal\n") ,
	},
	{
	 "-noprogress" , 
	 NULL , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.progres , 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-lmax" , 
	 "MaxLevel:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 0 , 
	 &cfg.condition.max_levels, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-lmax $nr          - allowed deep of tree\n") ,
	} ,
	{
	 "-dmax" , 
	 "MaxDocs:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 0 , 
	 &cfg.condition.max_documents, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-dmax $nr          - maximal number of downloaded documents\n") ,
	} ,
	{
	 "-sleep" , 
	 "SleepBetween:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 0 , 
	 &cfg.sleep, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-sleep $nr         - sleep for $nr seconds between trasfers, default 0 s\n") ,
	} ,
	{
	 "-retry" , 
	 "MaxRetry:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 1 , 
	 &cfg.nretry, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-retry $nr         - number of retries if anything failed\n") ,
	} ,
	{
	 "-nregets" , 
	 "MaxRegets:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 2 , 
	 &cfg.nreget, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-nregets $nr       - max number of regets on single file, default 2\n") ,
	} ,
	{
	 "-nredirs" , 
	 "MaxRedirections:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 5 , 
	 &cfg.nredir, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-nredirs $nr       - max number of followed HTTP redirections, default 5\n") ,
	} ,
	{
	 "-timeout" , 
	 "CommTimeout:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 0 , 
	 &cfg.ctimeout, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-timeout $nr       - timeout for network communications (min).\n"
		     "\t                     0 == no timeout,\n"
		     "\t                     default = 0\n") ,
	} ,
	{
	 "-rollback" , 
	 "RegetRollbackAmount:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 0 , 
	 &cfg.rollback, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-rollback $nr      - number of bytes from end of file, which is\n"
		     "\t                     in reget , to discard , default 0\n") ,

	} ,
	{
	 "-ddays" , 
	 "DocExpiration:" , 
	 NULL , 
	 PARAM_NUM , 
	 (void *) 5 , 
	 &cfg.ddays, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-ddays $nr         - number of days since last access when is document\n"
		     "\t                     checked in sync mode\n"),
	} ,
	{
	 "-nocache" , 
	 "UseCache:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.cache, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-nocache/-cache    - disallow caching of HTTP documents (on proxy cache)\n") ,
	} ,
	{
	 "-cache" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.cache, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-noRobots" , 
	 "UseRobots:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.allow_robots, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noRobots/-Robots  - care about \"robots.txt\" file ?\n") ,
	} ,
	{
	 "-Robots" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.allow_robots, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-noFTP" , 
	 "AllowFTP:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.ftp, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noFTP/-FTP        - don't download FTP files\n") ,
	} ,
	{
	 "-FTP" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.ftp, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-noHTTP" , 
	 "AllowHTTP:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.http, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noHTTP/-HTTP      - don't download HTTP files\n") ,
	} ,
	{
	 "-HTTP" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.http, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
#ifdef USE_SSL
	{
	 "-noSSL" , 
	 "AllowSSL:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.https, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noSSL/-SSL        - dont download (HTTPS) SSL files\n"),
	} ,
	{
	 "-SSL" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.https, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
#endif
	{
	 "-noGopher" , 
	 "AllowGopher:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.gopher, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noGopher/-Gopher - download Gopher files ?\n") ,
	} ,
	{
	 "-Gopher" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.gopher, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-noCGI" , 
	 "AllowCGI:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.cgi, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noCGI/-CGI        - download parametric CGI pages ?\n") ,
	} ,
	{
	 "-CGI" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.condition.cgi, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-noEnc" , 
	 "AllowGZEncoding:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.use_enc, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noEnc/-Enc        - allow transfer of encoded files ?\n") ,
	} ,
	{
	 "-Enc" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.use_enc, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-noRelocate" , 
	 "AllowRelocation:" , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.rewrite_links, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-noRelocate/-Relocate\n"
		     "\t                   - dont't rewrite links\n") ,
	} ,
	{
	 "-Relocate" , 
	 NULL , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.rewrite_links, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-FTPhtml" , 
	 "FTPhtml:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.ftp_html, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-FTPhtml/-noFTPhtml\n"
		     "\t                   - process HTML files downloaded over FTP\n") ,
	} ,
	{
	 "-noFTPhtml" , 
	 NULL , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.ftp_html, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-FTPlist" , 
	 "FTPListCMD:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.ftp_list, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-FTPlist/-noFTPlist\n"
		     "\t                   - use wide FTP directory listing\n") ,
	} ,
	{
	 "-noFTPlist" , 
	 NULL , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.ftp_list, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-FTPdir" , 
	 "AllowFTPRecursion:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.condition.ftpdir, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-FTPdir/-noFTPdir  - recurse FTP directory\n") ,
	} ,
	{ 
	 "-noFTPdir" , 
	 NULL , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.condition.ftpdir, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-store_index" , 
	 "StoreDirIndexFile:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) TRUE , 
	 &cfg.store_index, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-store_index/-nostore_index\n"
		     "\t                   - store directory URLs as index files\n") ,
	} ,
	{ 
	 "-nostore_index" , 
	 NULL , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) TRUE , 
	 &cfg.store_index, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-force_reget" , 
	 "ForceReget:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.freget, 
	 NULL , 
	 NULL ,
	 gettext_nop("\t-force_reget/-noforce_reget\n"
		     "\t                   - force reget of whole file when server doesn't\n"
		     "\t                     support reget\n") ,

	} ,
	{
	 "-noforce_reget" , 
	 NULL , 
	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.freget, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
#ifdef DEBUG
	{
	 "-debug" , 
	 "Debug:" , 
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.debug, 
 	 NULL , 
	 NULL ,
	 gettext_nop("\t-debug             - turn on debug mode\n") ,
	} ,
	{
	 "-nodebug" , 
	 NULL , 
 	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.debug, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
#endif
	{
	 "-asite" , 
	 "AllowedSites:" , 
	 NULL , 
	 PARAM_STRLIST , 
	 (void *) NULL , 
	 &cfg.condition.sites, 
	 (void *) TRUE , 
	 &cfg.condition.allow_site ,
	 gettext_nop("\t-asite $list       - comma-separated list of allowed sites\n") ,
	} ,
	{
	 "-dsite" , 
	 "DisallowedSites:" , 
	 NULL , 
	 PARAM_STRLIST , 
	 (void *) NULL , 
	 &cfg.condition.sites,
	 (void *) FALSE , 
	 &cfg.condition.allow_site ,
	 gettext_nop("\t-dsite $list       - comma-separated list of disallowed sites\n") ,
	} ,
	{
	 "-adomain" ,
	 "AllowedDomains:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.domains,
	 (void *) TRUE ,
	 &cfg.condition.allow_domain ,
	 gettext_nop("\t-adomain $list     - list of allowed domains\n") ,
	} ,
	{
	 "-ddomain" ,
	 "DisallowedDomains:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.domains,
	 (void *) FALSE ,
	 &cfg.condition.allow_domain ,
	 gettext_nop("\t-ddomain $list     - list of disallowed domains\n") ,
	} ,
	{
	 "-aprefix" ,
	 "AllowedPrefixes:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.dir_prefix,
	 (void *) TRUE ,
	 &cfg.condition.allow_prefix ,
	 gettext_nop("\t-aprefix $list     - list of allowed directory/file prefixes\n") ,
	} ,
	{
	 "-dprefix" , 
	 "DisallowedPrefixes:" , 
	 NULL , 
	 PARAM_STRLIST , 
	 (void *) NULL , 
	 &cfg.condition.dir_prefix, 
	 (void *) FALSE , 
	 &cfg.condition.allow_prefix ,
	 gettext_nop("\t-dprefix $list     - list of disallowed directory/file prefixes\n") ,
	} ,
	{
	 "-asfx" , 
	 "AllowedSufixes:" , 
	 NULL , 
	 PARAM_STRLIST , 
	 (void *) NULL , 
	 &cfg.condition.sufix,  
	 (void *) TRUE , 
	 &cfg.condition.allow_sufix ,
	 gettext_nop("\t-asfx $list        - list of allowed sufixes\n") ,
	} ,
	{
	 "-dsfx" , 
	 "DisallowedSufixes:" , 
	 NULL ,
	 PARAM_STRLIST ,
 	 (void *) NULL ,
	 &cfg.condition.sufix,
	 (void *) FALSE , 
	 &cfg.condition.allow_sufix ,
	 gettext_nop("\t-dsfx $list        - list of disallowed sufixes\n") ,
	} ,
	{
	 "-amimet" ,
	 "AllowedMIMETypes:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.mime ,
	 (void *) TRUE ,
	 &cfg.condition.allow_mime ,
	 gettext_nop("\t-amimet $list      - list of alloved MIME types\n") ,
	},
	{
	 "-dmimet" ,
	 "DisallowedMIMETypes:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.mime ,
	 (void *) FALSE ,
	 &cfg.condition.allow_mime ,
	 gettext_nop("\t-dmimet $list      - list of disallowed MIME types\n") ,
	},
	{
	 "-alang" , 
	 "PrefferedLanguages:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.accept_lang,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-alang $list       - list of preffered languages (only via HTTP)\n") ,
	} ,
	{
	 "-acharset" , 
	 "PrefferedCharset:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.accept_chars,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-acharset $list    - list of preffered character sets (only via HTTP)\n") ,
	} ,
	{
	 "-scnDir" ,
	 "ScenarioDirectory:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.scndir ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-scndir $dir       - directory ,wher your scenarios (config files)\n"
		     "\t                     stored are\n") ,
	} ,
	{
	 "-cdir" ,
	 "WorkingDir:" ,
	 "HTTPGDIR" , 
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.cache_dir,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-cdir $dir         - directory for storing documents\n") ,
	} ,
	{
	 "-subdir" ,
	 "WorkingSubDir:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.subdir,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-subdir $dir       - subdirectory of cdir to operate on\n") ,
	} ,
	{
	 "-scenario" ,
	 NULL ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.scenario ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-scenario $str     - name of scenario from scenario directory to load\n"
                     "\t                     and or run\n"),
	} ,
	{
	 "-auth_scheme" ,
	 "HTTPAuthorizationScheme:" ,
	 NULL ,
	 PARAM_AUTHSCH ,
	 (void *) 2 ,
	 &cfg.auth_scheme ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-auth_scheme 1/2/3 - HTTP authorization scheme 1-user 2-Basic 3-Digest,\n"
		     "\t                     default 2\n") ,
	} ,
	{
	 "-auth_name" ,
	 "HTTPAuthorizationName:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.name_auth ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-auth_name $str    - name for Authorization (only via HTTP)\n") ,
	} ,
	{
	 "-auth_passwd" ,
	 "HTTPAuthorizationPassword:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.passwd_auth ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-auth_passwd $str  - password for Authorization (only via HTTP)\n") ,
	} ,
	{
	 "-auth_reuse_nonce" ,
	 "AuthReuseDigestNonce:" ,
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.auth_reuse_nonce, 
 	 NULL , 
	 NULL ,
	 gettext_nop("\t-auth_reuse_nonce/-noauth_reuse_nonce\n"
		     "\t                   - reuse one HTTP digest access authorization\n"
		     "\t                     nonce for more requests\n") ,
	} ,
	{
	 "-noauth_reuse_nonce" , 
	 NULL , 
 	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.auth_reuse_nonce, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
	{
	 "-auth_reuse_proxy_nonce" ,
	 "AuthReuseProxyDigestNonce:" ,
	 NULL , 
	 PARAM_PBOOL , 
	 (void *) FALSE , 
	 &cfg.auth_reuse_proxy_nonce, 
 	 NULL , 
	 NULL ,
	 gettext_nop("\t-auth_reuse_proxy_nonce/-noauth_reuse_proxy_nonce\n"
		     "\t                   - reuse one HTTP proxy digest access\n"
		     "\t                     authorization nonce for more requests\n") ,
	} ,
	{
	 "-noauth_reuse_proxy_nonce" , 
	 NULL , 
 	 NULL , 
	 PARAM_NBOOL , 
	 (void *) FALSE , 
	 &cfg.auth_reuse_proxy_nonce, 
	 NULL , 
	 NULL ,
	 NULL ,
	} ,
#ifdef USE_SSL
	{
	 "-ssl_cert_passwd" ,
	 "SSLCertPassword:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.ssl_cert_passwd ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ssl_cert_passwd $str  - password for SSL certifiaction file\n") ,
	},
	{
	 "-ssl_cert_file" ,
	 "SSLCertFile:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.ssl_cert_file ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ssl_cert_file $str  - SSL certifiaction file\n") ,
	},
	{
	 "-ssl_key_file" ,
	 "SSLKeyFile:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.ssl_key_file ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ssl_key_file $str  - SSL certifiaction key file\n") ,
	},
#endif
	{
	 "-from" ,
	 "EmailAddress:" ,
	 NULL ,
	 PARAM_STR ,
	 (void*) NULL ,
	 &cfg.from ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-from $str         - e-mail address used for user identification\n") ,
	} ,
	{
	 "-send_from" ,
	 "SendFromHeader:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.send_from ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-send_from/-nosend_from\n"
		     "\t                   - send From: header in HTTP request with your e-mail address\n") ,
	} ,
	{
	 "-nosend_from" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.send_from ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-identity" ,
	 "Identity:" ,
	 NULL ,
	 PARAM_STR ,
	 (void*) NULL ,
	 &cfg.identity ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-identity $str      - contents of User-agent: field in HTTP request\n") ,
	} ,
	{
	 "-pattern" , 
	 "MatchPattern:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.pattern ,
	 NULL ,
	 NULL , 
	 gettext_nop("\t-pattern $list     - list of matching patterns for files\n") ,
	} ,
	{
	 "-skip_pattern" , 
	 "SkipMatchPattern:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.skip_pattern ,
	 NULL ,
	 NULL , 
	 gettext_nop("\t-skip_pattern $list\n"
	   	     "\t                   - list of skip matching patterns for files\n") ,
	} ,
	{
	 "-url_pattern" , 
	 "URLMatchPattern:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.url_pattern ,
	 NULL ,
	 NULL , 
	 gettext_nop("\t-url_pattern $list - list of matching patterns for urls\n") ,
	} ,
	{
	 "-skip_url_pattern" , 
	 "SkipURLMatchPattern:" ,
	 NULL ,
	 PARAM_STRLIST ,
	 (void *) NULL ,
	 &cfg.condition.skip_url_pattern ,
	 NULL ,
	 NULL , 
	 gettext_nop("\t-skip_url_pattern $list\n"
		     "\t                   - list of matching patterns for urls\n") ,
	} ,
	{
	 "-mode" , 
	 "DefaultMode:" ,
	 NULL ,
	 PARAM_MODE ,
	 (void *) MODE_NORMAL ,
	 &cfg.mode ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-mode $mode        - set operation mode\n"
		     "\t                         normal - recurse throught WWW (default)\n"
		     "\t                         linkupdate - update remote links in local tree\n"
		     "\t                         sync - synchronize local tree with remote WWW\n"
		     "\t                                servers\n"
		     "\t                         singlepage - single page with inline objects\n"
		     "\t                         singlereget - reget file until not whole\n"
		     "\t                         resumeregets - reget all files which are broken\n"
		     "\t                         dontstore - transfer documents, but don't store\n") ,
	} ,
	{
	 "-ftp_proxy" ,
	 "FTPProxy:" ,
	 "FTP_PROXY_SERVER" ,
	 PARAM_CONN ,
	 (void *) NULL ,
	 &cfg.ftp_proxy ,
	 (void *) DEFAULT_FTP_PROXY_PORT ,
	 &cfg.ftp_proxy_port ,
	 gettext_nop("\t-ftp_proxy $site[:$port]\n"
		     "\t                   - ftp proxy/cache server\n") ,
	} ,
	{
	 "-http_proxy" ,
	 "HTTPProxy:" ,
	 "HTTP_PROXY_SERVER" ,
	 PARAM_CONN ,
	 (void *) NULL ,
	 &cfg.http_proxy ,
	 (void *) DEFAULT_HTTP_PROXY_PORT ,
	 &cfg.http_proxy_port ,
	 gettext_nop("\t-http_proxy $site[:$port]\n"
		     "\t                   - http proxy/cache server\n") ,
	} ,
	{
	 "-gopher_proxy" ,
	 "GopherProxy:" ,
	 NULL ,
	 PARAM_CONN ,
	 (void *) NULL ,
	 &cfg.gopher_proxy ,
	 (void *) DEFAULT_GOPHER_PROXY_PORT ,
	 &cfg.gopher_proxy_port ,
	 gettext_nop("\t-gopher_proxy $site[:$port]\n"
		     "\t                   - gopher proxy/cache server\n") ,
	} ,
#ifdef USE_SSL
	{
	 "-ssl_proxy" ,
	 "SSLProxy:" ,
	 NULL ,
	 PARAM_CONN ,
	 (void *) NULL ,
	 &cfg.ssl_proxy ,
	 (void *) DEFAULT_SSL_PROXY_PORT ,
	 &cfg.ssl_proxy_port ,
	 gettext_nop("\t-ssl_proxy $site[:$port]\n"
		     "\t                   - ssl proxy server\n") ,
	} ,
#endif
	{
	 "-gopher_httpgw" ,
	 "GopherViaHTTPProxy:" ,
	 NULL ,
	 PARAM_PBOOL ,	
	 (void *) FALSE ,
	 &cfg.gopher_via_http ,
	 NULL ,	
	 NULL ,
	 gettext_nop("\t-gopher_httpgw/-nogopher_httpgw\n"
		     "\t                   - specified gopher proxy HTTP gateway for Gopher\n") ,

	} ,
	{
	 "-nogopher_httpgw" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,	
	 (void *) FALSE ,
	 &cfg.gopher_via_http ,
	 NULL ,	
	 NULL ,
	 NULL ,
	} , 
	{
	 "-ftp_httpgw" ,
	 "FTPViaHTTPProxy:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.ftp_via_http ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ftp_httpgw/-noftp_httpgw\n"
		     "\t                   - specified ftp proxy is HTTP gateway for FTP\n") ,

	} ,
	{
	 "-noftp_httpgw" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,	
	 (void *) FALSE ,
	 &cfg.ftp_via_http ,	
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-ftp_dirtyproxy" ,
	 "FTPDirtyProxy:" ,	
	 NULL ,
	 PARAM_PBOOL ,	
	 (void *) FALSE ,
	 &cfg.ftp_dirtyp ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ftp_dirtyproxy/-noftp_dirtyproxy\n"
		     "\t                   - use CONNECT request to HTTP proxy for FTP\n"
		     "\t                     connections\n") ,
	} ,
	{
	 "-noftp_dirtyproxy" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.ftp_dirtyp ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
#ifdef I_FACE
#ifdef WITH_TREE
	{
	 "-browser" ,
	 "Browser:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.browser ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-browser $str      - your preffered browser programm\n") ,
	} ,
#endif
	{
	 "-xmaxlog" ,
	 "XMaxLogSize:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) NULL ,
	 &cfg.xlogsize ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-xmaxlog           - maximal length of log window\n") ,
	} ,
#endif
	{
	 "-dumpscn" ,
	 NULL ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.save_scn ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-dumpscn $str      - save scenario in scndir with name $str\n") ,
	} ,
	{
	 "-maxsize" ,
	 "MaxSize:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) 0 ,
	 &cfg.condition.max_size ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-maxsize $nr       - maximal allowed size of document in bytes\n") ,
	} ,
	{
	 "-minsize" ,
	 "MinSize:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) 0 ,
	 &cfg.condition.min_size ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-minsize $nr       - minimal allowed size of document in bytes\n") ,
	} ,
	{
	 "-http_proxy_pass" ,
	 "HTTPProxyPass:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.http_proxy_pass ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-http_proxy_pass $str\n"
		     "\t                   - password for HTTP proxy authorization\n") ,
	} ,
	{
	 "-http_proxy_user" ,
	 "HTTPProxyUser:" ,
	 NULL ,	
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.http_proxy_user ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-http_proxy_user $str\n"
		     "\t                   - user name for HTTP proxy authorization\n") ,
	} ,
	{
	 "-http_proxy_auth" ,
	 "HTTPProxyAuth:" ,
	 NULL ,
	 PARAM_AUTHSCH ,
	 (void *) 2 ,
	 &cfg.proxy_auth_scheme ,
	 NULL ,	
	 NULL ,
	 gettext_nop("\t-http_proxy_auth 1/2/3\n"
		     "\t                   - authorization scheme for HTTP proxy authorization\n") ,
	} ,
	{
	 "-logfile" ,
	 "LogFile:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.logfile ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-logfile $file     - name of file where mesages are stored\n") ,
	} ,
	{
	 "-slogfile" ,
	 "SLogFile:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.short_logfile ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-slogfile $file    - name of file where short log will be stored\n") ,
	} ,
	{
	 "-stime" ,
	 "ShowDownloadTime:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.show_time ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-stime/-nostime    - write starting and ending time of transfer\n") ,
	} ,
	{
	 "-nostime" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.show_time ,
	 NULL ,
	 NULL ,
	 NULL ,
	},
	{
	 "-remove_old" ,
	 "RemoveOldDocuments:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.remove_old ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-remove_old/-noremove_old\n"
		     "\t                   - remove improper files or directories while running\n"
		     "\t                     in sync mode\n") ,

	},
	{
	 "-noremove_old" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.remove_old ,
	 NULL ,
	 NULL ,
	 NULL ,
	},
	{
	 "-auth_file" ,
	 "AuthFile:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.auth_file ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-auth_file $file   - file where you have stored your auth infos\n"
		     "\t                     see manual for format description\n"),
	},
	{
	 "-base_level" ,
	 "BaseLevel:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) 0 ,
	 &cfg.base_level ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-base_level $nr    - number of levels of directory tree to omit from top\n") ,
	},
	{
	 "-ftp_active" ,
	 "ActiveFTPData:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.ftp_activec ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ftp_active        - select active FTP data connection\n") ,
	} ,
	{
	 "-ftp_passive" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.ftp_activec ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-ftp_passive       - select passive FTP data connection\n") ,
	} ,
#ifdef GETTEXT_NLS
	{
	 "-msgcat" ,
	 "NLSMessageCatalogDir:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.msgcatd ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-msgcat $dir       - directory where message catalogs are stored\n") ,
	} ,
	{
	 "-language" ,
	 "Language:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.language ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-language $str     - set language for messages\n")
	} ,
#endif
	{
	 "-quiet" ,
	 "Quiet:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.quiet ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-quiet             - don't show output messages\n")
	} ,
	{
	 "-verbose" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.quiet ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-verbose           - show output messages (default)\n")
	} ,
	{
	 "-newer_than" ,
	 "NewerThan:" ,
	 NULL ,
	 PARAM_TIME ,
	 (void *) 0 ,
	 &cfg.condition.btime ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-newer_than $time  - download only documents newer than $time\n"
		     "\t                     format of $time: YYYY.MM.DD.hh:mm\n")
	} ,
	{
	 "-older_than" ,
	 "OlderThan:" ,
	 NULL ,
	 PARAM_TIME ,
	 (void *) 0 ,
	 &cfg.condition.etime ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-older_than $time  - download only documents older than $time\n"
		     "\t                     format of $time: YYYY.MM.DD.hh:mm\n")
	} ,
	{
	 "-schedule" ,
	 NULL ,
	 NULL ,
	 PARAM_TIME ,
	 (void *) 0 ,
	 &cfg.schtime ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-schedule $time    - schedule pavuk start at $time\n"
		     "\t                     format of $time: YYYY.MM.DD.hh:mm\n")
	} ,
	{
	 "-reschedule" ,
	 "Reschedule:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) 0 ,
	 &cfg.reschedh ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-reschedule $nr    - number of hours between two runs\n"
		     "\t                     for cyclic scheduling\n")
	} ,
	{
	 "-dont_leave_site" ,
	 "DontLeaveSite:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.condition.dont_leave_site ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-dont_leave_site/-levave_site\n"
		     "\t                   - don't/leave site of starting URL\n")
 	} ,
	{
	 "-leave_site" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.condition.dont_leave_site ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-dont_leave_dir" ,
	 "DontLeaveDir:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.condition.dont_leave_dir ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-dont_leave_dir/-leave_dir\n"
		     "\t                   - don't/leave directory of starting URL\n")
	} ,
	{
	 "-leave_dir" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.condition.dont_leave_dir ,
	 NULL ,
	 NULL ,
	 NULL
	} ,
	{
	 "-preserve_time" ,
	 "PreserveTime:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.preserve_time ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-preserve_time/-nopreserve_time\n"
		     "\t                   - preserve document modification time\n")
	} ,
	{
	 "-nopreserve_time" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.preserve_time ,
	 NULL ,
	 NULL ,
	 NULL
	} ,
	{
	 "-preserve_perm" ,
	 "PreservePermisions:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.preserve_perm ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-preserve_perm/-nopreserve_perm\n"
		     "\t                   - preserve document permisions\n")
	} ,
	{
	 "-nopreserve_perm" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.preserve_perm ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-preserve_slinks" ,
	 "PreserveAbsoluteSymlinks:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.preserve_links ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-preserve_slinks/-nopreserve_slinks\n"
		     "\t                   - preserve absolute symlinks\n")
	} ,
	{
	 "-nopreserve_slinks" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.preserve_links ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-leave_level" ,
	 "LeaveLevel:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) 0 ,
	 &cfg.condition.leave_level ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-leave_level $nr   - how many tree levels leave from starting\n"
		     "\t                     site , (0 == don't care) default 0\n")
	} ,
	{
	 "-cookie_file" ,
	 "CookieFile:" ,
	 NULL ,
	 PARAM_PATH ,
	 (void *) NULL ,
	 &cfg.cookie_file ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-cookie_file $file - file where are stored cookies in Netscapes format\n"),
	 
	} ,
	{
	 "-cookie_send" ,
	 "CookieSend:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.send_cookies ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-cookie_send/-nocookie_send\n"
		     "\t                   - send cookie info in HTTP request\n")
	} ,
	{
	 "-nocookie_send" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.send_cookies ,
	 NULL ,
	 NULL ,
	 NULL
	} ,
	{
	 "-cookie_recv" ,
	 "CookieRecv:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.recv_cookies ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-cookie_recv/-nocookie_recv\n"
		     "\t                   - accept cookies from HTTP response\n")
	} ,
	{
	 "-nocookie_recv" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.recv_cookies ,
	 NULL ,
	 NULL ,
	 NULL
	} ,
	{
	 "-cookie_update" ,
	 "CookieUpdate:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.update_cookies ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-cookie_update/-nocookie_update\n"
		     "\t                   - update cookies in cookies file\n")
	} ,
	{
	 "-nocookie_update" ,
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) FALSE ,
	 &cfg.update_cookies ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-cookies_max" ,
	 "CookiesMax:" ,
	 NULL ,
	 PARAM_NUM ,
	 (void *) 0 ,
	 &cfg.cookies_max ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-cookies_max $nr   - maximal number of cookies in cookie cache\n")
	} ,
	{
	 "-disabled_cookie_domains" , 
	 "DisabledCookieDomains:" , 
	 NULL , 
	 PARAM_STRLIST , 
	 (void *) NULL , 
	 &cfg.cookies_disabled_domains, 
	 (void *) NULL , 
	 NULL ,
	 gettext_nop("\t-disabled_cookie_domains $list\n"
		     "\t                   - comma-separated list of disabled cookie domains\n") ,
	} ,
	{
	 "-disable_html_tag" ,
	 "DisableHTMLTag:" ,
	 NULL ,
	 PARAM_HTMLTAG ,
	 (void *) TRUE ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-disable_html_tag $TAG,[$ATTRIB][;...]\n"
		     "\t                   - disable proccessing of URLs from\n"
		     "\t                     attribute $ATTRIB of HTML tag $TAG\n")
	} ,
	{
	 "-enable_html_tag" ,
	 "EnableHTMLTag:" ,
	 NULL ,
	 PARAM_HTMLTAG ,
	 (void *) FALSE ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-enable_html_tag $TAG,[$ATTRIB][;...]\n"
		     "\t                   - enable proccessing of URLs from\n"
		     "\t                     attribute $ATTRIB of HTML tag $TAG\n")
	} , 
#ifdef I_FACE
	{
	 "-gui_font" ,
	 "GUIFont:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.fontname ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-gui_font $font    - font name used in GUI interface\n")
	} ,
#endif
	{
	 "-user_condition" ,
	 "UserCondition:" ,
	 NULL , 
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.condition.uexit,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-user_condition $str\n"
		     "\t                   - user exit script used to restrict some URLs\n"
		     "\t                     for more informations see manpage\n") ,
	} ,
	{
	 "-tr_del_chr" ,
	 "TrDeleteChar:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.tr_del_chr ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-tr_del_chr $str   - characters that will be deleted from\n"
		     "\t                     local filename\n") ,
	} ,
	{
	 "-tr_str_str" ,
	 "TrStrToStr:" ,
	 NULL ,
	 PARAM_TWO_QSTR ,
	 (void *) NULL ,
	 &cfg.tr_str_s1 ,
	 (void *) NULL ,
	 &cfg.tr_str_s2 ,
	 gettext_nop("\t-tr_str_str $str1 $str2\n"
		     "\t                   - translate $str1 to $str2 in local filename\n") ,
	} ,
	{
	 "-tr_chr_chr" ,
	 "TrChrToChr:" ,
	 NULL ,
	 PARAM_TWO_QSTR ,
	 (void *) NULL ,
	 &cfg.tr_chr_s1 ,
	 (void *) NULL ,
	 &cfg.tr_chr_s2 ,
	 gettext_nop("\t-tr_chr_chr $chrset1 $chrset2\n"
		     "\t                   - translate $chrset1 to $chrset2 in local filename\n") ,
	} ,
	{
	 "-index_name" ,
	 "IndexName:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.index_name ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-index_name $str   - name of directory index instead of _._.html\n") ,
	} ,
	{
	 "-store_name" ,
	 "StoreName:" ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.store_name ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-store_name $str   - filename for document downloaded with\n"
		     "\t                     -mode singlepage\n") ,
	} ,
	{
	 "-htDig" ,
	 NULL ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) FALSE ,
	 &cfg.htdig ,
	 NULL ,
	 NULL ,
         NULL ,
	} ,
	{
	 "-check_size",
	 "CheckSize:" ,
	 NULL ,
	 PARAM_PBOOL ,
	 (void *) TRUE ,
	 &cfg.check_size ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-check_size/-nocheck_size\n"
		     "\t                   - compare received size of file with\n"
		     "\t                     that provided by remote server\n")
	} ,
	{
	 "-nocheck_size",
	 NULL ,
	 NULL ,
	 PARAM_NBOOL ,
	 (void *) TRUE ,
	 &cfg.check_size ,
	 NULL ,
	 NULL ,
	 NULL ,
	} ,
	{
	 "-urls_file",
	 NULL ,
	 NULL ,
	 PARAM_STR ,
	 (void *) NULL ,
	 &cfg.urls_file ,
	 NULL ,
	 NULL ,
	 gettext_nop("\t-urls_file $file   - URLs should be read from file, until\n"
		     "\t                     line with single \".\" occurss in input\n") 
	} ,
};


/********************************/
/* vypis prepinacov programu	*/
/********************************/
void usage(prg)
char *prg;
{
	int i;

	printf(gettext("Usage:  %s  [options]  [any number of URLS]\npavuk-%s\n") , prg , VERSION);

	for(i = 0 ; i < NUM_ELEM(params) ; i++)
	{
		if (params[i].help) printf(gettext(params[i].help));
	}

	printf(gettext("\t-h                 - help\n"));
	printf(gettext("\t-v                 - print version number\n"));

	exit(1);
}

static void usage_short(prg)
char *prg;
{

	printf("pavuk-%s\n" , VERSION);
	printf(gettext("Type \"%s -h\" for long help\n") , prg);


	exit(1);
}

char *get_1qstr(str)
char *str;
{
	static char *p;
	char *pom1 = NULL , *pom2 = NULL;
	int found;

	if (str) p = str;

	for ( ; p && *p && (!pom1 || !pom2) ; p++)
	{
		found = FALSE;

		if (*p == '\"')
		{
			if ((p == str) || (*(p-1) != '\\'))
				found = TRUE;
		}

		if (!pom1 && found)
			pom1 = p+1;
		else if (!pom2 && found)
			pom2 = p;
	}
	if (pom1 && pom2)
	{
		*pom2 = '\0';
		return pom1;
	}
	else
	{
		p = NULL;
		return NULL;
	}
}

int htmltag_set_disabled(tagstr , disable)
char *tagstr;
int disable;
{
	int i,j;
	bool tfound,afound;
	char *tag;
	char *attrib;
	char *pom = new_string(tagstr);

	tag = strtokc(pom , ',');
	attrib = strtokc(NULL , ';');

	while (tag) 
	{
	    tfound = FALSE;
	    afound = FALSE;
	    for (i = 0 ; i < NUM_ELEM(linky) ; i++)
	    {
		if (!strcasecmp(linky[i].tag , tag))
		{
			tfound = TRUE;
                	for(j = 0 ; linky[i].attribs[j].attrib ; j++)
                	{
				if (attrib && *attrib)
				{
					if (!strcasecmp(linky[i].attribs[j].attrib , attrib))
					{
						afound = TRUE;
						if (disable)
							linky[i].attribs[j].stat |= LINK_DISABLED;
						else
							linky[i].attribs[j].stat &= ~LINK_DISABLED;
						break;
					}
				}
				else
				{
					afound = TRUE;
					if (disable)
						linky[i].attribs[j].stat |= LINK_DISABLED;
					else
						linky[i].attribs[j].stat &= ~LINK_DISABLED;
				}
			}
			break;
		}
	    }
	    if (!(tfound && afound))
	    {
		xprintf(0 , gettext("HTML tag not supported : %s.%s\n") , tag , attrib ? attrib : "(null)");
	    }

	    tag = strtokc(NULL , ',');
	    attrib = strtokc(NULL , ';');
	}

	_free(pom);
	return -1;
}

void cfg_set_to_default(i)
int i;
{
	char **p;
	int x,j;

	switch (params[i].type)
	{
		case PARAM_NUM:
			*((int *) params[i].val_adr) = (int) params[i].default_val;
		break;
		case PARAM_PBOOL:
			*((bool *) params[i].val_adr) = (bool)(int) params[i].default_val;
		break;
		case PARAM_NBOOL:
			*((bool *) params[i].val_adr) = (bool)(int) params[i].default_val;
		break;
		case PARAM_PATH:
		case PARAM_STR:
			_free(*((char **) params[i].val_adr));
			*((char **) params[i].val_adr) = (char *) params[i].default_val;
		break;
		case PARAM_STRLIST:		
			for (p = *((char ***) params[i].val_adr) ; p && *p ; p++)
				_free(*p);
			_free(*(char ***) params[i].val_adr);

			*((char ***) params[i].val_adr) = (char **) params[i].default_val;
			if (params[i].mval_adr)
				*((bool *) params[i].mval_adr) = (bool)(int) params[i].mdefault_val;				
		break;
		case PARAM_CONN:
			_free(*((char **) params[i].val_adr));
			*((char **) params[i].val_adr) = (char *) params[i].default_val;
			if (params[i].mval_adr)
				*((int *) params[i].mval_adr) = (int) params[i].mdefault_val;
		break;
		case PARAM_AUTHSCH:
			*((int *) params[i].val_adr) = (int) params[i].default_val;
		break;
		case PARAM_MODE:
			*((int *) params[i].val_adr) = (int) params[i].default_val;
		break;
		case PARAM_TIME:
			*((time_t *) params[i].val_adr) = (time_t)0;
		break;
		case PARAM_HTMLTAG:
	    		for (x = 0 ; x < NUM_ELEM(linky) ; x++)
                		for(j = 0 ; linky[x].attribs[j].attrib ; j++)
					linky[x].attribs[j].stat &= ~LINK_DISABLED;
		break;
		case PARAM_TWO_QSTR:
			*((char **) params[i].val_adr) = (char *) params[i].default_val;
			*((char **) params[i].mval_adr) = (char *) params[i].mdefault_val;
		break;
	}
}

void cfg_setup_default()
{
	int i,x,j;

	for (i = 0; i < NUM_ELEM(params) ; i++)
	{
		switch (params[i].type)
		{
			case PARAM_NUM:
				*((int *) params[i].val_adr) = (int) params[i].default_val;
			break;
			case PARAM_PBOOL:
				*((bool *) params[i].val_adr) = (bool)(int) params[i].default_val;
			break;
			case PARAM_NBOOL:
				*((bool *) params[i].val_adr) = (bool)(int) params[i].default_val;
			break;
			case PARAM_PATH:
			case PARAM_STR:
				*((char **) params[i].val_adr) = (char *) params[i].default_val;
			break;
			case PARAM_STRLIST:
				*((char ***) params[i].val_adr) = (char **) params[i].default_val;
				if (params[i].mval_adr)
					*((bool *) params[i].mval_adr) = (bool)(int) params[i].mdefault_val;				
			break;
			case PARAM_CONN:
				*((char **) params[i].val_adr) = (char *) params[i].default_val;
				if (params[i].mval_adr)
					*((int *) params[i].mval_adr) = (int) params[i].mdefault_val;
			break;
			case PARAM_AUTHSCH:
				*((int *) params[i].val_adr) = (int) params[i].default_val;
			break;
			case PARAM_MODE:
				*((int *) params[i].val_adr) = (int) params[i].default_val;
			break;
			case PARAM_TIME:
				*((time_t *) params[i].val_adr) = (time_t)0;
			break;
			case PARAM_HTMLTAG:
	    			for (x = 0 ; x < NUM_ELEM(linky) ; x++)
                			for(j = 0 ; linky[x].attribs[j].attrib ; j++)
						linky[x].attribs[j].stat &= ~LINK_DISABLED;
			break;
			case PARAM_TWO_QSTR:
				*((char **) params[i].val_adr) = (char *) params[i].default_val;
				*((char **) params[i].mval_adr) = (char *) params[i].mdefault_val;
		}
	}
}

void cfg_setup_cmdln(argc , argv)
int argc;
char **argv;
{
	int i , j;
	bool found;
	int nr=0;
	char *p=NULL,*p2=NULL;
	char **pl=NULL;

	for (j = 1 ; j < argc ; j++)
	{
		found = FALSE;
		for (i = 0; i < NUM_ELEM(params) ; i++)
		{
			if (!strcasecmp(argv[j] , params[i].par_cmd))
			{
				found = TRUE;
				switch (params[i].type)
				{
					case PARAM_NUM:
						j++;
						if (j < argc)
						{
							nr = _atoi(argv[j]);
							if (errno == ERANGE)
							{
								printf(gettext("Please specify number with parameter \"%s\"\n") , params[i].par_cmd);
								usage_short(argv[0]);
							}
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						*((int *) params[i].val_adr) = nr;
					break;
					case PARAM_PBOOL:
						*((bool *) params[i].val_adr) = TRUE;
					break;
					case PARAM_NBOOL:
						*((bool *) params[i].val_adr) = FALSE;
					break;
					case PARAM_STR:
						j++;
						if (j < argc)
						{
							p = new_string(argv[j]);
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						*((char **) params[i].val_adr) = p;
					break;
					case PARAM_PATH:
						j++;
						if (j < argc)
						{
#ifdef _WIN32
							p = cvt_win32_to_unix_path(argv[j]);
#else
							p = get_abs_file_path(argv[j]);
#endif
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						*((char **) params[i].val_adr) = p;
					break;
					case PARAM_STRLIST:
						j++;
						if (j < argc)
						{
							pl = make_list(argv[j] , ",");
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						if (params[i].mval_adr)
							*((bool *) params[i].mval_adr) = (bool)(int) params[i].mdefault_val;
						*((char ***) params[i].val_adr) = pl;
					break;
					case PARAM_CONN:
						j++;
						if (j < argc)
						{
							p = new_string(argv[j]);
							strtok(p , ":");
							p2 = strtok(NULL , "");
							if (p2)
							{
								nr = _atoi(p2);
								if (errno == ERANGE)
								{
									struct servent *se;
									
									if ((se = getservbyname(p2 , "tcp")))
									{
										nr = se->s_port;
									}
									else
									{
										printf(gettext("Unknown port \"%s\"\n") , p2);
									}
								}
								if (params[i].mval_adr)
									*((int *) params[i].mval_adr) = (int) nr;
							}
							*((char **) params[i].val_adr) = new_string(p);
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}

					break;
					case PARAM_AUTHSCH:
						j++;
						if (j < argc)
						{
							nr = _atoi(argv[j]);
							if (errno == ERANGE)
							{
								printf(gettext("Bad auth scheme \"%s\"\n") , argv[j]);
								usage_short(argv[0]);
							}
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						if (!(nr == 1 || nr == 2 || nr == 3))
						{
							printf(gettext("Bad auth scheme \"%s\"\n") , argv[j]);
							usage_short(argv[0]);
						}
						else
						*((int *) params[i].val_adr) = nr;
					break;
					case PARAM_MODE:
						j++;
						if (j < argc)
						{
							if (!strcasecmp(argv[j] , "normal"))
							{
								cfg.mode = MODE_NORMAL;
							}
							else if (!strcasecmp(argv[j] , "linkupdate"))
							{
								cfg.mode = MODE_LNUPD;
							}
							else if (!strcasecmp(argv[j] , "sync"))
							{
								cfg.mode = MODE_SYNC;
							}
							else if (!strcasecmp(argv[j] , "singlepage"))
							{
								cfg.mode = MODE_SINGLE;
							}
							else if (!strcasecmp(argv[j] , "singlereget"))
							{
								cfg.mode = MODE_SREGET;
							}
							else if (!strcasecmp(argv[j] , "resumeregets"))
							{
								cfg.mode = MODE_RESUME;
							}
							else if (!strcasecmp(argv[j] , "dontstore"))
							{
								cfg.mode = MODE_NOSTORE;
							}
							else
							{
								printf(gettext("Unknow operation mode \"%s\"\n") , argv[j]);
								usage_short(argv[0]);
							}
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
					break;
					case PARAM_TIME:
						j++;
						if (j < argc)
						{
							*((time_t *) params[i].val_adr) = time_scn_cmd(argv[j]);
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
					break;					
					case PARAM_HTMLTAG:
						j++;
						htmltag_set_disabled(argv[j] , params[i].default_val);
					break;
					case PARAM_TWO_QSTR:
						j++;
						if (j < argc)
						{
							p = new_string(argv[j]);
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						*((char **) params[i].val_adr) = p;

						j++;
						if (j < argc)
						{
							p = new_string(argv[j]);
						}
						else
						{
							printf(gettext("Not enough number of parameters \"%s\"\n") , params[i].par_cmd);
							usage_short(argv[0]);
						}
						*((char **) params[i].mval_adr) = p;
					break;
				}
			}
		}
		if (!found)
		{
			if (!strcasecmp(argv[j] , "-h")) usage(argv[0]);
			else if (!strcasecmp(argv[j] , "-v")) 
			{
				printf("pavuk %s\n",VERSION);
				printf(gettext("Optional features available :\n"));
#ifdef DEBUG
				printf(gettext(" - Debug mode\n"));
#endif

#ifdef GETTEXT_NLS
				printf(gettext(" - GNU gettext internationalization of messages\n"));
#endif

#ifdef HAVE_FLOCK
				printf(gettext(" - flock() document locking\n"));
#endif

#ifdef HAVE_FCNTL_LOCK
				printf(gettext(" - fcntl() document locking\n"));
#endif

#ifdef I_FACE
#ifdef GTK_FACE
				printf(gettext(" - Gtk GUI interface\n"));
#endif
#ifdef X_FACE
				printf(gettext(" - Xt GUI interface\n"));
#endif 
#ifdef WITH_TREE
				printf(gettext(" - URL tree preview\n"));
#endif
#endif

#ifdef USE_SSL
				printf(gettext(" - http over SSL\n"));
#endif

#ifdef SOCKS
#ifdef SOCKS5
				printf(gettext(" - Socks v4 proxy support\n"));
#else
				printf(gettext(" - Socks v5 proxy support\n"));
#endif
#endif
				exit(0);
			}
			else if (*(argv[j]) == '-')
			{
				printf(gettext("Unknown switch %s\n") , argv[j]);
				usage_short(argv[0]);
			}
			else
			{
				cfg.prepurlstr = dllist_append(cfg.prepurlstr , new_string(argv[j]));
				cfg.total_cnt++;
			}
		}
	}
}

int cfg_load(filename)
char *filename;
{
	int i;
	bool found;
	int nr;
	char *p,*p2;
	char lnbuf[4096];
	char *lns;
	int sz;
	bufio *fd;

	if (!(fd = bufio_open(filename , O_BINARY | O_RDONLY)))
	{
		xperror(filename);
		return -1;
	}	

	while((sz = bufio_readln(fd , lnbuf , sizeof(lnbuf)) > 0))
	{
		strtok(lnbuf , "\n\r");
		for(lns = lnbuf ; *lns && isspace(*lns) ; lns++);
		if (*lns == '#' || !*lns) continue;

		found = FALSE;

		for (i = 0; i < NUM_ELEM(params) ; i++)
		{
			if (! params[i].par_entry) continue;
			if (!strncasecmp(lns , params[i].par_entry , strlen(params[i].par_entry)))
			{
				
				lns += strlen(params[i].par_entry);
				for( ; *lns && isspace(*lns) ; lns++);
				for(p = lns + strlen(lns) - 1 ; p >= lns && isspace(*p) ; p--) *p = '\0';

				if (!*lns) 
				{
					cfg_set_to_default(i);
					continue;
				}

				found = TRUE;
				switch (params[i].type)
				{
					case PARAM_NUM:
						nr = _atoi(lns);
						if (errno == ERANGE)
						{
							xprintf(0 , gettext("Please specify number \"%s\"\n") , params[i].par_entry);
						}
						else
							*((int *) params[i].val_adr) = nr;
					break;
					case PARAM_PBOOL:
					case PARAM_NBOOL:
						if (!strcasecmp(lns , "false"))
						{
							*((bool *) params[i].val_adr) = FALSE;
						}
						else if (!strcasecmp(lns , "true"))
						{
							*((bool *) params[i].val_adr) = TRUE;
						}
						else xprintf(0 , gettext("Only \"true\" & \"false\" is allowed : \"%s\"\n") , lns);
					break;
					case PARAM_STR:
						*((char **) params[i].val_adr) = *lns ? new_string(lns) : NULL;
					break;
					case PARAM_PATH:
#ifdef _WIN32
						*((char **) params[i].val_adr) = *lns ? cvt_win32_to_unix_path(lns) : NULL;
#else
						*((char **) params[i].val_adr) = *lns ? get_abs_file_path(lns) : NULL;
#endif
					break;
					case PARAM_STRLIST:
						if (params[i].mval_adr)
							*((bool *) params[i].mval_adr) = (bool)(int) params[i].mdefault_val;
						*((char ***) params[i].val_adr) = *lns ? make_list(lns , ",") : NULL;
					break;
					case PARAM_CONN:
						p = new_string(lns);
						strtok(p , ":");
						p2 = strtok(NULL , "");
						if (p2)
						{
							nr = _atoi(p2);
							if (errno == ERANGE)
							{
								struct servent *se;

								if ((se = getservbyname(p2 , "tcp")))
								{
									nr = se->s_port;
								}
								else
								{
									xprintf(0 , gettext("Unknown port \"%s\"\n") , p2);
								}
							}
							if (params[i].mval_adr)
								*((int *) params[i].mval_adr) = (int) nr;
						}
						*((char **) params[i].val_adr) = new_string(p);
						_free(p);
					break;
					case PARAM_AUTHSCH:
						nr = _atoi(lns);
						if (errno == ERANGE)
						{
							xprintf(0 , gettext("Bad auth scheme \"%s\"\n") , lns);
						}
						if (!(nr == 1 || nr == 2))
						{
							xprintf(0 , gettext("Bad auth scheme \"%s\"\n") , lns);
						}
						else
							*((int *) params[i].val_adr) = nr;
					break;
					case PARAM_MODE:
						if (!strcasecmp(lns , "normal"))
						{
							cfg.mode = MODE_NORMAL;
						}
						else if (!strcasecmp(lns , "linkupdate"))
						{
							cfg.mode = MODE_LNUPD;
						}
						else if (!strcasecmp(lns , "sync"))
						{
							cfg.mode = MODE_SYNC;
						}
						else if (!strcasecmp(lns , "singlepage"))
						{
							cfg.mode = MODE_SINGLE;
						}
						else if (!strcasecmp(lns , "singlereget"))
						{
							cfg.mode = MODE_SREGET;
						}
						else if (!strcasecmp(lns , "resumeregets"))
						{
							cfg.mode = MODE_RESUME;
						}
						else
						{
							xprintf(0 , gettext("Unknow operation mode \"%s\"\n") , lns);
						}
					break;
					case PARAM_TIME:
						*((time_t *) params[i].val_adr) = time_scn_cmd(lns);
						if (!params[i].val_adr)
						{
							xprintf(0 , gettext("Bad time parameter \"%s\"\n") , lns);
						}
					break;
					case PARAM_HTMLTAG:
						htmltag_set_disabled(lns , params[i].default_val);
					break;
					case PARAM_TWO_QSTR:
					{
						if (lns && *lns)
						{
							char *xp = new_string(lns);

							*((char **) params[i].val_adr) = new_string(get_1qstr(xp));
							*((char **) params[i].mval_adr) = new_string(get_1qstr(NULL));

							_free(xp);
						}
						else
						{
							*((char **) params[i].val_adr) = NULL;
							*((char **) params[i].mval_adr) = NULL;
						}
					}
					break;
				}
			}
		}
		if (!found && !strncasecmp(lns , "URL:" , 4))
		{
			lns += 4;
			for( ; *lns && isspace(*lns) ; lns++);
			for(p = lns + strlen(lns) ; *p && isspace(*p) ; p--) *p = '\0';
			if (!*lns) continue;

			cfg.prepurlstr = dllist_append(cfg.prepurlstr , new_string(lns));
			cfg.total_cnt++;
		}
		else if (!found) xprintf(0 , gettext("Unable to parse \"%s\"\n") , lns);
	}
	bufio_close(fd);

	return 0;
}

void cfg_load_setup()
{
	char *p;

#ifdef DEFAULTRC
	if (!access(DEFAULTRC , R_OK))
		cfg_load(DEFAULTRC);
#endif

	if ((p = getenv("HOME")))
	{
		char pom[PATH_MAX];
		sprintf(pom , "%s/%s" , p , ".pavukrc");
		if (!access(pom , R_OK))
			cfg_load(pom);
	}
}

int cfg_dump(filename)
char *filename;
{
	int i,j;
	int fd;
	char pom[8192];
	char pom2[20];
	char *p=NULL,**pl;

	if ((fd = open(filename , O_BINARY | O_CREAT | O_WRONLY , 0666)) < 0)
	{
		xperror(filename);
		return -1;
	}	

	ftruncate(fd , 0);
	if (cfg.prepurlstr)
	{
		dllist *dptr;

		for(dptr = cfg.prepurlstr ; dptr ; dptr = dptr->next)
		{
			write(fd , "URL: " , 5);
			write(fd , (char *)dptr->data , strlen((char *)dptr->data));
			write(fd , "\n" , 1);
		}
	}
	
	for (i = 0; i < NUM_ELEM(params) ; i++)
	{
		if (! params[i].par_entry ) continue;
		switch (params[i].type)
		{
			case PARAM_NUM:
				sprintf(pom , "%s %d\n" , params[i].par_entry , *((int *) params[i].val_adr));
				write(fd , pom , strlen(pom));
			break;
			case PARAM_NBOOL:
			case PARAM_PBOOL:
				if (*((bool *) params[i].val_adr))
					sprintf(pom , "%s true\n" , params[i].par_entry);
				else
					sprintf(pom , "%s false\n" , params[i].par_entry);
				write(fd , pom , strlen(pom));
			break;
			case PARAM_PATH:
			case PARAM_STR:
				if (*((char **) params[i].val_adr))
				{
					sprintf(pom , "%s %s\n" , params[i].par_entry , *((char **) params[i].val_adr));
					write(fd , pom , strlen(pom));
				}
			break;
			case PARAM_STRLIST:
				if (!params[i].mval_adr || (params[i].mval_adr &&  
				    (*((bool *) params[i].mval_adr) == (bool)(int) params[i].mdefault_val)))
				{
					pl = *((char ***) params[i].val_adr);
					if (pl && pl[0])
					{
						sprintf(pom , "%s %s" , params[i].par_entry , pl[0]);
						write(fd , pom , strlen(pom));				

						j = 1;
						while (pl[j])
						{
							write(fd , "," , 1);
							write(fd , pl[j] , strlen(pl[j]));
							j++;
						}
						write(fd , "\n" , 1);
					}
				}
			break;
			case PARAM_CONN:
				if (*((char **) params[i].val_adr))
				{
					sprintf(pom , "%s %s:%d\n" , params[i].par_entry , *((char **) params[i].val_adr),*((int *) params[i].mval_adr));
					write(fd , pom , strlen(pom));
				}
			break;
			case PARAM_AUTHSCH:
				sprintf(pom , "%s %d\n" , params[i].par_entry , *((int *) params[i].val_adr));
				write(fd , pom , strlen(pom));
			break;
			case PARAM_MODE:
				switch (cfg.mode)
				{
					case MODE_NORMAL:
						p = "normal";
					break;
					case MODE_LNUPD:
						p = "linkupdate";
					break;
					case MODE_SYNC:
						p = "sync";
					break;
					case MODE_SINGLE:
						p = "singlepage";
					break;
					case MODE_SREGET:
						p = "singlereget";
					break;
					case MODE_RESUME:
						p = "resumeregets";
					break;
					case MODE_NOSTORE:
						p = "dontstore";
					break;
					default:
						p = "unknown";
				}
				sprintf(pom , "%s %s\n" , params[i].par_entry , p);
				write(fd , pom , strlen(pom));
			break;
			case PARAM_TIME:
				if (*((time_t *) params[i].val_adr))
				{
					strftime(pom2 , sizeof(pom2) , "%Y.%m.%d.%H:%M" , localtime((time_t *)params[i].val_adr));
					sprintf(pom , "%s %s\n" ,  params[i].par_entry , pom2);
					write(fd , pom , strlen(pom));
				}
			break;
			case PARAM_HTMLTAG:
			{
				int x,y;
				bool first = TRUE;

				sprintf(pom , "%s " , params[i].par_entry);
				for (x = 0 ; x < NUM_ELEM(linky) ; x++)
				{
					for(y = 0 ; linky[x].attribs[y].attrib ; y++)
					{
						if (!(linky[x].attribs[y].stat & LINK_DISABLED) ==
						    !(params[i].default_val))
						{
							if (!first) strcat(pom , ";");
							strcat(pom , linky[x].tag);
							strcat(pom , ",");
							strcat(pom , linky[x].attribs[y].attrib);
							first = FALSE;
						}
					}
				}
				strcat(pom , "\n");
				if (!first)
					write(fd , pom , strlen(pom));
			}
			break;
			case PARAM_TWO_QSTR:
				if (*((char **) params[i].val_adr))
				{
					sprintf(pom , "%s \"%s\" \"%s\"\n" , params[i].par_entry , *((char **) params[i].val_adr) , *((char **) params[i].mval_adr));
					write(fd , pom , strlen(pom));
				}
			break;
		}
	}
         	

	close(fd);
	return 0;
}

int cfg_dump_cmd(filename)
char *filename;
{
	int i,j;
	int fd;
	char pom [PATH_MAX];
	char pom2[20];
	char *p=NULL,**pl;

	if ((fd = open(filename , O_BINARY | O_CREAT  | O_WRONLY , 0666)) < 0)
	{
		xperror(filename);
		return -1;
	}	

	ftruncate(fd , 0);

	write(fd , cfg.prg_path , strlen(cfg.prg_path));
	
	write(fd , " " , 1);

	if (cfg.prepurlstr)
	{
		dllist *dptr;

		for(dptr = cfg.prepurlstr ; dptr ; dptr = dptr->next)
		{
			write(fd , " '" , 2);
			write(fd , (char *)dptr->data , strlen((char *)dptr->data));
			write(fd , "' " , 2);
		}
	}

	for (i = 0; i < NUM_ELEM(params) ; i++)
	{
		if (!params[i].par_entry && (params[i].type != PARAM_PBOOL) && (params[i].type != PARAM_NBOOL)) continue;
		switch (params[i].type)
		{
			case PARAM_NUM:
				sprintf(pom , " %s %d " , params[i].par_cmd , *((int *) params[i].val_adr));
				write(fd , pom , strlen(pom));
			break;
			case PARAM_NBOOL:
				if (!*((bool *) params[i].val_adr))
				{
					write(fd , " " , 1);
					write(fd ,  params[i].par_cmd , strlen(params[i].par_cmd));
					write(fd , " " , 1);
				}
			break;
			case PARAM_PBOOL:
				if (*((bool *) params[i].val_adr))
				{
					write(fd , " " , 1);
					write(fd , params[i].par_cmd , strlen(params[i].par_cmd));
					write(fd , " " , 1);
				}
			break;
			case PARAM_PATH:
			case PARAM_STR:
				if (*((char **) params[i].val_adr))
				{
					sprintf(pom , " %s '%s' " , params[i].par_cmd , *((char **) params[i].val_adr));
					write(fd , pom , strlen(pom));
				}
			break;
			case PARAM_STRLIST:
				if (!params[i].mval_adr || (params[i].mval_adr &&
				    (*((bool *) params[i].mval_adr) == (bool)(int) params[i].mdefault_val)))
				{
					pl = *((char ***) params[i].val_adr);
					if (pl && pl[0])
					{
						sprintf(pom , " %s '%s" , params[i].par_cmd , pl[0]);
						write(fd , pom , strlen(pom));				

						j = 1;
						while (pl[j])
						{
							write(fd , "," , 1);
							write(fd , pl[j] , strlen(pl[j]));
							j++;
						}
						write(fd , "' " , 2);
					}
				}
			break;
			case PARAM_CONN:
				if (*((char **) params[i].val_adr))
				{
					sprintf(pom , " %s %s:%d " , params[i].par_cmd , *((char **) params[i].val_adr),*((int *) params[i].mval_adr));
					write(fd , pom , strlen(pom));
				}
			break;
			case PARAM_AUTHSCH:
				sprintf(pom , " %s %d " , params[i].par_cmd , *((int *) params[i].val_adr));
				write(fd , pom , strlen(pom));
			break;
			case PARAM_MODE:
				switch (cfg.mode)
				{
					case MODE_NORMAL:
						p = "normal";
					break;
					case MODE_LNUPD:
						p = "linkupdate";
					break;
					case MODE_SYNC:
						p = "sync";
					break;
					case MODE_SINGLE:
						p = "singlepage";
					break;
					case MODE_SREGET:
						p = "singlereget";
					break;
					case MODE_RESUME:
						p = "resumeregets";
					break;
				}
				sprintf(pom , " %s %s " , params[i].par_cmd , p);
				write(fd , pom , strlen(pom));
			break;
			case PARAM_TIME:
				if (*((time_t *) params[i].val_adr))
				{
					strftime(pom2 , sizeof(pom2) , "%Y.%m.%d.%H:%M" , localtime((time_t *)params[i].val_adr));
					sprintf(pom , "%s %s\n" ,  params[i].par_cmd , pom2);
					write(fd , pom , strlen(pom));
				}
			break;
			case PARAM_HTMLTAG:
			{
				int x,y;
				bool first = TRUE;

				sprintf(pom , "%s '" , params[i].par_cmd);
				for (x = 0 ; x < NUM_ELEM(linky) ; x++)
				{
					for(y = 0 ; linky[x].attribs[y].attrib ; y++)
					{
						if (!(linky[x].attribs[y].stat & LINK_DISABLED) ==
						    !(params[i].default_val))
						{
							if (!first) strcat(pom , ";");
							strcat(pom , linky[x].tag);
							strcat(pom , ",");
							strcat(pom , linky[x].attribs[y].attrib);
							first = FALSE;
						}
					}
				}
				strcat(pom , "' ");
				if (!first)
					write(fd , pom , strlen(pom));
			}
			break;
			case PARAM_TWO_QSTR:
				if (*((char **) params[i].val_adr))
				{
					sprintf(pom , " %s '%s' '%s'" , params[i].par_cmd , *((char **) params[i].val_adr) , *((char **) params[i].mval_adr));
					write(fd , pom , strlen(pom));
				}
		}
	}

	close(fd);
	return 0;
}

void cfg_free_params()
{
	int i;

	for (i = 0 ; i < NUM_ELEM(params) ; i++)
		cfg_set_to_default(i);
}

