/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#ifndef _config_h_
#define _config_h_

#define NeedFunctionPrototypes 1

#ifndef HAVE_LONG_FILE_NAMES
#warning "This program can't run successfuly on machine without long filenames support"
#endif

#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "mode.h"

#ifdef X_FACE
#include "xinterface.h"
#include "icons.h"
#else
#ifdef GTK_FACE
#include "ginterface.h"
#include "icons.h"

#else
#define _Xt_Serve
#define _X_EscapeLoop
#define _Xt_ServeLoop
#define iface_set_what(w)
#endif
#endif

#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifdef NO_SETENV
#define setenv my_setenv
#define unsetenv(var) my_setenv(var , "" , 1)
#endif

#ifdef HAVE_FLOCK
#define _flock(fd , blk) my_flock(fd , blk)
#define _funlock(fd) flock(fd , LOCK_UN)
#else
#ifdef HAVE_FCNTL_LOCK
#define _flock(fd , blk) my_flock(fd , blk)
#define _funlock(fd) my_funlock(fd)
#else
#define _flock(fd , blk) 0
#define _funlock(fd) 0
#endif
#endif

#include <limits.h>
#ifndef PATH_MAX
#define PATH_MAX	2048
#endif
#ifndef INT_MAX
#define INT_MAX		2147483647
#endif
#ifndef USHRT_MAX
#define	USHRT_MAX	65535
#endif

#ifdef SOCKS

#ifdef HAVE_SOCKS_H
#include <socks.h>
#endif

#ifndef SOCKS5
#define connect 	Rconnect
#define bind		Rbind
#define accept		Raccept
#define listen		Rlisten
#define select		Rselect
#define gethostbyname	Rgethostbyname
#define getsockname	Rgetsockname
#endif

#endif


#include "bufio.h"
#include "url.h"
#include "condition.h"
#include "mode.h"
#include "doc.h"
#include "errcode.h"
#include "nls.h"
#include "authinfo.h"
#include "cookie.h"
#include "http.h"
#include "dllist.h"

#ifdef USE_SSL
#include <stdio.h>
#include <ssl.h>
#endif

#define HASH_TBL_SIZE	233

typedef struct _htbl {
	int	nr;
	url	**urls;
} htbl;


typedef struct {
/*** CMDline parameters ***/
	char 	*urls_file;	/*** urls will be rad from file ***/
	char	*cookie_file;	/*** cookie file in NS format ***/
	char	*auth_file;	/*** file for authentification informations ***/
	char	*save_scn;	/*** name for scenario saving ***/
	char    *scndir;	/*** directory where scenarios are stored ***/
	char	*scenario;	/*** scenario file to load ***/
	char	*subdir;	/*** subdirectory of cache_dir to focus  ***/
	char	*cache_dir;	/*** dir where your local tree is located ***/
	char 	*logfile;	/*** logging file ***/
	char	*short_logfile;	/*** newstyle log file name ***/

	int	base_level;	/*** base tree level from which is built local file tree ***/
	int	nretry;		/*** number of retries on error ***/
	int	nreget;		/*** number of regets ***/
	int	nredir;		/*** max number of redirections ***/
	int	ddays;		/*** delay in days for document reload ***/
	int	rollback;	/*** size to go back when reget-ing ***/
	int	sleep;		/*** sleeptime between transfers ***/
	int	cookies_max;	/*** maximal number of cookies ***/
	int	ctimeout;	/*** timeout for network communication ***/
	int	mode;		/*** working mode ***/
	int	reschedh;	/*** number of hours for rescheduling ***/
	struct tm *time;	/*** when to schedule execution time ***/
	time_t	schtime;	/*** cmdln param for scheduling time ***/

	char	*http_proxy;	/*** HTTP proxy server ***/
	int	http_proxy_port;
	char	*http_proxy_pass; /*** password for access to HTTP proxy server ***/
	char	*http_proxy_user; /*** username for access to HTTP proxy server ***/
	int	proxy_auth_scheme; /*** authorization scheme for proxy server ***/
	char	*ftp_proxy;	/*** FTP proxy server ***/
	int	ftp_proxy_port;
	bool	ftp_via_http;	/*** gatewaying FTP through HTTP proxy ***/
	bool	ftp_dirtyp;	/*** dirty FTP proxying via CONNECT request to HTTP proxy ***/
	char	*ssl_proxy;	/*** SSL tuneling proxy ***/
	int	ssl_proxy_port;
	char	*gopher_proxy;	/*** Gopher proxy ***/
	int	gopher_proxy_port;
	bool	gopher_via_http; /*** gatewaying Gopher trough HTTP proxy ***/
	char	*name_auth;	/*** meno pouzivatela pri HTTP autentifikacii ***/
	char	*passwd_auth;	/*** password for HTTP authorization ***/
	int	auth_scheme;	/*** authorization scheme 1- user, 2- Basic, 3- Digest ***/


	bool	store_index;	/*** store directory URLS as index files ***/
	bool	send_from;	/*** send From: header in request ***/
	bool	check_size;	/*** some broken HTTP servers send wrong Content-Length: header ***/
	bool	htdig;		/*** to dump HTTP response - used by htDig ***/
	bool	send_cookies;	/*** send available cookies in HTTP request ***/
	bool	recv_cookies;	/*** receive cookie infos ***/
	bool	update_cookies; /*** update cookie file ***/
	bool	preserve_time;	/*** preserve time of remote document ***/
	bool	preserve_perm;	/*** preserve permisions of remote document ***/
	bool	preserve_links;	/*** preserve absolute symlinks ***/
	bool	quiet;		/*** ? output messages ***/
	bool	auth_reuse_nonce;
	bool	auth_reuse_proxy_nonce;
	bool	show_time;	/*** show start and end time of download ? */
	bool	remove_old;	/*** remove old documents (when not occurrs on remote site) ***/
	bool	report_size;	/*** show progres internal ***/
	bool	progres;	/*** show retrieving progres , when on console ***/
	bool	ftp_activec;	/*** use active FTP data connection instead of passive ***/
	bool	ftp_list;	/*** retrieve FTP directory with LIST cmd insted of NLST ***/
	bool	ftp_html;	/*** process HTML files downloaded over FTP protocol ***/
	bool	cache;		/*** disallow caching of HTTP documents (on proxy cache) ***/
	bool	rewrite_links;	/*** indikacia ci ma maju byt odkazy v HTML dokumentoch prepisovane ***/
	bool	freget;		/*** force reget whole file when server doesn't support reget ***/
	bool	use_enc;	/*** indikacia ci sa ma pouzivat gzip/compress kodovanie pri prenose ***/

	char	*index_name;	/*** name of directory index file instead of _._.html ***/
	char	*store_name;    /*** filename of document transfered with -mode singlepage ***/
	char	*from;		/*** polozka From: alebo heslo pri anonymous FTP spojeni ***/
	char	*identity;	/*** User-agent: field contents ***/
	char	**accept_lang;	/*** list of preffered languages ***/
	char	**accept_chars;	/*** list of preffered character sets ***/
	char	**cookies_disabled_domains; /*** domains from which cookies are not acceptable ***/

	cond	condition;	/*** struktura v ktorej su obmedzujuce podmienky pre vyber dokumentov na spracovanie ***/
	dllist	*prepurlstr;	/*** zoznam nakonfigurovanych url ***/

	char	*tr_del_chr;	/*** set of characters to delete while doing name transformation ***/
	char	*tr_str_s1;	/*** strfrom in transformation ***/ 
	char	*tr_str_s2;	/*** strto in transformation ***/
	char	*tr_chr_s1;	/*** setfrom in transformation ***/
	char	*tr_chr_s2;	/*** setto in transformation ***/
#ifdef DEBUG
	bool	debug;		/*** debug mode on/off ***/
	int	debug_level;	/*** debug level ***/
#endif
#ifdef USE_SSL
	char	*ssl_cipher_list;
	char	*ssl_cert_file;
	char	*ssl_key_file;
	char	*ssl_cert_passwd;
#endif
#ifdef GETTEXT_NLS
	char	*msgcatd;	/*** explicit message catalog directory ***/
	char	*language;	/*** language for LC_MESSAGES ***/
#endif
#ifdef I_FACE
	char	*fontname;	/*** default font used in interface ***/
	int	xlogsize;	/*** max number of lines in LOG widget ***/
	bool	log_autoscroll;	/*** autoscroll of log window ***/
	bool	xi_face;	/*** indikacia ci je funkcne rozhranie X-windows ***/
	bool	run_iface;	/*** if immediately run download after start of pavuk in GUI interface ***/
#ifdef WITH_TREE
	char	*browser;	/*** command to execute your preffered browser ***/
#endif
#endif


/*** GLOBALdata ***/
	bool	check_limits;
	int	cookies_num;	/*** total number of_cookies ***/
	cookie_entry *cookies;	/*** cookies data ***/
	cookie_entry *last_cookie; /*** last cookie entry ***/
	int	cookie_changed; /*** was mem cookies changed after last update on disk ***/
	time_t	cookie_updated;	/*** time when cookies were last updated on disc ***/
	char	*local_host;	/*** hostname of local machine ***/
	int	dns_pid;	/*** pid of DNS resolver process ***/
	bufio	*dns_w;		/*** write side of pipe to DNS resolver ***/
	bufio	*dns_r;		/*** read side of pipe to DNS resolver ***/
	int	fail_cnt;	/*** counter for failed transfers ---> return code of pavuk ***/
	authinfo **authdata;	/*** authentification informations ***/
	int	authdata_num;	/*** number of entries in authdata ***/
	char	*prg_path;	/*** nazov programu pod akym bol spusteny ***/
	url	**urlstack;	/*** zoznam vsetkych spracovavnych URL ***/
	int	actual_url;	/*** index aktualne spracovavaneho URL ***/
	int	total_cnt;	/*** celkovy pocet URL v zozname ***/
	int	prev_mode;	/*** previous active mode ***/
	bool	mode_started;	/*** mode startup finisched ***/
	int	logfd;		/*** descriptor of log file ***/
	long	gmdiff;		/*** differenc between localtime and gmtime ***/
	bool	rbreak;		/*** immediately stop transfer ***/
	bool	stop;		/*** stop after this document will be processed ***/
	htbl	url_hash_tbl[HASH_TBL_SIZE];	/*** hash table for better performance URL lookup ***/

#ifdef X_FACE
	Xnfo	X;
#endif
#ifdef GTK_FACE
	Gtk_nfo	gtk;
#endif
#ifdef I_FACE
#ifdef WITH_TREE
	urltype_icon icon;	/*** icons in URL tree preview ***/
#endif
	bool	done;		/*** was done startup ? ***/
	bool	processing;	/*** some URL is actualy in processing ***/
#endif


/*** TEMPORARY information --> will go to separate structure in future ***/
#ifdef USE_SSL
	SSL	*ssl_con;
	BIO	*ssl_bio;
	SSL_CTX	*ssl_ctx;
	SSL_METHOD *ssl_method;
#endif
	http_digest_info *auth_digest;	/*** HTTP digest access authentification info ***/
	http_digest_info *auth_proxy_digest;	/*** HTTP digest access proxy authentification info ***/
	bool	ftp_fatal_err;	/*** was the FTP error fatal ? ***/
	int	origsize;	/*** size of local copy of document ***/
	int	total_len;	/*** celkova dlzka aktualneho dokumentu ***/
	int	errcode;	/*** chybovy kod ***/
	bufio	*ftp_control;	/*** deskriptor soketu riadiaceho FTP spojenia ***/
	bufio	*ftp_datasock;	/*** active FTP data connection socket ***/
	int	ftp_respc;	/*** last ftp response ***/
	int	sync_cnt;	/*** number of URLs to synchronize from begining of urlstack ***/ 
	int	t_error;	/*** indikacia chyby pri prenose ***/
	doc	*act_doc;	/*** aktualny dokument ***/
	bufio	*s_sock;	/*** deskriptor suboru do ktoreho sa online zapisuje ***/
	char *	lock_fn;	/*** name of lock file ***/
	bool	remove_lock;	/*** remove lock file if required ***/
	int	totsz;		/*** celkova velkost ak je znama ***/
} _config_struct;

extern _config_struct cfg;

typedef enum {
	PARAM_NUM ,
	PARAM_PBOOL ,
	PARAM_NBOOL ,
	PARAM_STR ,
	PARAM_STRLIST ,
	PARAM_CONN ,
	PARAM_AUTHSCH ,
	PARAM_MODE ,
	PARAM_PATH ,
	PARAM_TIME ,
	PARAM_HTMLTAG ,
	PARAM_TWO_QSTR ,
} par_type;

typedef struct _cfg_param {
	char		*par_cmd;
	char		*par_entry;
	char		*par_env;
	par_type	type;
	void		*default_val;
	void		*val_adr;
	void		*mdefault_val;
	void		*mval_adr;
	char		*help;
} cfg_param;

extern void report_error(int , char *);
extern void usage(char *);
extern void cfg_setup_default();
extern void cfg_setup_cmdln(int , char **);
extern int cfg_load(char *);
extern void cfg_load_setup();
extern int cfg_dump(char *);
extern int cfg_dump_cmd(char *);
extern void cfg_free_params();

#endif
