/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_FLOCK
#include <sys/file.h>
#endif
#include <sys/time.h>
#include <time.h>
#include <utime.h>

#ifdef HAVE_SYS_MODE_H
#include <sys/mode.h>
#endif

#include "config.h"
#include "url.h"
#include "doc.h"
#include "tools.h"
#include "mime.h"
#include "http.h"
#include "ftp.h"
#include "gopher.h"
#include "html.h"
#include "decode.h"
#include "abstract.h"
#include "mode.h"
#include "times.h"
#include "uexit.h"

#ifdef I_FACE
static void set_doc_info();
#endif



/********************************************************/
/* nacitanie dokumentu + specificke upravy		*/
/********************************************************/
int download_document(docu , load , b_lock)
doc * docu;
int load;
int b_lock;
{
	char buf[32768];
	char *p=NULL,*p1,*p3;
	int len,totallen = 0;
	http_response *resp;
	int retcode = 0;
	int sz;
	bufio *sock;
	int mres;
	bool fnd;
	char *fn,*inname;
	struct stat estat;
	bool havelock = FALSE;
	struct tm *modtime = docu->dtime ? new_tm(docu->dtime) : NULL;


	cfg.remove_lock = TRUE;
	cfg.lock_fn = NULL;
	docu->is_html = FALSE;
	docu->contents = NULL;
	docu->mime = NULL;
	docu->type_str = NULL;
	docu->save_online = FALSE;
	docu->size = 0;
	docu->rest_pos = 0;
	docu->stime = time(NULL);
	cfg.act_doc = docu;
	cfg.s_sock = NULL;
	cfg.t_error = 0;
	cfg.totsz = 0;
	cfg.origsize = 0;

	fn = url_to_filename(docu->doc_url);

	if (cfg.mode != MODE_SYNC)
	{

		if (docu->doc_url->type != FILET && !access(fn , R_OK))
		{
			if (!stat(fn , &estat))
			{
				if (!S_ISDIR(estat.st_mode))
				{
					docu->doc_url->status |= URL_REDIRECT;
				}
				else
				{
					char pom[2048];
					char *savepath = url_get_path(docu->doc_url);

					sprintf(pom , "%s/%s" , fn , cfg.index_name);
					if (!stat(pom , &estat))
					{
						if (!S_ISDIR(estat.st_mode))
		                                {
							_free(fn);
							fn = new_string(pom);

							url_remove_from_hash_tab(docu->doc_url);

							if (docu->doc_url->type != FILET)
								sprintf(pom , "%s/" , savepath);
							if (docu->doc_url->type == FTP)
								docu->doc_url->p.ftp.dir = TRUE;

							url_set_path(docu->doc_url , pom);
							_free(savepath);
                		                        docu->doc_url->status |= URL_REDIRECT;

							url_add_to_hash_tab(docu->doc_url);
                                		}
					}
				}
			}
		}

		if ((docu->doc_url->type == FILET || (docu->doc_url->status & URL_REDIRECT)) && !load)
		{
			if (!stat(fn , &estat))
			{
				if (S_ISDIR(estat.st_mode))
				{
					cfg.errcode= ERR_DIR_URL;
					_free(fn);
					_free(modtime);
					return -1;
				}
			}
			else
			{
				cfg.errcode= ERR_FILE_OPEN;
				_free(fn);
				_free(modtime);
				return -1;
			}

			if ((!cfg.ftp_html && docu->doc_url->type == FTP && !docu->doc_url->p.ftp.dir) || !file_is_html(fn))
			{
				_free(fn);
				_free(modtime);
				docu->is_html = FALSE;
				docu->save_online = TRUE;
				docu->size = estat.st_size;
#ifdef I_FACE
				if (cfg.xi_face)
					set_doc_info(docu);
#endif
				xprintf(1 , gettext("File redirect\n"));
				return 0;
			}
			else
			{
				if (!strcasecmp("css" , get_extension(fn)))
					docu->doc_url->status |= URL_STYLE;
				docu->is_html = TRUE;
			}
		}
	}
	else
	{
		if (!stat(fn , &estat))
		{
			cfg.origsize = estat.st_size;
		}
	}

	_free(fn);

	if ((inname = url_to_in_filename(docu->doc_url)))
	{
		if (!stat(inname , &estat) && !S_ISDIR(estat.st_mode))
		{
			if (!(cfg.s_sock = 
				bufio_copen(inname , O_BINARY | O_RDWR , S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR)))
			{
				xperror(inname);
				_free(inname);
				_free(modtime);
				cfg.errcode = ERR_STORE_DOC;
				return -1;
			}
			else
			{
#ifdef DEBUG
				if (cfg.debug)
					xprintf(1 , gettext("Locking %s\n") ,inname);
#endif
				if (_flock(bufio_getfd(cfg.s_sock) , b_lock))
				{
					cfg.errcode = ERR_LOCKED;
					_free(inname);
					_free(modtime);
					return -1;
				}
				havelock = TRUE;
				cfg.lock_fn = new_string(inname);
			}
			docu->rest_pos = estat.st_size - cfg.rollback;
			if (docu->rest_pos)
			{
				xprintf(1 , gettext("Trying to resume from position %d \n") ,
					docu->rest_pos);

				_free(modtime);
				modtime = new_tm(localtime(&estat.st_mtime));
				docu->stime = estat.st_mtime;
			}
		}
	}

	_free(inname);

	if (cfg.report_size)
		iface_set_what(gettext("Opening connection"));

	if (!(sock = get_data_socket(docu->doc_url , docu->rest_pos , modtime)))
	{
		cfg.remove_lock = FALSE;
		close_socket(sock , docu->doc_url , 0);
		_free(modtime);
		return -1;
	}

	_free(modtime);

	if (((docu->doc_url->type == FTP && !docu->doc_url->p.ftp.dir)
	   || (docu->doc_url->type == GOPHER && docu->doc_url->p.gopher.type != '1') || 
	   	(docu->doc_url->type == HTTP || docu->doc_url->type == HTTPS)) &&
	   	!(docu->doc_url->status & URL_REDIRECT))
	{
		docu->save_online = TRUE;
	}


	if ((inname = url_to_in_filename(docu->doc_url)))
	{
		if (cfg.mode != MODE_NOSTORE)
			makealldirs(inname);
	}
	else
	{
		close_socket(sock , docu->doc_url , 0);
		cfg.errcode = ERR_UNKNOWN;
		return -1;		
	}

	if (cfg.ftp_html && (docu->doc_url->type == FTP) && ext_is_html(docu->doc_url->p.ftp.path))
		docu->is_html = TRUE;

	if (access(inname , R_OK))
	{
	    if (!havelock)
	    {
		if (!(cfg.s_sock = 
			bufio_copen(inname , O_BINARY | O_CREAT | O_TRUNC | O_RDWR , S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR)))
		{
			xperror(inname);
			_free(inname);
			cfg.errcode = ERR_STORE_DOC;
			close_socket(sock , docu->doc_url , 0);
			return -1;
		}
		else
		{
#ifdef DEBUG
			if (cfg.debug)
				xprintf(1 , gettext("Locking %s\n") ,inname);
#endif
			if (_flock(bufio_getfd(cfg.s_sock) , b_lock))
			{
				cfg.errcode = ERR_LOCKED;
				_free(inname);
				close_socket(sock , docu->doc_url , 0);
				return -1;
			}
			havelock = TRUE;
			cfg.lock_fn = new_string(inname);
		}
	    }
	}
	else
	{
	    if (!havelock)
	    {
		if (!(cfg.s_sock = 
			bufio_copen(inname , O_BINARY | O_RDWR , S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR)))
		{
			xperror(inname);
			_free(inname);
			cfg.errcode = ERR_STORE_DOC;
			close_socket(sock , docu->doc_url , 0);
			return -1;
		}
		else
		{
#ifdef DEBUG
			if (cfg.debug)
				xprintf(1 , gettext("Locking %s\n") ,inname);
#endif
			if (_flock(bufio_getfd(cfg.s_sock) , b_lock))
			{
				cfg.errcode = ERR_LOCKED;
				_free(inname);
				close_socket(sock , docu->doc_url , 0);
				return -1;
			}
			havelock = TRUE;
			cfg.lock_fn = new_string(inname);
		}
	    }
	}
	_free(inname);

	if (docu->doc_url->type == FTP  && !(docu->doc_url->status & URL_REDIRECT) &&
	    cfg.errcode == ERR_FTP_NOREGET && !cfg.freget &&
	    (!cfg.ftp_proxy || (cfg.ftp_proxy && !cfg.ftp_via_http)))
	{
		close_socket(sock , docu->doc_url , 0);
		cfg.remove_lock = FALSE;
		return -1;
	}
	else if (docu->doc_url->type == FTP  && !(docu->doc_url->status & URL_REDIRECT) &&
		 cfg.errcode == ERR_FTP_NOREGET && cfg.freget &&
            (!cfg.ftp_proxy || (cfg.ftp_proxy && !cfg.ftp_via_http)))
	{
		docu->rest_pos = 0;
	}

	if ((docu->doc_url->type == HTTP || docu->doc_url->type == HTTPS ||
	    (docu->doc_url->type == FTP && cfg.ftp_proxy && cfg.ftp_via_http && !cfg.ftp_dirtyp) ||
	    (docu->doc_url->type == GOPHER && cfg.gopher_proxy && cfg.gopher_via_http)) 
		&& !(docu->doc_url->status & URL_REDIRECT))
	{
		if ((mres = read_mime_header(sock , &p , &totallen)) > 0)
		{
			docu->mime = p;

			if (cfg.htdig)
				printf("%s" , docu->mime);
#ifdef DEBUG
			if (cfg.debug)
			{
				xprintf(1 , gettext("*********** HTTP Server response MIME header **********\n"));
				xprintf(1 , "%s" , docu->mime);
				xprintf(1 , "*******************************************************\n");
			}
#endif
			
			docu->type_str = 
				get_mime_param_val_str("Content-type:" , docu->mime);

			if (docu->type_str && (!strncasecmp("text/html" , docu->type_str,9) ||
				!strncasecmp("text/css" , docu->type_str , 8)))
			{
				if (!strncasecmp("text/css" , docu->type_str , 8))
					docu->doc_url->status |= URL_STYLE;
				docu->is_html = TRUE;

				/* dirty hack to detect that FTP document downloaded */
				/* trought HTTP gateway is directory content */ 
				if (docu->doc_url->type == FTP && !docu->doc_url->p.ftp.dir)
				{
					resp = get_response(docu->mime);
					if (resp->ret_code == 200)
					{
						url_remove_from_hash_tab(docu->doc_url);
						docu->doc_url->p.ftp.dir = TRUE; 
						url_add_to_hash_tab(docu->doc_url);
					}
					_free(resp->text);
					_free(resp);
				}
			}
			else
			{
				docu->is_html = FALSE;
			}

			p3 = get_mime_param_val_str("Content-length:" , docu->mime);
			if (p3)
			{
				cfg.totsz = _atoi(p3);
				if (errno == ERANGE) cfg.totsz = 0;
			}
			_free(p3);

			if (docu->rest_pos)
			{
				p3 = get_mime_param_val_str("Content-Range:" , docu->mime);
				if (!p3)
				{
					if (cfg.freget)
					{
						xprintf(1 , gettext("Regeting whole file\n"));
						docu->rest_pos = 0;
					}
					else
					{
						cfg.errcode = ERR_HTTP_NOREGET;
						close_socket(sock , docu->doc_url);
						cfg.remove_lock = FALSE;
						_free(p3);
						return -1;
					}	
				}
				else
				{
					resp = get_response(docu->mime);
					if (resp->ret_code != 206)
					{
						docu->rest_pos = 0;
						xprintf(1 , gettext("Modified from last download - regeting whole\n"));
					}
					else
						cfg.totsz += docu->rest_pos;
					
					_free(resp->text);
					_free(resp);
					_free(p3);
				}
			}

			/**** get document creation time ****/
			p = get_mime_param_val_str("Last-Modified:" , docu->mime);
			if (p)
			{
				docu->dtime = new_tm(scntime(p));
				_free(p);
			}

			/**************/
			if (cfg.mode == MODE_SYNC)
			{
				resp = get_response(docu->mime);
				_free(resp->text);

				if (resp->ret_code == 304)
				{
					cfg.errcode = ERR_HTTP_ACTUAL;
					cfg.remove_lock = TRUE;
					close_socket(sock , docu->doc_url , 0);
					if (load)
					{
						p = url_to_filename(docu->doc_url);
						if (!(sock = bufio_open(p,O_BINARY | O_RDONLY)))
						{
							_free(p);
							_free(resp);
							cfg.errcode = ERR_FILE_OPEN;
							return -1;
						}
						_free(p);
						docu->doc_url->status |= URL_REDIRECT;
					}
					else
					{ 
						_free(resp);
						return 0;
					}
				}
				else if (modtime)
				{
					if (difftime(mktime(cfg.act_doc->dtime) , mktime(modtime)) > 0)
					{
						if (docu->rest_pos)
						{
							docu->rest_pos = 0;
							xprintf(1 , gettext("Modified from last download - regeting whole\n"));							}
						}
					else
					{
						cfg.errcode = ERR_HTTP_ACTUAL;
						cfg.remove_lock = TRUE;
						close_socket(sock , docu->doc_url , 0);
						if (load)
						{
							p = url_to_filename(docu->doc_url);
							if (!(sock = bufio_open(p,O_BINARY | O_RDONLY)))
							{
								_free(p);
								_free(resp);
								cfg.errcode = ERR_FILE_OPEN;	
								return -1;
							}
							_free(p);
						}
						else
						{
							_free(resp);
							return 0;
						}
					}
				}
				_free(resp);
			}
			else if (cfg.act_doc->dtime)
			{
				if (cfg.check_limits && cfg.condition.btime &&
				    (difftime(mktime(cfg.act_doc->dtime) , 
				   	     cfg.condition.btime) < 0))
				{
					cfg.errcode = ERR_OUTTIME;
					return -1;
				}
				if (cfg.check_limits && cfg.condition.etime &&
				    (difftime(mktime(cfg.act_doc->dtime) , 
				   	     cfg.condition.etime) > 0))
				{
					cfg.errcode = ERR_OUTTIME;
					return -1;
				}
			}

			p = NULL;
			totallen = 0;
		}
		else if (!mres)
		{
			if (p)
			{
				if (docu->save_online)
				{
					bufio_write(cfg.s_sock , p , totallen);
				}
				docu->is_html = TRUE;
			}
		} else
		{
			close_socket(sock , docu->doc_url , 0);
			cfg.remove_lock = FALSE;
			return -1;
		}
	}

	if (cfg.check_limits && cfg.condition.max_size && cfg.totsz && 
		cfg.condition.max_size < cfg.totsz &&
		docu->doc_url->type != FILET && !(docu->doc_url->status & URL_REDIRECT))
	{
		cfg.errcode = ERR_BIGGER;
		close_socket(sock , docu->doc_url , 0);
		return -1;
	}

	if (cfg.check_limits && cfg.condition.min_size && cfg.totsz && 
		cfg.condition.min_size < cfg.totsz &&
		docu->doc_url->type != FILET && !(docu->doc_url->status & URL_REDIRECT))
	{
		cfg.errcode = ERR_SMALLER;
		close_socket(sock , docu->doc_url , 0);
		return -1;
	}

        if (cfg.check_limits && docu->type_str && cfg.condition.mime && *cfg.condition.mime)
        {
		fnd = is_in_list(docu->type_str , cfg.condition.mime);

		if ((cfg.condition.allow_mime && !fnd) || 
		    (!cfg.condition.allow_mime && fnd))
		{
			cfg.errcode = ERR_NOMIMET;
			close_socket(sock , docu->doc_url , 0);
			return -1;
		}
        }

	if (cfg.check_limits && cfg.condition.uexit)
	{
		if (!uexit_condition(cfg.act_doc->doc_url ,
			&cfg.totsz ,
			cfg.act_doc->dtime))
		{
			cfg.errcode = ERR_SCRIPT_DISABLED;
			close_socket(sock , docu->doc_url , 0);
			return -1;
		}
	}



	if (docu->save_online)
	{
		ftruncate(bufio_getfd(cfg.s_sock) , docu->rest_pos);
	 	lseek(bufio_getfd(cfg.s_sock) , docu->rest_pos , SEEK_SET);
		bufio_reset(cfg.s_sock);
	}

	if (cfg.report_size)
		iface_set_what(gettext("Transfering data"));

	show_progres(docu->rest_pos);

	while ((len = bufio_nbfread(sock, buf, sizeof(buf))) > 0 || errno == EWOULDBLOCK)
	{
		if (cfg.act_doc->save_online)
		{
			if (write(bufio_getfd(cfg.s_sock) , buf ,len) != len)
			{
				cfg.errcode = ERR_STORE_DOC;
				xperror("storing doc");
				break;
			}
		}

		totallen += len;

		show_progres(totallen + docu->rest_pos);

		if (load || docu->is_html ||
		    (docu->doc_url->type == FTP && docu->doc_url->p.ftp.dir) ||
		    (docu->doc_url->type == GOPHER && 
		     docu->doc_url->p.gopher.type == '1'))
		{
			p = _realloc(p , totallen + 1);
			memmove(p + totallen - len , buf , len);
		}
	}

	if (cfg.report_size)
		iface_set_what(gettext("Data transfer done"));

	if (cfg.progres && cfg.report_size
#ifdef I_FACE
		&& !cfg.xi_face
#endif
	) printf("\n");

	if (len < 0) 
	{
		cfg.t_error = errno;
		xperror("read");
	}
		

	close_socket(sock,docu->doc_url , 1);

	if (docu->doc_url->type == FTP && cfg.errcode == ERR_FTP_TRUNC)
	{
		cfg.remove_lock = FALSE;
		retcode = -1;
	}

	/*** ak nezaciname od zaciatku tak precitaj znovu zo suboru ***/
	if (p) 
	{
		if (docu->rest_pos)
		{
			_free(p);
			totallen = 0;
			lseek(bufio_getfd(cfg.s_sock) , 0 , SEEK_SET);
			bufio_reset(cfg.s_sock);

			while ((len = bufio_read(cfg.s_sock, buf, sizeof(buf))) > 0 || errno == EWOULDBLOCK)
			{
				totallen += len;
				p = _realloc(p , totallen + 1);
				memmove(p + totallen - len , buf , len);
			}	
		}
		if (p) *(p + totallen) = '\0';
	}

	docu->size = totallen;
	docu->contents = p;

	if (docu->doc_url->type == GOPHER && !(docu->doc_url->status & URL_REDIRECT) &&
	    !(cfg.gopher_proxy && cfg.gopher_via_http))
	{
		docu->is_html = FALSE;
		/**** gopher adresar prerobime na HTML dokument ****/
		if (docu->doc_url->p.gopher.type == '1')
		{
			if (!(docu->doc_url->status & URL_REDIRECT))
				gopher_dir_to_html(docu);
			docu->is_html = TRUE;
		}
	}
	else if (docu->doc_url->type == FTP && !(docu->doc_url->status & URL_REDIRECT) &&
		 !(cfg.ftp_proxy && cfg.ftp_via_http && !cfg.ftp_dirtyp))
	{
		docu->is_html = ext_is_html(docu->doc_url->p.ftp.path);

		/**** ftp adresar prerobime na HTML dokument ****/
		if (docu->doc_url->p.ftp.dir)
		{
			if (!(docu->doc_url->status & URL_REDIRECT))
			{
				if (cfg.ftp_list)
					ftp_dir_list_to_html(docu);
				else
					ftp_dir_to_html(docu);
			}
			docu->is_html = TRUE;
		}
	}
	else if ((docu->doc_url->type == HTTP || docu->doc_url->type == HTTPS ||
		 (docu->doc_url->type == FTP && cfg.ftp_proxy && cfg.ftp_via_http && !cfg.ftp_dirtyp) ||
		 (docu->doc_url->type ==GOPHER && cfg.gopher_proxy && cfg.gopher_via_http)) && 
		!(docu->doc_url->status & URL_REDIRECT))
	{
		/**** zistime navratovy kod a ci sa jedna o HTTP/1.0 ****/
		resp = get_response(docu->mime);
		if (resp)
		{
			_free(resp->text);

			/**** chybny dokument ****/
			if (resp->ret_code >= 400)
			{
				cfg.errcode = 2000 + resp->ret_code;
				_free(resp);
				return -1;
			}
	
			/**** presmerovanie dokumentu na ine URL ****/
			if (resp->ret_code == 302 || resp->ret_code == 301)
			{
				url *pomurl;
				char *pomcr = NULL;

				free(resp);
				pomcr = get_mime_param_val_str("URI:" , docu->mime);

				if (pomcr)
				{
					strtokc(pomcr , ';');
				}
				else
				{
					pomcr = get_mime_param_val_str("Location:" , 
						docu->mime);

					if (!pomcr)
					{
						dllist *stc;
#ifdef I_FACE					
						bool sxi = cfg.xi_face;
						cfg.xi_face = FALSE;
#endif
			
						stc = html_get_all_links(docu , FALSE);
						if (stc)
						{
							pomcr = url_to_urlstr((url *)stc->data);
							while(stc)
							{
								free_deep_url((url *)stc->data);
								stc = dllist_remove_entry(stc , stc);
							}
						}	
#ifdef I_FACE
						cfg.xi_face = sxi;	
#endif
					}
				}
			
				if (pomcr)
				{

					pomurl = parse_url(pomcr);
					free(pomcr);

					cfg.errcode = ERR_HTTP_REDIR;

					url_redirect_to(docu->doc_url , pomurl);
				}
				else 
					cfg.errcode = ERR_HTTP_UNKNOWN;
				return -1;
			}

			/*** not modified ***/
			else if (resp->ret_code == 304 && cfg.mode == MODE_SYNC)
			{
				free(resp);
				cfg.errcode = ERR_HTTP_ACTUAL;
				return -1;
			}
			free(resp);
		}

		
		/**** kontrola ci bol dokument preneseny cely ****/
		if (cfg.check_size)
		{
			p = get_mime_param_val_str("Content-Length:" , docu->mime);
			if (p)
			{
				sz = _atoi(p);	
				if (errno != ERANGE && sz != docu->size - (docu->contents ? docu->rest_pos : 0))
				{
					cfg.errcode = ERR_HTTP_TRUNC;

					cfg.remove_lock = FALSE;
					retcode = -1;
					xprintf(1 , gettext("File may be truncated\n"));
				}
				_free(p);
			}
		}

		
		/**** zistenie ci bol dokument komprimovany a dekomprimacia ****/
		p = get_mime_param_val_str("Content-Encoding:" , docu->mime);
		if (!retcode && p && cfg.use_enc && 
			strcmp(docu->type_str , "application/x-gzip") &&
			strcmp(docu->type_str , "application/gzip") &&
			strcmp(docu->type_str , "application/x-compress") &&
			strcmp(docu->type_str , "application/compress"))
		{
			if ((!strcasecmp(p,"x-gzip")) ||
			    (!strcasecmp(p,"gzip")) ||
			    (!strcasecmp(p,"x-compress")) ||
			    (!strcasecmp(p,"compress")))
			{
				if (!gzip_decode(docu->contents , docu->size , 
					&p1 , &len , 
					(docu->contents ? NULL : cfg.lock_fn)))
				{
					docu->size = len;
					free(docu->contents);
					docu->contents = p1;
					xprintf(1 , gettext("OK Encoded\n"));
				}
				else
				{
					xperror("Encoding\n");
				}
			}
			else
			{
				xprintf(1 , gettext("Unknown Encoding\n"));
			}
		}
		else if (p && !retcode)
		{
			xprintf(1 , gettext("Received Encoded file but encoding not allowed (untouched)\n"));	
		}
		_free(p);
	}
	else
	{
		if (docu->doc_url->type == FILET || (docu->doc_url->status & URL_REDIRECT))
		{
			p1 = url_to_filename(docu->doc_url);

			if (file_is_html(p1))
			{
				docu->is_html = TRUE;
			}
			free(p1);
		}
		else
		{
			docu->is_html = FALSE;
		}
	}

	if (cfg.t_error)
	{
		if (!cfg.errcode) cfg.errcode = ERR_READ;
		cfg.remove_lock = FALSE;
		retcode = -1;
	}

	if (docu->size == 0)
	{
		if (!cfg.errcode) cfg.errcode = ERR_ZERO_SIZE;
		cfg.remove_lock = FALSE;
		retcode = -1;
	}
#ifdef I_FACE
	if (cfg.xi_face)
		set_doc_info(docu);
#endif
	if (!retcode && cfg.lock_fn && docu->save_online && 
		!docu->contents && (cfg.mode != MODE_NOSTORE) &&
		!(docu->doc_url->status & URL_REDIRECT) && (docu->doc_url->type != FILET))
	{
		p1 = url_to_filename(docu->doc_url);
		if (!access(p1 , F_OK))
		{
			if (unlink(p1)) xperror(p1);
		}
		if (link(cfg.lock_fn , p1))
		{
			if (errno != EPERM)
				xperror(p1);
			else
			{
				if (copy_fd_to_file(bufio_getfd(cfg.s_sock) , p1))
					xperror(p1);
			}
		}

		if (docu->dtime && cfg.preserve_time)
		{
			struct utimbuf utmbf;

			stat(p1 , &estat);
			utmbf.actime = estat.st_atime;
			utmbf.modtime = mktime(docu->dtime) + cfg.gmdiff;
			utime(p1 , &utmbf);
		}

		if (docu->doc_url->type == FTP &&
		    docu->doc_url->extension &&
		    cfg.preserve_perm)
		{
			chmod(p1 , ((ftp_url_extension *)docu->doc_url->extension)->perm);
		}

		free(p1);
	}

	return retcode;
}

/********************************************************/
/* ulozi dokument ak je to potrebne vytvori adresare	*/
/********************************************************/
int store_document(docu , overwrite)
doc *docu;
int overwrite;
{
	char *pom;
	int f;
	struct utimbuf utmbf;
	struct stat estat;

	if (cfg.mode == MODE_NOSTORE) return 0;

	/*** don't store directory indexes ***/
	if (!cfg.store_index && url_is_dir_index(docu->doc_url))
			return 0;

	pom = url_to_filename(docu->doc_url);
	makealldirs(pom);

	if (!access(pom , R_OK) && !overwrite)
	{
		free(pom);
		return 0;
	}
	
	if ((f = open(pom , O_BINARY | O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR)) == -1)
	{
		if (!access(pom , R_OK)) unlink(pom);
		xperror(pom);
		free(pom);
		return -1;
	}

	if (write(f , docu->contents , docu->size) != docu->size)
	{
		if (!access(pom , R_OK)) unlink(pom);
		xperror(pom);
		close(f);
		free(pom);
		return -1;
	}

	close(f);

	if (docu->dtime && cfg.preserve_time)
	{
		utmbf.modtime = mktime(docu->dtime) + cfg.gmdiff;
	}
	else
	{
		utmbf.modtime = docu->stime;
	}

	stat(pom , &estat);
	utmbf.actime = estat.st_atime;
	utime(pom , &utmbf);

	if (docu->doc_url->type == FTP &&
	    docu->doc_url->extension &&
	    cfg.preserve_perm)
	{
		chmod(pom , ((ftp_url_extension *)docu->doc_url->extension)->perm);
	}

	free(pom);

	return 0;
}

/*** remove improper documents if required ***/
int remove_document(urlr)
url *urlr;
{
	char *fn;
	int len;

#ifdef DEBUG
	if (cfg.debug)
	{
		fn = url_to_urlstr(urlr);
		xprintf(1 , gettext("Removing improper document : %s\n") , fn); 
		_free(fn);
	}
#endif
	
	fn = url_to_filename(urlr);

	if (urlr->type == FTP)
	{
		len = strlen(fn);

		/*** if URL FTPdir index ***/
		if (!strcmp(fn + len - (( len >= strlen(cfg.index_name)) ? strlen(cfg.index_name) : 0) , cfg.index_name))
			 *(fn + len - strlen(cfg.index_name)) = '\0';

		return unlink_recursive(fn);	
	}
	else
	{
		if (unlink(fn))
		{
			xperror(fn);
			return -1;
		}
		
	}

	return 0;
}

#ifdef I_FACE
/********************************************************/
/* nastavenie info dokumentu pre informaciu pouzivatela	*/
/********************************************************/

static void set_doc_info(docp)
doc *docp;
{
#ifdef WITH_TREE
	url_prop *prp = _malloc(sizeof(url_prop));

	prp->size = docp->size;
	prp->type = NULL;
	switch (docp->doc_url->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			if (docp->type_str)
				prp->type = new_string(docp->type_str);
			break;
		case FILET:
			prp->type = new_string(gettext_nop("Local file"));
			break;
		case GOPHER:
			switch (docp->doc_url->p.gopher.type)
			{
				case '0':
					prp->type = new_string(gettext_nop("Gopher/Text File"));
					break;
				case '1':
					prp->type = new_string(gettext_nop("Gopher/Directory"));
					break;
				case '2':
					prp->type = new_string(gettext_nop("Gopher/CSO phone book"));
					break;
				case '3':
					prp->type = new_string(gettext_nop("Gopher/Error"));
					break;
				case '4':
					prp->type = new_string(gettext_nop("Gopher/BINHEX"));
					break;
				case '5':
					prp->type = new_string(gettext_nop("Gopher/DOS bin"));
					break;
				case '6':
					prp->type = new_string(gettext_nop("Gopher/UUencoded"));
					break;
				case '7':
					prp->type = new_string(gettext_nop("Gopher/Search index"));
					break;
				case '8':
					prp->type = new_string(gettext_nop("Gopher/Telnet session"));
					break;
				case '9':
					prp->type = new_string(gettext_nop("Gopher/bin"));
					break;
				case '+':
					prp->type = new_string(gettext_nop("Gopher/Duplicated server"));
					break;
				case 'T':
					prp->type = new_string(gettext_nop("Gopher/TN3270"));
					break;
				case 'g':
					prp->type = new_string(gettext_nop("Gopher/GIF"));
					break;
				case 'I':
					prp->type = new_string(gettext_nop("Gopher/Image"));
					break;
			}
			break;
		case FTP:
			if (docp->doc_url->p.ftp.dir)
				prp->type = new_string(gettext_nop("FTP/Directory"));
			else
				prp->type = new_string(gettext_nop("FTP/File"));
			break;
		default:
			prp->type = new_string(gettext_nop("Unsupported type"));
			break;
	}

	if (!prp->type) prp->type = new_string(gettext_nop("Local file"));

	docp->doc_url->prop = prp;
#endif
}
#endif
