/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <netdb.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>
#include <time.h>

#include "config.h"
#include "url.h"
#include "http.h"
#include "net.h"
#include "base64.h"
#include "tools.h"
#include "cookie.h"
#include "html.h"
#include "mime.h"

#ifdef USE_SSL

#include <rsa.h>
#include <crypto.h>
#include <x509.h>
#include <pem.h>
#include <ssl.h>
#include <err.h>
#include <md5.h>

#define _MD5_CTX 	MD5_CTX
#define _MD5Init 	MD5_Init
#define _MD5Update	MD5_Update
#define _MD5Final	MD5_Final

extern int ssl_do_connect();

#else

#include "md5c.h"

#endif

/* this function is stolen from apache md5_util.c by Jeff Hostetler */
char *_md5(data)
unsigned char *data;
{
        _MD5_CTX md5ctx;
	unsigned char md5s[16];
        unsigned char result[33];
	unsigned char pom[3];
        int i;

        _MD5Init(&md5ctx);
        _MD5Update(&md5ctx , data , strlen(data));
        _MD5Final(md5s , &md5ctx);

	result[0] = '\0';
        for (i = 0 ; i < 16 ; i++)
	{
                sprintf(pom , "%02x" , md5s[i]);
		strcat(result , pom);
	}

	return new_string(result);
}

void http_digest_deep_free(digest)
http_digest_info *digest;
{
	_free(digest->nonce);
	_free(digest->opaque);
	_free(digest->realm);
	_free(digest->site);
	_free(digest);
}

http_digest_info *http_digest_parse(mime_header , urlp , proxy_auth)
char *mime_header;
url *urlp;
int *proxy_auth;
{
	char *authtag = NULL;
	http_digest_info *retv = NULL;

	*proxy_auth = FALSE;

	if ((authtag = get_mime_param_val_str("Proxy-Authenticate:" , mime_header)))
		*proxy_auth = TRUE;	
	else authtag = get_mime_param_val_str("WWW-Authenticate:" , mime_header);

	if (authtag)
	{
		if (!strncmp(authtag , "Digest " , 7))
		{
			retv = _malloc(sizeof(http_digest_info));
			retv->nonce = get_url_from_tag(authtag , "nonce");
			retv->opaque = get_url_from_tag(authtag , "opague");
			retv->realm = get_url_from_tag(authtag , "realm");
			if (*proxy_auth)
			{
				retv->site = NULL;
				retv->port = 0;
			}
			else
			{
				retv->site = new_string(url_get_site(urlp));
				retv->port = url_get_port(urlp);
			}
		}
		_free(authtag);
	}

	return retv;
}

int http_dumy_proxy_connect(sock , host , port)
bufio *sock;
char *host;
int port;
{
	http_response *mresp=NULL;
	char *p;
	int len,res;
	char pom[1024];

	sprintf(pom , "CONNECT %s:%d HTTP/1.0\r\n\r\n" , 
		host , port);
#ifdef DEBUG
	if (cfg.debug)
	{
		xprintf(1 , gettext("****************** Proxy connect request *****************\n"));
		xprintf(1 , "%s" , pom);
		xprintf(1 , "**********************************************************\n");
	}
#endif
	if (bufio_write(sock , pom , strlen(pom)) != strlen(pom))
	{
		xperror("proxy req");
		cfg.errcode = ERR_PROXY_CONNECT;
		return -1;
	}
	p = NULL;
	if ((res = read_mime_header(sock , &p , &len)) > 0)
	{
#ifdef DEBUG
		if (cfg.debug)
		{
			xprintf(1 , gettext("***************** Proxy connect response *****************\n"));
			xprintf(1 , "%s" , p);
			xprintf(1 , "**********************************************************\n");
		}
#endif
		mresp = get_response(p);
		_free(mresp->text);
		if (mresp->ret_code < 400)
		{
			_free(mresp);
		}
		else
		{
			_free(mresp);	
			_free(p);
			cfg.errcode = ERR_PROXY_CONNECT;
			return -1;
		}
	}
	else 
	{
		_free(mresp);
		_free(p);
		cfg.errcode = ERR_PROXY_CONNECT;
		return -1;
	}
	_free(p);

	return 0;
}

/************************************************/
/* na dany soket posle HTTP request		*/
/************************************************/
int http_get_request(sock , urlp , pos , modtime)
bufio *sock;
url *urlp;
int pos;
struct tm *modtime;
{
	char pom[1024];
	char *p,*p2;
	char **al;
	char buf[32768];
	char auth[2048];
	char *user;
	char *pass;
	int auth_scheme;


	if ((!cfg.http_proxy && urlp->type == HTTP) || urlp->type == HTTPS)
	{
		p = encode_url(urlp->p.http.document);
		if (urlp->p.http.searchstr)
		{
			strcpy(buf , p);
			_free(p);
			p = encode_searchstr(urlp->p.http.searchstr);
			strcat(buf , "?");
			strcat(buf , p);
			_free(p);
			p = new_string(buf);
		}
	}
	else
	{
		p = url_to_urlstr_woauth(urlp);
	}

	sprintf(buf , "GET %s HTTP/1.0\r\n"
		"Accept: */*\r\n"
		"Host: %s\r\n" , p , url_get_site(urlp));

	_free(p);

	if (cfg.identity)
	{
		sprintf(pom , "User-Agent: %s\r\n" , cfg.identity);
	}
	else
	{
		sprintf(pom , "User-Agent: pavuk-%s\r\n" , VERSION);
	}
	strcat(buf , pom);

	if (cfg.send_from && cfg.from)
	{
		sprintf(pom , "From: %s\r\n" , cfg.from);
		strcat(buf , pom);
	}

	if (cfg.send_cookies && (p = cookie_get_field(urlp)))
	{
		strcat(buf , p);
		_free(p);
	}

	auth[0] = '\0';
	user = url_get_user(urlp , cfg.auth_digest ? cfg.auth_digest->realm : NULL);
	pass = url_get_pass(urlp , cfg.auth_digest ? cfg.auth_digest->realm : NULL);
	auth_scheme = url_get_auth_scheme(urlp , cfg.auth_digest ? cfg.auth_digest->realm : NULL);

	/*** informacie pre autorizaciu ****/
	if (user)
	{
		if (pass)
		{
			if (cfg.auth_digest)
			{
				char *a1,*a2;
				char *d = encode_url(urlp->p.http.document);

				sprintf(pom , "%s:%s:%s" , user , cfg.auth_digest->realm , pass);
				a1 = _md5(pom);
				sprintf(pom , "GET:%s" , d);
				a2 = _md5(pom);
				sprintf(pom , "%s:%s:%s" , a1 , cfg.auth_digest->nonce , a2);
				p = _md5(pom);

				sprintf(auth , "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", response=\"%s\"" , 
					user , cfg.auth_digest->realm , cfg.auth_digest->nonce , d , p);

				if (cfg.auth_digest->opaque)
				{
					strcat(auth , ", opaque=\"");
					strcat(auth , cfg.auth_digest->opaque);
					strcat(auth , "\"");
				}
				_free(d);
			}
			else if (auth_scheme == 1)
			{
				sprintf(auth , "user %s:%s" , 
					user , pass);
			}
			else if (auth_scheme == 2)
			{
				sprintf(pom , "%s:%s" , 
					user , pass);
				p2 = base64_encode(pom);
				sprintf(auth , "Basic %s", p2);
				free(p2);
			}
			if (auth[0])
			{
				sprintf(pom , "Authorization: %s\r\n" , auth);
				strcat(buf , pom);
			}
		}
		else
		{
			sprintf(pom , "Authorization: user %s\r\n" , user);
			strcat(buf , pom);
		}
	}

        /**** informatins for HTTP proxy autorization ****/
	auth[0] = '\0';
        if (cfg.http_proxy_user)
        {
                if (cfg.http_proxy_pass)
                {
			if (cfg.auth_proxy_digest)
			{
				char *a1,*a2;
				char *d = url_to_urlstr_woauth(urlp);

				sprintf(pom , "%s:%s:%s" , user , cfg.auth_proxy_digest->realm , pass);
				a1 = _md5(pom);
				sprintf(pom , "GET:%s" , d);
				a2 = _md5(pom);
				sprintf(pom , "%s:%s:%s" , a1 , cfg.auth_proxy_digest->nonce , a2);
				p = _md5(pom);

				sprintf(auth , "Digest username=\"%s\", realm=\"%s\", nonce=\"%s\", uri=\"%s\", response=\"%s\"" , 
					user , cfg.auth_proxy_digest->realm , cfg.auth_proxy_digest->nonce , d , p);

				if (cfg.auth_proxy_digest->opaque)
				{
					strcat(auth , ", opaque=\"");
					strcat(auth , cfg.auth_proxy_digest->opaque);
					strcat(auth , "\"");
				}
				_free(d);
			}
                        if (cfg.proxy_auth_scheme == 1)
                        {
                                sprintf(auth , "user %s:%s" ,
                                        cfg.http_proxy_user , cfg.http_proxy_pass);
                        }
                        else if (cfg.proxy_auth_scheme == 2)
                        {
                                sprintf(pom , "%s:%s" ,
                                        cfg.http_proxy_user , cfg.http_proxy_pass);
                                p2 = base64_encode(pom);
                                sprintf(auth , "Basic %s", p2);
                                free(p2);
                        }
			if (auth[0])
			{
                        	sprintf(pom , "Proxy-Authorization: %s\r\n" , auth);
                        	strcat(buf , pom);
			}
                }
                else
                {
                        sprintf(pom , "Proxy-Authorization: user %s\r\n" , cfg.http_proxy_user);
			strcat(buf , pom);
                }
        }

	/**** preferovane jazyky ****/
	if (cfg.accept_lang)
	{
		bool f=FALSE;
		al = cfg.accept_lang;
		
		if (*al)
		{
			sprintf(pom , "Accept-Language: %s" , *al);
			al++;
			f = TRUE;
		}
		while(*al)
		{
			strcat(pom , ",");
			strcat(pom , *al);
			al++;
		}
		if (f)
		{
			strcat(pom , "\r\n");
			strcat(buf , pom);
		}
	}

	/*** preffered character sets ***/
	if (cfg.accept_chars)
	{
		bool f=FALSE;
		al = cfg.accept_chars;
		
		if (*al)
		{
			sprintf(pom , "Accept-Charset: %s" , *al);
			al++;
			f = TRUE;
		}
		while(*al)
		{
			strcat(pom , ",");
			strcat(pom , *al);
			al++;
		}
		if (f)
		{
			strcat(pom , "\r\n");
			strcat(buf , pom);
		}
	}
	/**** referencne URL ****/
	if (urlp->parent_url && urlp->parent_url[0])
	{
		if (urlp->parent_url[0]->type != FILET)
		{
			strcat(buf , "Referer: ");
			p = url_to_urlstr_woauth(urlp->parent_url[0]);
			strcat(buf,p);
			free(p);
			strcat(buf, "\r\n");
		}
	}

	/**** povolenie kodovanie gzipom, compressom ****/
	if (cfg.use_enc)
	{
		strcat (buf , "Accept-Encoding: x-gzip, gzip, x-compress, compress\r\n");
	}

	/**** reget (iba niektore servery podporuju) ****/
	if (pos)
	{
		sprintf(pom , "Range: bytes=%d-\r\n" , pos); 
		strcat(buf , pom);
		if (cfg.act_doc->etag)
		{
			sprintf(pom , "If-Range: %s\r\n" , cfg.act_doc->etag);
			strcat(buf , pom);
		}
	}

	/**** synchronizacia ****/
	if (modtime && !pos)
	{
		strftime(pom , sizeof(pom) ,
			"If-Modified-Since: %a, %d %b %Y %H:%M:%S GMT" , modtime);
		strcat(buf , pom);
		strcat(buf , "\r\n");
	}

	/**** no caching ****/
	if (!cfg.cache)
	{
		strcat(buf , "Pragma: no-cache\r\n");
	}

	strcat(buf, "\r\n");

#ifdef DEBUG
	if (cfg.debug)
	{
		xprintf(1 , gettext("************ Client HTTP MIME header ***************\n"));
		xprintf(1 , "%s" , buf);
		xprintf(1 , "****************************************************\n");
	}
#endif

	/**** zaslanie requestu ****/
	if (bufio_write(sock , buf , strlen(buf)) == -1)
	{
		return -1;
	}

	return 0;
}

bufio *get_http_data_socket(urld , pos , modtime)
url *urld;
int pos;
struct tm *modtime;
{
	bufio *sock;
	char *host=NULL;
	int port=0;

	cfg.errcode = ERR_NOERROR;

	switch (urld->type)
	{
	    case HTTP:
		if (!cfg.http_proxy)
		{
			host = urld->p.http.host;
			port = urld->p.http.port;
		}
		else
		{
			host = cfg.http_proxy;
			port = cfg.http_proxy_port;
		}
		break;
	    case HTTPS:
		if (!cfg.ssl_proxy)
		{
                	host = urld->p.http.host;
               	 	port = urld->p.http.port;
		}
		else
		{
			host = cfg.ssl_proxy;
			port = cfg.ssl_proxy_port;
		}
		break;
	    case FTP:
		if (cfg.ftp_proxy)
		{
			host = cfg.ftp_proxy;
			port = cfg.ftp_proxy_port;
		}
		else
			xprintf(1 , gettext("HUH: don't know how to gateway FTP when proxy not specified\n"));
		break;
	    case GOPHER:
		if (cfg.gopher_proxy)
		{
			host = cfg.gopher_proxy;
			port = cfg.gopher_proxy_port;
		}
		else
			xprintf(1 , gettext("HUH: don't know how to Gateway Gopher when proxy not specified\n"));
		break;
	     default:
		xprintf(1 , gettext("HUH: this protocol is not supported for gatewaing\n"));
		
	}

	sock = bufio_fdopen(net_connect(host , port));

#ifdef USE_SSL	
	if (sock && urld->type == HTTPS)
	{
		if (!ssl_do_connect(urld , sock))
		{
			cfg.errcode = ERR_HTTPS_CONNECT;
			shutdown(bufio_getfd(sock) , 2);
			bufio_close(sock);
			return NULL;
		}
	}
#endif

	if (!sock)
	{
		if (h_errno != 0)
			xherror(host);
		else
			xperror("net_connect");

		cfg.errcode = ERR_HTTP_CONNECT;
		return NULL;
	}	
	else if (http_get_request(sock , urld , pos , modtime) < 0)
	{
		xperror("http_simple_get_request");
		cfg.errcode = ERR_HTTP_SNDREQ;

#ifdef USE_SSL	
		if(urld->type == HTTPS)
		{
#ifdef DEBUG
			ERR_print_errors_fp(stdout);
#endif
			SSL_free (cfg.ssl_con);
			SSL_CTX_free (cfg.ssl_ctx);
			cfg.ssl_con = NULL;
			cfg.ssl_ctx = NULL;
 			shutdown(bufio_getfd(sock) , 2);

		}	
#endif
		bufio_close(sock);
		return NULL;
	}
	return sock;
}

/**********************************************/
/* parsovanie navratoveho kodu z HTTP servera */
/**********************************************/

http_response * get_response(doc_txt)
char *doc_txt;
{
	char *ptr,*p;
	char pom[1024];

	if (!doc_txt) return NULL;

	ptr = strchr(doc_txt , '\n');

	if (ptr)
	{
		strncpy(pom , doc_txt , ptr - doc_txt);

		*(pom + (ptr - doc_txt)) = '\0';

		if (!strncmp(pom,"HTTP/",5))
		{
			http_response *ret_val = (http_response *)
				_malloc(sizeof(http_response));

			p = pom + 5;
			memcpy(ret_val->version , p , 3);

			p = pom + 9;
			ret_val->ret_code = _atoi(p);

			p = pom + 12;
			while(isspace(*p)) p++;

			ret_val->text = new_string(p);

			return ret_val;
		}
	}
	return NULL;
}

int read_mime_header(sock , buf , len)
bufio *sock;
char **buf;
int *len;
{
	char pom[4096];
	char *rb;
	int alen,tlen=0;
	rb = NULL;

	*buf = NULL;

	if ((alen = bufio_readln(sock , pom , sizeof(pom)))>0)
	{
		if (!strncmp(pom , "HTTP/" , 5))
		{
			rb = _malloc(alen + 1);
			tlen += alen;
			memmove(rb , pom , alen + 1);
			while((alen = bufio_readln(sock , pom , sizeof(pom))) > 0)
			{
				tlen += alen;
				rb = _realloc(rb , tlen + 1);
				memmove(rb + tlen - alen , pom , alen + 1);
				if (!strtok(pom , "\n\r")) break;
			}
			if (alen < 0)
			{
				cfg.errcode = ERR_READ;
				xperror("http_response");
				_free(rb);
				*len = 0;
				*buf = NULL;
				return -1;
			}
			else
			{
				*len = tlen;
				*buf = rb;
				return 1;
			}
		}
		else
		{
			*len = tlen;
			*buf = rb;
			return 0;
		}
	}
	else
	{
		*buf = NULL;
		*len = 0;
		if (alen < 0)
		{
			cfg.errcode = ERR_READ;
			xperror("http_response");
			return -1;
		}
		return 0;
	}
}

