/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "config.h"
#include "url.h"
#include "doc.h"

static char *errcodetype(ecode)
{
	switch (ecode)
	{
		case ERR_NOERROR:
			return "OK";
		case ERR_STORE_DOC:
		case ERR_FILE_OPEN:
		case ERR_DIR_URL:
		case ERR_UNKNOWN:
		case ERR_PROXY_CONNECT:
		case ERR_FTP_UNKNOWN:
		case ERR_FTP_BUSER:
		case ERR_FTP_BPASS:
		case ERR_HTTP_UNKNOWN:
		case ERR_HTTP_AUTH:
		case ERR_HTTP_PAY:
		case ERR_HTTP_BADRQ:
		case ERR_HTTP_FORB:
		case ERR_HTTP_SERV:
		case ERR_GOPHER_UNKNOWN:
			return "FATAL";
		case ERR_LOCKED:
		case ERR_BIGGER:
		case ERR_NOMIMET:
		case ERR_BREAK:
		case ERR_OUTTIME:
		case ERR_SCRIPT_DISABLED:
		case ERR_SMALLER:
		case ERR_ZERO_SIZE:
		case ERR_PROCESSED:
		case ERR_UDISABLED:
		case ERR_RDISABLED:
		case ERR_FTP_NOREGET:
		case ERR_FTP_ACTUAL:
		case ERR_FTP_NOTRANSFER:
		case ERR_FTP_NOMDTM:
		case ERR_FTP_DIRNO:
		case ERR_HTTP_NOREGET:
		case ERR_HTTP_REDIR:
		case ERR_HTTP_ACTUAL:
			return "WARN";
		case ERR_READ:
		case ERR_FTP_BDIR:
		case ERR_FTP_CONNECT:
		case ERR_FTP_DATACON:
		case ERR_FTP_GET:
		case ERR_FTP_NODIR:
		case ERR_FTP_TRUNC:
		case ERR_HTTP_CONNECT:
		case ERR_HTTP_SNDREQ:
		case ERR_HTTP_TRUNC:
		case ERR_HTTP_CYCLIC:
		case ERR_HTTP_NFOUND:
		case ERR_GOPHER_CONNECT:
		case ERR_HTTPS_CONNECT:
			return "ERR";
		default:
			return "UNKNOWN";
	}
}

void short_log(docp , urlp)
doc *docp;
url *urlp;
{
	int fd;
	char pom[2048];
	char *p;
	time_t t = time(NULL);;

	if (!cfg.short_logfile)
		return;

	fd = open(cfg.short_logfile , O_CREAT | O_APPEND | O_WRONLY , S_IRUSR | S_IWUSR);

	if (fd < 0)
	{
		xperror("shortlog");
		return;
	}

	if (_flock(fd , TRUE))
	{
		close(fd);
		return;
	}

	sprintf(pom , "%d %s %d/%d %s %d " , getpid() , 
		strtok(ctime(&t) , "\n"), 
		cfg.actual_url + 1, cfg.total_cnt ,
		errcodetype(cfg.errcode) , cfg.errcode);

	write(fd , pom , strlen(pom));

	p = url_to_urlstr(urlp);

	write(fd , p , strlen(p));

	_free(p);

	if (urlp->parent_url && urlp->parent_url[0])
	{
		write(fd , " " , 1);
		p = url_to_urlstr(urlp->parent_url[0]);
		write(fd , p , strlen(p));
		_free(p);
	}
	else
	{
		write(fd , " [none]" , 7);
	}

	p = url_to_filename(urlp);

	write(fd , " " , 1);
	write(fd , p , strlen(p));

	if (docp->mime)
	{
		int l = strcspn(docp->mime , "\r\n");

		write(fd , " " , 1);
		write(fd , docp->mime , l);
	}

	write(fd , "\n" , 1);

	_funlock(fd);
	close(fd);
}

