/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>

#include "config.h"
#include "dns.h"

#ifdef X_FACE
static void TOut(client_data , timer)
XtPointer client_data;
XtIntervalId *timer;
{
	cfg.X.inid = 0;

	_X_EscLoop();
}

static void ConnectCB(ptr , src , id)
XtPointer ptr;
int *src;
XtInputId *id;
{
	XtRemoveInput(*id);
	cfg.X.iid = 0;
#ifdef DEBUG
	if (cfg.debug)
		xprintf(1 , gettext("Async connect - connected\n"));
#endif
	_X_EscLoop();
}
#endif


#ifdef GTK_FACE
static void TOut(data)
gpointer data;
{
	cfg.gtk.inid = 0;

	_X_EscLoop();
}

static void ConnectCB(data , source , condition)
gpointer data;
gint source;
GdkInputCondition condition;
{
	gdk_input_remove(cfg.gtk.iid);
	cfg.gtk.iid = 0;

#ifdef DEBUG
	if (cfg.debug)
		xprintf(1 , gettext("Async connect - connected\n"));
#endif
	_X_EscLoop();
}
#endif

/*****************************************************************/
/* otvori spojenie na dany port servera  s (IP , DNS) adresou	 */
/*****************************************************************/
int net_connect(hostname , port_no)
char *hostname;
int port_no;
{
	struct sockaddr_in addr;
	int sock;
	int rv;

	h_errno = 0;

	memset(&addr , '\0' ,  sizeof(addr));
	if ((addr.sin_addr.s_addr = inet_addr(hostname)) == -1)
	{
		if (dns_gethostbyname(hostname , &rv  , &(addr.sin_addr)))
		{
			return -1;
		}
	}
	_Xt_Serve
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port_no);


	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		return -1;
	}

	fcntl(sock , F_SETFL , O_NONBLOCK);

	rv = connect(sock, (struct sockaddr*)&addr, sizeof(addr));
	if (rv && (errno != EINPROGRESS) && (errno != EISCONN))
	{
		close(sock);
		return -1;
	}


#ifdef I_FACE
	if (cfg.xi_face)
	{
		if (rv && (errno == EINPROGRESS))
		{

#ifdef X_FACE
			cfg.X.inid = cfg.ctimeout ? XtAppAddTimeOut(cfg.X.app_context , 
					cfg.ctimeout * 60000, TOut , NULL) : 0;
#endif

#ifdef GTK_FACE
			cfg.gtk.inid = cfg.ctimeout ? gtk_timeout_add(cfg.ctimeout * 60000 ,
					(GtkFunction)TOut , NULL) : 0;
#endif
#ifdef X_FACE
			cfg.X.iid = XtAppAddInput(cfg.X.app_context ,
				sock , (XtPointer) (XtInputExceptMask | 
				XtInputReadMask | XtInputWriteMask) , 
				ConnectCB , NULL);
#endif

#ifdef GTK_FACE
			cfg.gtk.iid = gdk_input_add(sock , 
				GDK_INPUT_READ | GDK_INPUT_WRITE | GDK_INPUT_EXCEPTION, 
				(GdkInputFunction) ConnectCB , (gpointer) NULL);
#endif


			_Xt_ServeLoop

#ifdef X_FACE
			if (cfg.X.inid)
			{
				XtRemoveTimeOut(cfg.X.inid);
				cfg.X.inid = 0;
			}
#endif   

#ifdef GTK_FACE
			if (cfg.gtk.inid)
			{
				gtk_timeout_remove (cfg.gtk.inid);
				cfg.gtk.inid = 0;
			}
#endif
			if (cfg.rbreak || cfg.stop)
			{
#ifdef X_IFACE
				if (cfg.X.iid) 
#endif
#ifdef GTK_FACE
				if (cfg.gtk.iid) 
#endif
				{
					close(sock);
#ifdef X_IFACE
					XtRemoveInput(cfg.X.iid);
#endif

#ifdef GTK_FACE
					gdk_input_remove(cfg.gtk.iid);
#endif
					return -1;
				}
			}
			
		}
	}
	else
#endif
	{
		if (rv && (errno == EINPROGRESS))
		{
			while ((rv = _selectw(sock , 60 * cfg.ctimeout)) == -1 
				&& errno == EINTR);
			if (rv <= 0)
			{
				if (rv == 0) errno = ETIMEDOUT;
				return -1;
			}
		}
	}

	if ((connect(sock, (struct sockaddr*)&addr, sizeof(addr)) == -1) &&
	     (errno && (errno != EISCONN)))
	{
		close(sock);
		return -1;
	}
#ifdef DEBUG
	if (cfg.debug)
		xprintf(1 , gettext("Successfully connected to %s,%d\n") , hostname , port_no);
#endif

	return(sock);
}

int net_bindport(port)
int port;
{
	int sock;
	int n;
	struct sockaddr_in addr;
	char pom[150];

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror("socket");
		return -1;
	}

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;

	if (bind(sock, (struct sockaddr*) &addr, sizeof(addr)) == -1)
	{
		sprintf(pom , "bind to port(%d)" , port);
		xperror(pom);
		close(sock);
		return -1;
	}

	n = sizeof(addr);
	if (getsockname(sock , (struct sockaddr*) &addr, &n))
	{
		xperror("getsockname");
		close(sock);
		return -1;
	}

#ifdef DEBUG
	if (cfg.debug)
		xprintf(1 , gettext("BINDING to port : %d\n") , ntohs(addr.sin_port));
#endif

	if (listen(sock,1) == -1)
	{
		xperror("listen");
		close(sock);
		return -1;
	}

	return sock;
}

int net_accept(sock)
int sock;
{
	struct sockaddr_in caller;
	int p,rsock = -1;
	int rv;

	p = sizeof(caller);

	fcntl(sock , F_SETFL , O_NONBLOCK);

	rsock = accept(sock, (struct sockaddr*)&caller, &p);
	if ((rsock < 0) && (errno != EWOULDBLOCK))
	{
		close(sock);
		return -1;
	}

#ifdef I_FACE
	if (cfg.xi_face)
	{

		if ((rsock < 0) && (errno == EWOULDBLOCK))
		{
#ifdef X_FACE
			cfg.X.inid = cfg.ctimeout ? XtAppAddTimeOut(cfg.X.app_context , 
					cfg.ctimeout * 60000, TOut , NULL) : 0;
#endif

#ifdef GTK_FACE
			cfg.gtk.inid = cfg.ctimeout ? gtk_timeout_add(cfg.ctimeout * 60000 ,
					(GtkFunction)TOut , NULL) : 0;
#endif
#ifdef X_FACE
			cfg.X.iid = XtAppAddInput(cfg.X.app_context ,
				sock , (XtPointer) (XtInputExceptMask | 
				XtInputReadMask) , 
				ConnectCB , (XtPointer) NULL);
#endif

#ifdef GTK_FACE
			cfg.gtk.iid = gdk_input_add(sock , 
				GDK_INPUT_READ | GDK_INPUT_EXCEPTION, 
				(GdkInputFunction) ConnectCB , (gpointer) NULL);
#endif

			_Xt_ServeLoop

#ifdef X_FACE
			if (cfg.X.inid)
			{
				XtRemoveTimeOut(cfg.X.inid);
				cfg.X.inid = 0;
			}
#endif   

#ifdef GTK_FACE
			if (cfg.gtk.inid)
			{
				gtk_timeout_remove (cfg.gtk.inid);
				cfg.gtk.inid = 0;
			}
#endif
			if (cfg.rbreak || cfg.stop)
			{
#ifdef X_FACE
				if (cfg.X.iid) 
#endif
#ifdef GTK_FACE
				if (cfg.gtk.iid) 
#endif
				{
#ifdef X_FACE
					XtRemoveInput(cfg.X.iid);
					cfg.X.iid = 0;
#endif
#ifdef GTK_FACE
					gdk_input_remove(cfg.gtk.iid);
					cfg.gtk.iid = 0;
#endif
					return -1;
				}
			}
		}
	}
	else
#endif
	{
		if ((rsock < 0) && (errno == EWOULDBLOCK))
		{
			while ((rv = _selectr(sock , 60 * cfg.ctimeout)) == -1 
				&& errno == EINTR);
			if (rv <= 0)
			{
				if (rv == 0) errno = ETIMEDOUT;
				close(sock);
				return -1;
			}
		}
	}

	if ((rsock < 0) && ((rsock = accept(sock, (struct sockaddr*)&caller, &p)) == -1))
	{
		close(sock);
		return -1;
	}

#ifdef DEBUG
	if (cfg.debug)
		xprintf(1 , gettext("ACCEPTING connection from: %s\n") , 
			inet_ntoa(caller.sin_addr));
#endif

	return rsock;
}
