/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include "config.h"
#include "http.h"
#include "bufio.h"

#ifdef USE_SSL

#include <rsa.h>
#include <crypto.h>
#include <x509.h>
#include <pem.h>
#include <ssl.h>
#include <err.h>


/*
Most of this code is inspired by example programs from SSLeay package
*/

static int ssl_passwd_callback(char *buf, int num, int verify)
{
	if(verify)
		xprintf(1, "%s\n", buf);
	else
	{
		if(num > strlen(cfg.ssl_cert_passwd))
		{
			strcpy(buf, cfg.ssl_cert_passwd);
			return strlen(buf);
		}
	}
	return 0;
}

int ssl_do_connect(urld , sock)
url *urld;
bufio *sock;
{
	int rv;

	if (cfg.ssl_proxy)
	{
		if (http_dumy_proxy_connect(sock , url_get_site(urld) , url_get_port(urld)))
		{
			xperror("ssl proxy connect req");
			return FALSE;
		}
	}

	if (!cfg.ssl_method)
	{
		SSL_load_error_strings();
		cfg.ssl_method = SSLv23_client_method();
		SSLeay_add_ssl_algorithms();
	}

	cfg.ssl_ctx = SSL_CTX_new (cfg.ssl_method);
	
/* SSL Certification stuff */
	
	if (cfg.ssl_cert_file != NULL)
	{
		SSL *ssl;
		X509 *x509;

		if(cfg.ssl_cert_passwd)
		{
			SSL_CTX_set_default_passwd_cb(cfg.ssl_ctx , ssl_passwd_callback);
		}

		if (SSL_CTX_use_certificate_file(cfg.ssl_ctx , cfg.ssl_cert_file,
				SSL_FILETYPE_PEM) <= 0)
		{
			xprintf(1 , gettext("Unable to set certificate file (wrong password?)\n"));
			return FALSE;
		}

		if (cfg.ssl_key_file == NULL) 
			cfg.ssl_key_file = new_string(cfg.ssl_cert_file);

		if (SSL_CTX_use_PrivateKey_file(cfg.ssl_ctx , cfg.ssl_key_file,
				SSL_FILETYPE_PEM) <= 0) 
		{
			xprintf(1, gettext("Unable to set public key file\n"));
			return FALSE;
		}

		ssl = SSL_new (cfg.ssl_ctx);
	    
		x509 = SSL_get_certificate(ssl);
    
		if (x509 != NULL)
			EVP_PKEY_copy_parameters(X509_get_pubkey(x509),
			SSL_get_privatekey(ssl));

		SSL_free(ssl);

		if (!SSL_CTX_check_private_key(cfg.ssl_ctx))
		{
			xprintf(1, gettext("Private key does not match the certificate public key\n"));
			return FALSE;
		}    
	}	
/* End SSL Certification stuff */
	
	cfg.ssl_con = SSL_new (cfg.ssl_ctx);
	cfg.ssl_bio = BIO_new_socket(bufio_getfd(sock), BIO_NOCLOSE);
	SSL_set_bio (cfg.ssl_con , cfg.ssl_bio , cfg.ssl_bio);
	SSL_set_connect_state(cfg.ssl_con);
	while((rv = SSL_in_init(cfg.ssl_con)) && 
		SSL_get_error(cfg.ssl_con , rv) == SSL_ERROR_WANT_CONNECT)
	{
		printf("notconnected\n");
		_Xt_Serve;
	}

	if (rv < 0)
	{
#ifdef DEBUG
		int retv = SSL_get_error(cfg.ssl_con , -1);
		ERR_print_errors_fp(stdout);
#endif
		SSL_free (cfg.ssl_con);
		SSL_CTX_free (cfg.ssl_ctx);
		cfg.ssl_con = NULL;
		cfg.ssl_ctx = NULL;
		return FALSE;
	}
#if 0
/*** this is broken because of non-blocking IO ***/
#ifdef DEBUG
	else if (cfg.debug)
	{
		X509* server_cert;
		char *str;

		xprintf(1 , gettext("SSL connection is using %s\n"), SSL_get_cipher(cfg.ssl_con));

		server_cert = SSL_get_peer_certificate(cfg.ssl_con);
		xprintf(1 , gettext("Server certificate:\n"));
  
		str = X509_NAME_oneline(X509_get_subject_name(server_cert) ,
					NULL , 0);
		xprintf(1 , gettext("\t subject: %s\n"), str);
		_free (str);

		str = X509_NAME_oneline(X509_get_issuer_name(server_cert) ,
					NULL , 0);
		xprintf(1 , gettext("\t issuer: %s\n"), str);
		_free (str);

		X509_free (server_cert);
	}
#endif
#endif

	return TRUE;
}

#endif
