/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_FLOCK
#include <sys/file.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>
#include <ctype.h>
#include <time.h>
#include <stdarg.h>
#include <utime.h>
#include <dirent.h>

#ifdef HAVE_SYS_MODE_H
#include <sys/mode.h>
#endif

#include "config.h"
#include "url.h"
#include "doc.h"
#include "ftp.h"

void show_progres(sz)
int sz;
{
	if (cfg.progres && cfg.report_size && !cfg.quiet
#ifdef I_FACE
	    && !cfg.xi_face
#endif
           )
	{
		if (cfg.totsz || 
		    (cfg.act_doc->doc_url->type == FTP && cfg.act_doc->doc_url->extension)) 

		{
			int size = cfg.totsz ? cfg.totsz:
				((ftp_url_extension *)cfg.act_doc->doc_url->extension)->size;
			printf(gettext("Retrieved size %11d / %d B [%3d%%]\r") , sz, size, 
				(int)(100 * (double)sz / size));
		} 
/*
			printf(gettext("Retrieved size %11d / %d B \r") , sz , 
				cfg.totsz ? cfg.totsz :
				((ftp_url_extension *)cfg.act_doc->doc_url->extension)->size);
*/
		else printf(gettext("Retrieved size %11d B \r") , sz);
		fflush(stdout);
	}

#ifdef I_FACE
	if (cfg.report_size && cfg.xi_face)
	{
		iface_set_size(sz);
	}
#endif
}

#ifdef NO_SETENV
int my_setenv(var , val , ovr)
char *var;
char *val;
int ovr;
{
	char *pom=_malloc(strlen(var) + strlen(val) + 2);

	sprintf(pom , "%s=%s" , var , val);
	return putenv(pom);
}
#endif

static char *xstrherror(enr)
int enr;
{
	char *p;

	switch (enr)
	{
#ifdef NETDB_INTERNAL
		case NETDB_INTERNAL:
			p = strerror(errno);
			break;
#endif
#ifdef NETDB_SUCCESS
		case NETDB_SUCCESS:
			p = gettext("no error");
			break;
#endif
#ifdef HOST_NOT_FOUND
		case HOST_NOT_FOUND:
			p = gettext("host not found");
			break;
#endif
#ifdef TRY_AGAIN
		case TRY_AGAIN:
			p = gettext("temporary error (try again later)");
			break;
#endif
#ifdef NO_RECOVERY
		case NO_RECOVERY:
			p = gettext("non recoverable error");
			break;
#endif
#ifdef NO_ADDRESS
		case NO_ADDRESS:
			p = gettext("name is valid, but doesn't have an IP address");
			break;
#endif
		default:
			p = gettext("unknown hostname translation error");
	}
	return p;
}

void xherror(str)
char *str;
{
		xprintf(1 , "%s: %s\n" , str , xstrherror(h_errno));
} 
void xperror(str)
char *str;
{
		xprintf(1 , "%s: %s\n" , str , strerror(errno));
} 

void xprintf(int log ,char *strs , ... ) 
{
	va_list args;

	va_start(args , strs);

#ifdef I_FACE
	if (cfg.xi_face && cfg.done && !cfg.quiet)
	{
		char buf[8192];

		buf[0] = '\0';

		vsprintf(buf , strs , args);

#ifdef X_FACE
		{
			char *p;
			for(p = buf , strtok(buf , "\n\r") ; p  ; p = strtok(NULL , "\n\r"))
				LogAppendLine(cfg.X.logw , p);
		}
#endif

#ifdef GTK_FACE
		{
			gtk_text_freeze(GTK_TEXT (cfg.gtk.logw));
			gtk_text_insert (GTK_TEXT (cfg.gtk.logw), 
				NULL, &cfg.gtk.logw->style->black , NULL, buf , -1);
			gtk_text_thaw(GTK_TEXT(cfg.gtk.logw));
			if (cfg.log_autoscroll)
			{
				cfg.gtk.logvadj->value = cfg.gtk.logvadj->upper >  
					cfg.gtk.logvadj->page_size ?  cfg.gtk.logvadj->upper -  
					cfg.gtk.logvadj->page_size : 0;
				gtk_signal_emit_by_name (GTK_OBJECT(cfg.gtk.logvadj), "value_changed");
			}
		}
#endif


	}
	else	
#endif
		if (!cfg.quiet)	vprintf(strs , args);

	if (cfg.logfd >= 0 && log)
	{
		char buf[4096];

		buf[0] = '\0';

		vsprintf(buf , strs , args);
		write(cfg.logfd , buf , strlen(buf));
	}

	va_end(args);
}

unsigned int hash_func(str,num)
char *str;
int num;
{
	char *p = str;
	unsigned int rv = 0;

	while(*p)
	{
		rv = (rv + (unsigned char)*p) % num;
		p++;
	}
	rv = rv % num;
	return rv;
}

int my_flock(fd , b_lock)
int fd;
{

	int i = 0;
#ifdef HAVE_FLOCK
	if (b_lock)
	{
		if (flock(fd , LOCK_EX | LOCK_NB))
		{
			if (errno == EWOULDBLOCK)
			{
				xprintf(1 , gettext("waiting to release lock on FD : %d\n") , fd);
				i = flock(fd , LOCK_EX);
			}
		}
		else i = 0;
	}
	else
		i = flock(fd , LOCK_EX | LOCK_NB);
#else
#ifdef HAVE_FCNTL_LOCK
	struct flock fl;

	if (b_lock)
	{
		memset(&fl , '\0' , sizeof(fl));
		fl.l_type = F_WRLCK;
		if (fcntl(fd , F_SETLK , &fl))
		{
			if (errno == EWOULDBLOCK)
			{
				xprintf(1 , gettext("waiting to release lock on FD : %d\n") , fd);
				memset(&fl , '\0' , sizeof(fl));
				fl.l_type = F_WRLCK;
				i = fcntl(fd , F_SETLKW , &fl);
			}
		}
		else i = 0;
	}
	else
	{
		memset(&fl , '\0' , sizeof(fl));
		fl.l_type = F_WRLCK;
		i = fcntl(fd , b_lock ? F_SETLKW : F_SETLK , &fl);
	}
#endif
#endif
	return i;
}

#ifdef HAVE_FCNTL_LOCK
int my_funlock(fd)
int fd;
{
        struct flock fl;
	int i;

        memset(&fl , '\0' , sizeof(fl));
        fl.l_type = F_UNLCK;
        i = fcntl(fd , F_SETLK , &fl);
        
        return i;
}
#endif

/****************************************************/
/* Unlock document and remove lock file if required */
/****************************************************/
void remove_lock()
{
	struct utimbuf utmbf;

	if (cfg.s_sock)
	{
#ifdef DEBUG
		if (cfg.debug)
			xprintf(1 , gettext("Unlocking document\n"));
#endif
		_funlock(bufio_getfd(cfg.s_sock));
		bufio_close(cfg.s_sock);

		utmbf.actime = time(NULL);

		if (cfg.act_doc->dtime && cfg.preserve_time)
			utmbf.modtime = mktime(cfg.act_doc->dtime) + cfg.gmdiff;
		else
			utmbf.modtime = cfg.act_doc->stime;

		utime(cfg.lock_fn , &utmbf);

		if (cfg.remove_lock)
		{
			unlink(cfg.lock_fn);
		}
		_free(cfg.lock_fn);
	}
}

long int _atoi(str) 
char *str;
{
	char *__eptr__;
	long int rv;

	rv = strtol(str , (char **) &__eptr__ , 10);
	if (*__eptr__ != '\0') errno = ERANGE;
	else errno = 0;

	return rv;
}

char * _strtrchr(str , cfrom , cto)
char *str;
int cfrom;
int cto;
{
	char *p = str;

	while((p = strchr(p , cfrom))) *p = cto;

	return str;
}

void *_malloc(sz)
int sz;
{
	void *ret = malloc(sz);
	if (!ret)
	{
		perror("malloc");
	}
	return ret;
}

void *_realloc(from , sz)
void *from;
int sz;
{
	void *ret = realloc(from , sz);
	if (!ret)
	{
		perror("realloc");
	}
	return ret;
}

/**************************************/
/* nahrada standartnej funkcie strtok */
/* jeden separator -> jeden parameter */
/* nie je reentrantna                 */
/**************************************/
char *strtokc(str , chr)
char *str;
int chr;
{
	static char *p;
	char *pom;
	char *ret;

	if (str) p = str;

	ret = p;

	if (p)
	{
		pom = strchr(p , chr);

		if (pom)
		{
			*pom = '\0';
			p = pom  + 1;
		}
		else p = NULL;
	}
	
	return ret;
}

/**************************************/
/* najdi poziciu n-teho vyskytu znaku */
/**************************************/

char *strfindnchr(str , chr , n)
char *str;
int chr;
int n;
{
	int cnt;
	char *p;
	
	for (p = str , cnt = 0 ; *p && cnt < n ; p++)
	{
		if (*p == chr) cnt ++;
	}
	if (cnt != n) return NULL;
	else return p;
}

/********************************************/
/* search in list for string occurrence     */
/********************************************/
bool is_in_list(str , list)
char *str;
char **list;
{
        char **p = list;

        while(*p)
        {
                if (!strcasecmp(*p , str))
                {
                        return TRUE;
                }
                p++;
        }
        return FALSE;
}

/*******************************************************/
/* z retazca uroby zoznam casti oddelenych oddelovacom */
/*******************************************************/
char **make_list(liststr,sep)
char *liststr;
char *sep;
{
	char *p,*pom = new_string(liststr);
	int i = 0;
	char **ret_val = NULL;

	p = strtok(pom , sep);

	if (p)
	{
		ret_val = _malloc(sizeof(char **));
		ret_val[0] = NULL;
	}

	while(p)
	{
		ret_val = _realloc(ret_val , sizeof(char **) * (i + 2));
		ret_val[i] = new_string(p);
		ret_val[i+1] = NULL;
		p = strtok(NULL , sep);
		i++;
	}

	free(pom);

	return ret_val;
}

/***********************************************/
/* male pismena v retazci zkonvertuje na velke */
/***********************************************/

char *upperstr(s) 
char *s;
{
	char *p ; 

	for(p = s ; *p != '\0' ; p ++) *p = toupper(*p); 

	return s;
}

/***********************************************/
/* velke pismena v retazci zkonvertuje na male */
/***********************************************/

char *lowerstr(s)
char *s;
{
        char *p ;

        for(p = s ; *p != '\0' ; p ++) *p = tolower(*p);

        return s;
}

/***********************************************/
/* alokuje miesto v pameti a skopiruje retezec */
/***********************************************/

char *new_string(s)
char *s;
{
	char *p = NULL; 

	if (s)
	{
		p = (char *) _malloc(strlen(s) + 1);
		strcpy(p , s);
	}

	return p;
}

/**************************************************************/
/* alokuje miesto v pameti a skopiruje retezec dlaky l znakov */
/**************************************************************/

char *new_n_string(s , l)
char *s;
int l;
{
	char *p = (char *) _malloc(l + 1);

	strncpy(p , s , l);
	*(p + l) = '\0';
	return p;
}

/*****************************************************/
/* pre dane meno suboru vytvory vsetky neexistujuce  */
/* adresare v ceste ak je to mozne 		     */
/*****************************************************/

int makealldirs(path)
char *path;
{
	char pom[PATH_MAX + 1];
	char *p2 = new_string(path);
	char *p;

	pom[0]='\0';

	if (path)
	{
		p = p2;
#ifdef _WIN32
	if (strlen(p2) > 2 && p2[0] == '/' && p2[1] == '/' && 
	    isalpha(p2[2]) &&
	    (p2[3] == '\0' || p2[3] == '/'))
	{
		strncpy(pom , p2 ,3);
		pom[3] = '\0';
		p = p2 + 3;
	}
#endif
		p = strtok(p,"/");
		while(p)
		{
			strcat(pom , "/");
			strcat(pom , p);

			p = strtok(NULL , "/");			

			if (p && access(pom,F_OK))
			{
				mkdir(pom, S_IRWXU | S_IRGRP | S_IROTH | 
					S_IXGRP | S_IXOTH);
			}
		}
	}
	free(p2);
	return 0;
}

/***********************************************/
/* z relativnej cesty urobi absolutnu a vyhodi */
/* vsetky "." a ".." adresare		       */
/***********************************************/

char *get_abs_file_path(path)
char *path;
{
	char *p,pom[PATH_MAX],*tmp,result[PATH_MAX]="/";

	for(p = path; isspace(*p) && *p ; p++);

	if (*p != '/')
	{
		tmp = (char *) getcwd(NULL,PATH_MAX);
		sprintf(pom , "%s/%s" , cfg.cache_dir ,p);
		_free(tmp);
	}
	else
	{
		sprintf(pom , "%s" , p);
	}

	p = pom;

#ifdef _WIN32
	if (strlen(pom) >2 && pom[0] == '/' && pom[1] == '/' && 
	    isalpha(pom[2]) &&
	    (pom[3] == '\0' || pom[3] == '/'))
	{
		strncpy(result , pom ,3);
		result[3] = '\0';
		p = pom + 3;
	}
#endif
	if (!(p = strtok(p , "/"))) strcpy(result,"/");

	while(p)
	{
		if (strcmp(p , "."))
		{
			if (strcmp(p , ".."))
			{
				if (result[strlen(result) - 1] != '/') strcat(result , "/");
				strcat(result , p);
			}
			else
			{
				tmp = strrchr(result , '/');
				*(tmp + 1 - (tmp != result) ) = '\0';
			}
		}
		p = strtok(NULL , "/");
	}

	if ((*(path + strlen(path) - 1) != '/') && 
		result[strlen(result) - 1] == '/') result[strlen(result) - 1] = '\0';

	if ((*(path + strlen(path) - 1) == '/') && 
		result[strlen(result) - 1] != '/')
	{
		result[strlen(result) + 1] = '\0';
		result[strlen(result)] = '/';
	}

	return new_string(result);
}

#ifdef _WIN32
char *cvt_win32_to_unix_path(path)
char *path;
{
	char pom[PATH_MAX];
	char *p = new_string(path);
	char *p1;

	pom[0] = '\0';
	p1 = p;

	if(strlen(p) >= 2 && isalpha(p[0]) && p[1] == ':')
	{
		sprintf(pom , "//%c" , p[0]);
		p += 2;
	}

	_strtrchr(p , '\\' , '/');

	strcat(pom , p);
	_free(p1);
		
	return get_abs_file_path(pom);
}
#endif

/**********************************/
/* spocita vyskyt znaku v retazci */
/**********************************/

static int str_cnt_chr(s,c)
char *s;
int c;
{
	char *p = s;
	int ret = 0;

	while(*p)
	{
		if (*p == c) ret++;

		p++;
	}

	return ret;
}

/********************************************/
/* urci relativnu cestu z adresara v ktorom */
/* sa nachadza prvy subor na druhy subor    */
/********************************************/

char *get_relative_path(fromabs,toabs)
char *fromabs;
char *toabs;
{
	char *p1 ,*p2 , *pom1 , *pom2;
	int offset = 0 , i , plom;
	char pom[PATH_MAX];

	pom1 = p1 = get_abs_file_path(fromabs);
	pom2 = p2 = get_abs_file_path(toabs);

	while (*p1 && *p2 && *p1 == *p2)
	{
		p1++;
		p2++;
	}

	while((p1 >= pom1) && *p1 != '/') p1--;

	if (*p1 != '/') 
	{
		free(pom1);
		free(pom2);

		return NULL;
	}
	offset = p1 - pom1;

	plom = str_cnt_chr(p1 + 1 , '/');

	pom[0] = '\0';

	for (i = 0 ; i < plom ; i++)
	{
		strcat(pom , "../");
	}

	strcat(pom , pom2 + offset + 1);

	free(pom1);
	free(pom2);

	return new_string(pom);		
}

/********************************/
/* vrati poziciu pripony suboru */
/********************************/

char * get_extension(fname)
char *fname;
{
	char *p1,*p2;

	p1 = strrchr(fname , '.');
	p2 = strrchr(fname , '/');

	if (p1 > p2)
	{
		return(p1 + 1);
	}
	else return "";
	
}

static char *html_tag_tab[] = {
"<HTML>" ,
"<META" ,
"<TITLE>" ,
"<HEAD>" ,
"<BODY>",
"<html>" ,
"<title>" ,
"<head>" ,
"<body>",
};

bool ext_is_html(fn)
char *fn;
{
	char *ext = get_extension(fn);

	return str_is_in_list(0 , ext , "html" , "htm" , "shtml" , "phtml" , "css" , NULL);
}

bool file_is_html(fn)
char *fn;
{
	int i,j,len;
	char pom[512];
	bufio *fd;

	if (ext_is_html(fn))
		return TRUE;

	if ((fd = bufio_open(fn , O_BINARY | O_RDONLY)))
	{
		for (j = 0 ; j < 10 ; j++)
		{
			if ((len = bufio_readln(fd , pom , sizeof(pom))) > 0)
			{
				for (i = 0 ; i < NUM_ELEM(html_tag_tab) ; i++)
				    if (strstr(pom , html_tag_tab[i]))
				    {
				    	bufio_close(fd);
				    	return TRUE;
				    }
			}
			else
			{
				if (len < 0) xperror("file_is_html");
				bufio_close(fd);
				return FALSE;
			}
		}
		bufio_close(fd);
	}
	return FALSE;
}


#ifdef X_FACE

static void TOut(client_data , timer)
XtPointer client_data;
XtIntervalId *timer;
{
	cfg.X.inid = 0;

	_X_EscLoop();
}

static void rwCB(ptr , src , id)
XtPointer ptr;
int *src;
XtInputId *id;
{
	XtRemoveInput(*id);
	cfg.X.iid = 0;

	_X_EscLoop();
}

#endif

#ifdef GTK_FACE

static void TOut(data)
gpointer data;
{
	cfg.gtk.inid = 0;

	_X_EscLoop();
}

static void rwCB(data , source , condition)
gpointer data;
gint source;
GdkInputCondition condition;
{

	gdk_input_remove(cfg.gtk.iid);
	cfg.gtk.iid = 0;

	_X_EscLoop();
}

#endif

int _selectr(sock , timeout)
int sock;
int timeout;
{
	fd_set readfds, exceptfds;
	struct timeval tout, *toutptr = NULL;
	int rv;

	FD_ZERO(&readfds);
	FD_SET(sock , &readfds);
	FD_ZERO(&exceptfds);
	FD_SET(sock , &exceptfds);

	if (timeout)
	{
		tout.tv_sec = timeout;
		tout.tv_usec = 0;
		toutptr = &tout;
	}

	rv = select(sock + 1 , &readfds , NULL , &exceptfds , toutptr);

	return rv;
}

int _selectw(sock , timeout)
int sock;
int timeout;
{
	fd_set writefds, exceptfds;
	struct timeval tout, *toutptr = NULL;
	int rv;

	FD_ZERO(&writefds);
	FD_SET(sock , &writefds);
	FD_ZERO(&exceptfds);
	FD_SET(sock , &exceptfds);

	if (timeout)
	{
		tout.tv_sec = timeout;
		tout.tv_usec = 0;
		toutptr = &tout;
	}

	rv = select(sock + 1 , NULL , &writefds , &exceptfds , toutptr);

	return rv;
}

int my_read(sock , buf , len)
int sock;
char *buf;
int len;
{
	int rv = 0;

	do
	{
#ifdef I_FACE
		if (cfg.xi_face && cfg.done)
		{
#ifdef X_FACE
			cfg.X.inid = cfg.ctimeout ? XtAppAddTimeOut(cfg.X.app_context , 
				cfg.ctimeout * 60000, TOut , NULL) : 0;
#endif
#ifdef GTK_FACE
			cfg.gtk.inid = cfg.ctimeout ? gtk_timeout_add(cfg.ctimeout * 60000 ,
					(GtkFunction)TOut , NULL) : 0;
#endif
		}
		else
#endif
		{
			while ((rv = _selectr(sock , 60 * cfg.ctimeout)) == -1 
				&& errno == EINTR);
			if (rv <= 0)
			{
				if (rv == 0) errno = ETIMEDOUT;
				return -1;
			}
		}
#ifdef I_FACE
		if (cfg.xi_face && cfg.done)
		{
#ifdef X_FACE
			cfg.X.iid = XtAppAddInput(XtDisplayToApplicationContext(cfg.X.dpy),
				 	sock , (XtPointer) XtInputReadMask , 
					rwCB , (XtPointer)NULL);

			_Xt_ServeLoop


			if (cfg.X.inid)
			{
				XtRemoveTimeOut(cfg.X.inid);
				cfg.X.inid = 0;
			}
			else if (cfg.ctimeout)
			{
				errno = ETIMEDOUT;
				rv = -1;
			}

			if (cfg.X.iid) 
			{
				XtRemoveInput(cfg.X.iid);
				cfg.X.iid = 0;
			}

#endif /*X_FACE*/

#ifdef GTK_FACE
			cfg.gtk.iid = gdk_input_add(sock , GDK_INPUT_READ , 
				(GdkInputFunction) rwCB , (gpointer) NULL);

			_Xt_ServeLoop

			if (cfg.gtk.inid)
			{
				gtk_timeout_remove(cfg.gtk.inid);
				cfg.gtk.inid = 0;
			}
			else if (cfg.ctimeout)
			{
				errno = ETIMEDOUT;
				rv = -1;
			}

			if (cfg.gtk.iid) 
			{
				gdk_input_remove(cfg.gtk.iid);
				cfg.gtk.iid = 0;
			}

#endif /*GTK_FACE*/

			if (rv) return rv;
		}
#endif /*I_FACE*/

		if (cfg.rbreak)
		{
			errno = EINTR;
			rv = -1;
			cfg.stop = TRUE;
			break;
		}

#ifdef USE_SSL
		if (cfg.ssl_con)
		{
			bool stopread = FALSE;

			while (!stopread)
			{
				rv = SSL_read (cfg.ssl_con , buf , len);
				switch (SSL_get_error(cfg.ssl_con , rv))
				{
					case SSL_ERROR_WANT_READ:
					case SSL_ERROR_WANT_WRITE:
						my_msleep(10);
						_Xt_Serve
						break;
					case SSL_ERROR_SSL:
						ERR_print_errors(stdout);
						stopread = TRUE;
						break;
					default:
						stopread = TRUE;
						break;
				}
			}
		}
		else
#endif
			rv = read(sock , buf , len);

	} while (rv == -1 && errno == EINTR && !cfg.rbreak);


#ifdef X_FACE
	if (cfg.X.inid)
	{
		XtRemoveTimeOut(cfg.X.inid);
		cfg.X.inid = 0;
	}

	if (cfg.X.iid) 
	{
		XtRemoveInput(cfg.X.iid);
		cfg.X.iid = 0;
	}

#endif   

#ifdef GTK_FACE

	if (cfg.gtk.inid)
	{
		gtk_timeout_remove (cfg.gtk.inid);
		cfg.gtk.inid = 0;
	}

	if (cfg.gtk.iid) 
	{
		gdk_input_remove(cfg.gtk.iid);
		cfg.gtk.iid = 0;
	}

#endif   
	return rv;
}

int my_write(sock , buf , len)
int sock;
char *buf;
int len;
{
	int rv = 0;

	do
	{
#ifdef I_FACE
		if (cfg.xi_face && cfg.done)
		{
#ifdef X_FACE
			cfg.X.inid = cfg.ctimeout ? XtAppAddTimeOut(cfg.X.app_context , 
				cfg.ctimeout * 60000, TOut , NULL) : 0;
#endif
#ifdef GTK_FACE
			cfg.gtk.inid = cfg.ctimeout ? gtk_timeout_add(cfg.ctimeout * 60000 ,
					(GtkFunction)TOut , NULL) : 0;
#endif
		}
		else
#endif
		{
	
			while ((rv = _selectw(sock , 60 * cfg.ctimeout)) == -1 
				&& errno == EINTR);
			if (rv <= 0)
			{
				if (rv == 0) errno = ETIMEDOUT;
				return -1;
			}
		}

#ifdef I_FACE
		if (cfg.xi_face && cfg.done)
		{
#ifdef X_FACE
			cfg.X.iid = XtAppAddInput(XtDisplayToApplicationContext(cfg.X.dpy),
				 	sock , (XtPointer) XtInputWriteMask , 
					rwCB , (XtPointer)NULL);

			_Xt_ServeLoop


			if (cfg.X.inid)
			{
				XtRemoveTimeOut(cfg.X.inid);
				cfg.X.inid = 0;
			}
			else if (cfg.ctimeout)
			{
				errno = ETIMEDOUT;
				rv = -1;
			}

			if (cfg.X.iid) 
			{
				XtRemoveInput(cfg.X.iid);
				cfg.X.iid = 0;
			}

#endif /*X_FACE*/

#ifdef GTK_FACE
			cfg.gtk.iid = gdk_input_add(sock , GDK_INPUT_WRITE , 
				(GdkInputFunction) rwCB , (gpointer) NULL);

			_Xt_ServeLoop

			if (cfg.gtk.inid)
			{
				gtk_timeout_remove(cfg.gtk.inid);
				cfg.gtk.inid = 0;
			}
			else if (cfg.ctimeout)
			{
				errno = ETIMEDOUT;
				rv = -1;
			}

			if (cfg.gtk.iid) 
			{
				gdk_input_remove(cfg.gtk.iid);
				cfg.gtk.iid = 0;
			}

#endif /*GTK_FACE*/


			if (rv) return rv;
		}
#endif /*I_FACE*/

		if (cfg.rbreak)
		{
			errno = EINTR;
			rv = -1;
			cfg.stop = TRUE;
			break;
		}

#ifdef USE_SSL
		if (cfg.ssl_con)
			while (len)
			{
				rv = SSL_write (cfg.ssl_con , buf , len);
				switch (SSL_get_error(cfg.ssl_con , rv))
				{
					case SSL_ERROR_NONE:
						len -= rv;
						buf += rv;
						if (rv <= 0) len = 0;
						break;
					case SSL_ERROR_WANT_WRITE:
					case SSL_ERROR_WANT_READ:
						my_msleep(10);
						_Xt_Serve
						break;
					case SSL_ERROR_SSL:
						ERR_print_errors(stdout);
						break;
					default:
						len = 0;
						break;
				}
			}
		else
#endif
			rv = write(sock , buf , len);

	} while (rv == -1 && errno == EINTR && !cfg.rbreak);
   

#ifdef X_FACE
	if (cfg.X.inid)
	{
		XtRemoveTimeOut(cfg.X.inid);
		cfg.X.inid = 0;
	}

	if (cfg.X.iid) 
	{
		XtRemoveInput(cfg.X.iid);
		cfg.X.iid = 0;
	}

#endif   

#ifdef GTK_FACE

	if (cfg.gtk.inid)
	{
		gtk_timeout_remove (cfg.gtk.inid);
		cfg.gtk.inid = 0;
	}

	if (cfg.gtk.iid) 
	{
		gdk_input_remove(cfg.gtk.iid);
		cfg.gtk.iid = 0;
	}

	

#endif   

	return rv;
}

#ifdef I_FACE

void my_sleep(s)
unsigned int s;
{

	if (cfg.xi_face)
	{
#ifdef X_FACE
		XtAppAddTimeOut(cfg.X.app_context , s * 1000 , TOut , NULL);
#endif
#ifdef GTK_FACE
		gtk_timeout_add(s * 1000 , (GtkFunction) TOut , NULL);
#endif

		_Xt_ServeLoop
	}
	else sleep(s);
}
#endif

void my_msleep(ms)
unsigned int ms;
{
#ifdef I_FACE
	if (cfg.xi_face)
	{
#ifdef X_FACE
		XtAppAddTimeOut(cfg.X.app_context , ms , TOut , NULL);
#endif
#ifdef GTK_FACE
		gtk_timeout_add(ms , (GtkFunction) TOut , NULL);
#endif

		_Xt_ServeLoop
	}
	else 
#endif
#ifdef HAVE_USLEEP
		usleep(ms * 1000);
#else
	{
		struct timeval tout;

		tout.tv_sec = ms / 1000;
		tout.tv_usec = (ms % 1000) * 1000;
	}
#endif
}

int unlink_recursive(fn)
char *fn;
{
	struct stat estat;

	if (lstat(fn , &estat))
	{
		xperror(fn);
		return -1;
	}

	if (!S_ISDIR(estat.st_mode))
	{
		if (unlink(fn))
		{
			xperror(fn);
			return(-1);
		}	
	}
	else
	{
		DIR *dir;
		struct dirent *dent;
		char next_dir[PATH_MAX];

	        if (!(dir = opendir(fn)))
        	{
                	xperror(fn);
	                return -1;
        	}

	        while((dent = readdir(dir)))
        	{
                	sprintf(next_dir , "%s/%s" , fn , dent->d_name);
	                if (!strcmp(dent->d_name , ".")) continue;
        	        if (!strcmp(dent->d_name , "..")) continue;

			unlink_recursive(next_dir);
        	}

	        closedir(dir);

	        if (rmdir(fn))
	        {
	        	xperror(next_dir);
	        	return -1;
	        }
	}

	return 0;
}

int str_is_in_list(int casesensitive , char *str , ...)
{
	char *which;
	va_list	args;
	int found = FALSE;

	va_start(args , str);

	for (which = va_arg(args , char *) ; which ; which = va_arg(args , char *))
	{
		if (casesensitive ? !strcmp(str , which) : !strcasecmp(str , which))
		{
			found = TRUE;
			break;
		}		
	}

	va_end(args);

	return found;
}

int copy_fd_to_file(fd , filename)
int fd;
char *filename;
{
	char pom[8192];
	int len;
	int dfd;

	if ((dfd = open(filename , O_BINARY | O_WRONLY | O_CREAT , 
		S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR)) < 0)
	{
		xperror(filename);
		return -1;
	}

	lseek(fd , 0 , SEEK_SET);
	while((len = read(fd , pom , sizeof(pom))) > 0)
	{
		if (len != write(dfd , pom , len))
			return -1;
	}

	close(dfd);

	lseek(fd , 0 , SEEK_END);

	return len;
}

void make_clean_dir(urlp)
url *urlp;
{
	char *p,*ustr;
	int rv;

	ustr = url_to_filename(urlp);

	p = strrchr(ustr , '/');
	if (p) *p = '\0';

	while ((strlen(ustr) + 1) > strlen(cfg.cache_dir))
	{

		rv = rmdir(ustr);

		if (rv)
		{
			if (errno != ENOTEMPTY && errno != ENOENT && errno != EEXIST)
				xperror(ustr);
			break;
		}

		p = strrchr(ustr , '/');
		if (p) *p = '\0';
	}
}
