/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include "config.h"
#include "tools.h"

#ifndef isblank
#define isblank(c) ((c) == ' ' || (c) == '\t')
#endif

#ifndef isgraph
#define isgraph(c) (isprint (c) && !isspace (c))
#endif

#define HEXASC2HEXNR(x) (((x) >= '0' && (x) <= '9') ? \
	((x) - '0') : (tolower(x) - 'a' + 10))

typedef enum {
	TR_ALPHA ,
	TR_ALNUM ,
	TR_NUM ,
	TR_XNUM ,
	TR_SPACE ,
	TR_BLANK ,
	TR_CTRL ,
	TR_PRINTABLE ,
	TR_UPPER ,
	TR_LOWER ,
	TR_PUNCT ,
	TR_GRAPH ,
	TR_NONPRINTABLE ,
	TR_BADCLS 
} tr_cls;

typedef struct {
	char 	*name;
	tr_cls	cls;
} tr_cls_str;

static tr_cls_str tr_cls_map[] = {
	{"[:upper:]" , TR_UPPER } ,
	{"[:lower:]" , TR_LOWER } ,
	{"[:alpha:]" , TR_ALPHA } ,
	{"[:alnum:]" , TR_ALNUM } ,
	{"[:digit:]" , TR_NUM } ,
	{"[:xdigit:]" , TR_XNUM } ,
	{"[:space:]" , TR_SPACE } ,
	{"[:blank:]" , TR_BLANK } ,
	{"[:cntrl:]" , TR_CTRL } ,
	{"[:print:]" , TR_PRINTABLE } ,
	{"[:nprint:]" , TR_NONPRINTABLE },
	{"[:punct:]" , TR_PUNCT } ,
	{"[:graph:]" , TR_GRAPH } 
};

static int tr_is_cls(c , cls)
unsigned char c;
tr_cls cls;
{
	switch(cls)
	{
        	case TR_ALPHA :
			return isalpha(c);
		break;
        	case TR_ALNUM :
			return isalnum(c);
		break;
        	case TR_NUM :
			return isdigit(c);
		break;
        	case TR_XNUM :
			return isxdigit(c);
		break;
        	case TR_SPACE :
			return isspace(c);
		break;
        	case TR_BLANK :
			return isblank(c);
		break;
        	case TR_CTRL :
			return iscntrl(c);
		break;
        	case TR_PRINTABLE :
			return isprint(c);
		break;
        	case TR_UPPER :
			return isupper(c);
		break;
        	case TR_LOWER :
			return islower(c);
		break;
        	case TR_PUNCT :
			return ispunct(c);
		break;
        	case TR_GRAPH :
			return isgraph(c);
		break;
        	case TR_NONPRINTABLE :
			return !isprint(c);
		break;
		default:
			return FALSE;
	}
}

static unsigned int tr_append_cls(p , cls)
char *p;
tr_cls cls;
{
	unsigned int i;
	unsigned int apnd = 0;

	for (i = 0 ; i <= 255 ; i++)
	{
		if (tr_is_cls((unsigned char) i , cls))
		{
			p[apnd] = (unsigned char) i;
			p[apnd+1] = '\0';
			apnd ++;
		}	
	}

	return apnd;
}

static unsigned char tr_get_escaped_str(p)
char **p;
{
	unsigned char *ps = *p;
	unsigned char rc;

	switch (ps[1])
	{
		case 'n':
			rc = '\n';
			(*p) ++;
		break;
		case 'r':
			rc = '\r';
			(*p) ++;
		break;
		case 't':
			rc = '\t';
			(*p) ++;
		break;
		break;
		case '0':
			if (ps[2] == 'x')
			{
				if (isxdigit(ps[3])  && isxdigit(ps[4]))
				{
					rc = (HEXASC2HEXNR(ps[3]) << 4) + HEXASC2HEXNR(ps[4]);
					(*p) += 4;
				}
				else
					rc = ps[0];
			}
			else
				rc = ps[0];
		break;
		default:
			rc = ps[1];
			(*p) ++;
	}

	return rc;
}

char *tr_expand_str(str , rlen)
unsigned char *str;
int *rlen;
{
	unsigned char *p;
	unsigned char pom[2048];
	int i;

	pom[0] = '\0';

	for (i = 0 , p = str ; *p ; p++)
	{
		switch (*p)
		{
			case '\\':
				pom[i] = tr_get_escaped_str(&p);
				pom[i+1] = '\0';
				i++;
			break;
			case '[':
			{
				int j;
				tr_cls cls = TR_BADCLS;

				for (j = 0 ; j < TR_BADCLS ; j++)
				{
					if (!strncmp(p , tr_cls_map[j].name , strlen(tr_cls_map[j].name)))
					{
						cls = tr_cls_map[j].cls;
						p += strlen(tr_cls_map[j].name) - 1;
						break;
					}
				}
				if (cls != TR_BADCLS)
					i += tr_append_cls((pom + i) , cls);
				else
				{
					pom[i] = *p;
					pom[i+1] = '\0'; 
					i++;	
				}
			}
			break;
			case '-':
			{
				unsigned char strtc;
				unsigned char endc;
				unsigned int pc;

				if (i)
					strtc = pom[i-1] + 1;
				else
					strtc = '\0';

				if (*(p+1))
				{
					p++;
					if (*p == '\\')	
					{
						endc = tr_get_escaped_str(&p);
					}
					else
						endc = *p;
				}
				else
					endc = '\255';

				for (pc = strtc ; pc <= endc ; pc++)
				{
					pom[i] = (unsigned char) pc;
					i++;
				}
			}		
			break;
			default:
				pom[i] = *p;
				pom[i+1] = '\0'; 
				i++;	
		}
	}

	if (rlen)
		*rlen = i;

	return (new_string(pom));
}

char *tr_chr_chr(fset , tset , str)
unsigned char *fset;
unsigned char *tset;
unsigned char *str;
{
	unsigned char *p , *d;
	int i;
	int tsetlen = strlen(tset);
	unsigned char *retv = new_string(str); 

	for (p = str , d = retv; *p ; p++ , d++)
	{
		for (i = 0 ; fset[i] ; i++)
		{
			if (fset[i] == *p)
			{
					*d = tset[(tsetlen > i) ? i : (tsetlen - 1)];
					break;
			}
		}
	}

	return retv;
}

char *tr_del_chr(set , str)
unsigned char *set;
unsigned char *str;
{
	unsigned char *p , *d;
	int i;
	unsigned char *retv = new_string(str); 
	int found;

	for (p = str , d = retv; *p ; p++)
	{
		found = FALSE;
		for (i = 0 ; set[i] ; i++)
		{
			if (set[i] == *p)
			{
				found = TRUE;
				break;
			}
		}
		if (!found)
		{
			*d = *p;
			d++;
			*d = '\0';
		}
	}

	return retv;
}

char *tr_str_str(s1 , s2 , str)
unsigned char *s1;
unsigned char *s2;
unsigned char *str;
{
	unsigned char *p = str , *p1 , *retv;
	int i = 0;

	while (p)
	{
		if ((p = strstr(p , s1)))
		{
			i++;
			p += strlen(s1);
		}
	}

	retv = (unsigned char *) malloc (1 + strlen(str) - i * strlen(s1) + i * (s2 ? strlen(s2) : 0));
	memset(retv , '\0' , 1 + strlen(str) - i * strlen(s1) + i * (s2 ? strlen(s2) : 0));

	p = p1 = str;

	while(p1)
	{
		p1 = strstr(p , s1);
		if (p1)
		{
			strncat(retv , p , p1 - p);
			if (s2) strcat(retv , s2);
		}
		else
			strcat(retv , p);

		p = p1 + strlen(s1);
	}

	return retv;
}

char *tr(str)
char *str;
{
	char *p1,*p2;
	char *s1,*s2;

	if (cfg.tr_str_s1 && cfg.tr_str_s2)
	{
		p1 = tr_str_str(cfg.tr_str_s1 , cfg.tr_str_s2 , str);
	}
	else
	{
		p1 = new_string(str);
	}

	if (cfg.tr_del_chr)
	{
		s1 = tr_expand_str(cfg.tr_del_chr , NULL);
		p2 = tr_del_chr(s1 , p1);
		_free(s1);
		_free(p1);
	}
	else
	{
		p2 = p1;
	}

	if (cfg.tr_chr_s1 && cfg.tr_chr_s2)
	{
		s1 = tr_expand_str(cfg.tr_chr_s1 , NULL);
		s2 = tr_expand_str(cfg.tr_chr_s2 , NULL);
		p1 = tr_chr_chr(s1 , s2 , p2);
		_free(s1);
		_free(s2);
		_free(p2);
	}
	else
	{
		p1 = p2;
	}

	return p1;
}

