/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>

#include "config.h"
#include "http.h"
#include "ftp.h"
#include "gopher.h"
#include "url.h"
#include "html.h"
#include "tools.h"
#include "authinfo.h"
#include "tr.h"

/* refers to RFC 1738 */
#define UNSAFE_CHARS " <>\"#%{}|\\^~[]`+!@=&()"
#define UNSAFE_SCHARS " <>\"#%{}|\\^~[]`!@()"
#define IS_UNSAFE(c) (strchr(UNSAFE_CHARS , c ) || ((unsigned char)c > 0x7f) || ((unsigned char)c < 0x20))
#define IS_SUNSAFE(c) (strchr(UNSAFE_SCHARS , c ) || ((unsigned char)c > 0x7f) || ((unsigned char)c < 0x20))


/* for hexadicimal encoding */
static char hexa[] = "0123456789ABCDEF";
#define HEXASC2HEXNR(x) (((x) >= '0' && (x) <= '9') ? \
	((x) - '0') : (toupper(x) - 'A' + 10))

#define HEX2CHAR(x) (HEXASC2HEXNR(*(x + 1)) << 4) + HEXASC2HEXNR(*(x + 2))


protinfo prottable[] = {
	{UNKNOWN , NULL , NULL , NULL , 0 , FALSE} ,
	{BROKEN , NULL , NULL , NULL , 0 , FALSE} ,
	{HTTP , "http" , "http" , "http://" , 80 , TRUE} ,
#ifdef USE_SSL
	{HTTPS , "https" , "https" , "https://" , 443 , TRUE} ,
#else
	{HTTPS , "https" , "https" , "https://" , 443 , FALSE} ,
#endif
	{SHTTP , "shttp" , "shttp" , "shttp://" , 0 , FALSE} ,
	{FTP , "ftp" , "ftp" , "ftp://" , 21 , TRUE} ,
	{FILET , NULL , "file" , "file://" , 0 , TRUE} ,
	{GOPHER , "gopher" , "gopher" , "gopher://" , 70 , TRUE} ,
	{NEWS , "news" , "news" , "news:" , 119 , FALSE} ,
	{NNTP , "nntp" , "nntp" , "nntp://" , 119 , FALSE} ,
	{WAIS , NULL , "wais" , "wais://", 0 , FALSE} ,
	{MAILTO , NULL , "mailto" , "mailto:", 25 , FALSE} ,
	{TELNET , NULL , "telnet" , "telnet://" , 21 , FALSE} ,
	{RLOGIN , NULL , "rlogin" , "rlogin://" , 513 , FALSE} ,
	{TN3270 , NULL , "tn3270" , "tn3270://" , 0 , FALSE} ,
	{X500 , NULL , "x500" , "x500://" , 0 , FALSE} ,
	{WHOIS , NULL , "whois" , "whois://" , 0 , FALSE} ,
	{PROSPERO , NULL , "prospero" , "prospero://" , 0 , FALSE} ,
	{JAVASCRIPT , NULL , "javascript" , "javascript:" , 0 , FALSE} ,
	{URN , NULL , "urn" , "urn://" , 0 , FALSE} ,
	{LDAP , NULL , "ldap" , "ldap://" , 0 , FALSE} ,
	{Z39_50S , NULL , "z39_50s" , "z39_50s://" , 0 , FALSE} ,
	{Z39_50R , NULL , "z39_50r" , "z39_50r://" , 0 , FALSE} ,
};


/***********************************/
/* analyza jednotlivych typov URL  */
/***********************************/
url *parse_url(_urlstr)
char *_urlstr;
{
	char *p , *p3 , *p4;
	char *p2;
	char *urlstr;
	url ret_url;
	int i;


	if (strchr(_urlstr,'%'))
	{
		urlstr = decode_url(_urlstr);
	}
	else
		urlstr = new_string(_urlstr);

	ret_url.type = UNKNOWN;
	ret_url.status = 0;
	ret_url.parent_url = NULL;
	ret_url.moved_to = NULL;
	ret_url.ref_cnt = 1;
	ret_url.parent_cnt = 0;
	ret_url.level = 0;
	ret_url.extension = NULL;

#ifdef WITH_TREE
#ifdef I_FACE
	ret_url.prop = NULL;
	ret_url.tree_nfo = NULL;
#endif
#endif

	p2 = new_string(urlstr);
	p = strtok(p2,":");
	p = strtok(NULL,":");
	free(p2);

	p2 = new_string(urlstr);

	if (!p)
	{
		ret_url.type = FILET;
                ret_url.p.file.anchor_name = NULL;
                p = strtokc(p2 , '#');
                p = strtokc(NULL , '#');
                ret_url.p.file.filename = _strtrchr(new_string(p2) , '\\' , '/');
                if (p) ret_url.p.file.anchor_name = new_string(p);

                free(p2);
		free(urlstr);

		return new_url(&ret_url);
	}

	p = strtok(p2," :");

	lowerstr(p);

	for( i = 0 ; i < NUM_ELEM(prottable) ; i++) 
	{
		if (prottable[i].urlid)
		{
			if (!strcmp(prottable[i].urlid , p))
			{
				ret_url.type = i;
				break;
			}
		}
	}

	switch (i)
	{
	    case HTTP:
#ifdef USE_SSL
	    case HTTPS:
#endif
	    {
		ret_url.type = i;

		free(p2);

		/* check for separator */
		if (strncmp(urlstr + strlen(prottable[i].urlid) , "://" , 3) != 0)
		{
			ret_url.type = FILET;
			p2 = new_string(urlstr + strlen(prottable[i].urlid) + 1);
			p3 = strtok(p2,"#");
			ret_url.p.file.filename = _strtrchr(new_string(p2) , '\\' , '/');
			p3 = strtok(NULL,"#");
			if (p3)
			{
				ret_url.p.file.anchor_name = new_string(p3);
			}
			else
			{
				ret_url.p.file.anchor_name = NULL;
			}
			free(p2);
			free(urlstr);
			return new_url(&ret_url);
		}

		ret_url.p.http.password = NULL;
		ret_url.p.http.user = NULL;

		/* determine hostname */
		p2 = new_string(urlstr);
		
		strtok(p2 + strlen(prottable[i].typestr) , "/");

		p = p2 + strlen(prottable[i].typestr);
		p3 = strrchr(p,':');
		p4 = strchr(p,'@');
		if ((p4 && p3 && (p3 > p4)) || (p3 && !p4)) *p3 = '\0';
		ret_url.p.http.host = new_string(p);

		free(p2);

		/* determine port number */
		ret_url.p.http.port = prottable[i].default_port;

		if (*(urlstr + strlen(prottable[i].typestr) + strlen(ret_url.p.http.host)) == ':')
		{
			p2 =  new_string(urlstr);
			p = strtok(p2 + strlen(prottable[i].typestr) + 1 + strlen(ret_url.p.http.host) , "/");

			if (p)
			{
				ret_url.p.http.port = _atoi(p);
				if (errno == ERANGE)
				{
					ret_url.p.http.port = prottable[i].default_port;
				}
			}
			else
			{
				ret_url.p.http.port = 
					_atoi(urlstr + strlen(prottable[i].typestr) + 1 + strlen(ret_url.p.http.host));
				if (errno == ERANGE)
				{
					ret_url.p.http.port = prottable[i].default_port;
				}
			}
			
			free(p2);
		}

		/* determine document path */
		ret_url.p.http.anchor_name = NULL;
		ret_url.p.http.document = NULL;
		ret_url.p.http.searchstr = NULL;
		p2 = new_string(urlstr);
		p = strchr(p2 + strlen(prottable[i].typestr) + strlen(ret_url.p.http.host) , '/');
		if (p)
		{
			char *sstr,*astr;

			sstr = strchr(p , '?');
			astr = strchr(p , '#');

			if ((sstr && astr && sstr < astr) || sstr)
			{
				*sstr = '\0';
				ret_url.p.http.searchstr = new_string(sstr+1);
			}
			else if (astr)
			{
				*astr = '\0';
				ret_url.p.http.anchor_name = new_string(astr+1);
			}
			ret_url.p.http.document = get_abs_file_path(_strtrchr(p , '\\' , '/'));
			
		} else 	ret_url.p.http.document = new_string("/");

		if (strchr(ret_url.p.http.host,'@'))
		{
			p = strtok(ret_url.p.http.host,"@");
			ret_url.p.http.host = new_string(strtok(NULL,""));

			p3 = strtokc(p , ':');
			ret_url.p.http.user = new_string(p3 ? (*p3 ? p3 : NULL) : NULL);
			ret_url.p.http.password = new_string(strtokc(NULL , '\0'));
			_free(p);
		}

		free(p2);
		free(urlstr);
		lowerstr(ret_url.p.http.host);

		return new_url(&ret_url);
	    }
	    break;	
	    case FILET:
	    {
		free(p2);	
		p2 = new_string(urlstr);

		ret_url.p.file.anchor_name = NULL;
		/* check for separator */
		if (strncmp(urlstr + 4 , "://" , 3) != 0)
		{
			p = strtok(p2,"#");
			p = strtok(NULL , "#");
			ret_url.p.file.filename = _strtrchr(new_string(p2) , '\\' , '/');
			if (p) ret_url.p.file.anchor_name = new_string(p);

			free(p2);
			free(urlstr);
			return new_url(&ret_url);
		}
		p = strtok(p2,"#");
		p = strtok(NULL , "#");
		ret_url.p.file.filename = _strtrchr(new_string(p2 + 7) , '\\' , '/');
		if (p) ret_url.p.file.anchor_name = new_string(p);

		free(p2);
		free(urlstr);

		return new_url(&ret_url);
	    }
	    break;
	    case FTP:
	    {
	    	ret_url.p.ftp.dir = FALSE;
		ret_url.p.ftp.host = NULL;
		ret_url.p.ftp.password = NULL;
		ret_url.p.ftp.user = NULL;
		ret_url.p.ftp.path = NULL;
		ret_url.p.ftp.anchor_name = NULL;

		free(p2);

		/* check for separator */
		if (strncmp(urlstr + 3 , "://" , 3) != 0)
		{
			ret_url.type = FILET;
			p2 = new_string(urlstr + 4);
			p3 = strtok(p2,"#");
			ret_url.p.file.filename =  new_string(_strtrchr(p2 , '\\', '/' ));
			p3 = strtok(NULL,"#");
			if (p3)
			{
				ret_url.p.file.anchor_name = new_string(p3);
			}
			else
			{
				ret_url.p.file.anchor_name = NULL;
			}
			free(urlstr);
			free(p2);
			return new_url(&ret_url);
		}

		strtok(urlstr,";");

		/* determine hostname */
		p2 = new_string(urlstr);
		
		p = strtok(p2 + 6 , "/");

		p = p2+6;
		p3 = strrchr(p,':');
		p4 = strchr(p,'@');
		if ((p4 && p3 && (p3 > p4)) || (p3 && !p4)) *p3 = '\0';
		ret_url.p.ftp.host = new_string(p);	
		free(p2);

		/* determine port number */
		ret_url.p.ftp.port = prottable[i].default_port;
		if (*(urlstr + 6 + strlen(ret_url.p.ftp.host)) == ':')
		{
			p2 =  new_string(urlstr);
			p = strtok(p2 + 7 + strlen(ret_url.p.ftp.host) , "/");

			if (p)
			{
				ret_url.p.ftp.port = _atoi(p);
				if (errno == ERANGE)
				{
					ret_url.p.ftp.port = prottable[i].default_port;
				}
			}
			else
			{
				ret_url.p.ftp.port = 
					_atoi(urlstr + 7 + strlen(ret_url.p.ftp.host));
				if (errno == ERANGE)
				{
					ret_url.p.ftp.port = prottable[i].default_port;
				}
			}
			
			free(p2);
		}

		/* determine document path */
		ret_url.p.ftp.path = NULL;
		p2 = new_string(urlstr);
		p = strchr(p2 + 6 + strlen(ret_url.p.ftp.host) , '/');
		if (p)
		{
			p = strtokc(p , '#');
			ret_url.p.ftp.anchor_name = new_string(strtokc(NULL , '\0'));
			ret_url.p.ftp.path = get_abs_file_path(_strtrchr(p , '\\' , '/'));
			
		} else 	ret_url.p.ftp.path = new_string("/");

		free(urlstr);
		free(p2);

		if (strchr(ret_url.p.ftp.host,'@'))
		{
			p = strtok(ret_url.p.ftp.host,"@");
			ret_url.p.ftp.host = new_string(strtok(NULL,""));

			p3 = strtokc(p , ':');
			ret_url.p.ftp.user = new_string(p3 ? (*p3 ? p3 : NULL) : NULL);
			ret_url.p.ftp.password = new_string(strtokc(NULL,'\0'));

			_free(p);
		}
		ret_url.p.ftp.dir = 
			*(ret_url.p.ftp.path + strlen(ret_url.p.ftp.path) - 1) == '/';

		lowerstr(ret_url.p.ftp.host);	

		return new_url(&ret_url);

	    }
	    break;
	    case GOPHER:
	    {
		free(p2);
		ret_url.p.gopher.type = '1';
		ret_url.p.gopher.selector = NULL;
		ret_url.p.gopher.host = NULL;
		ret_url.p.gopher.port = DEFAULT_GOPHER_PORT;

		/* check for separator */
		if (strncmp(urlstr + 6 , "://" , 3) != 0)
		{
			ret_url.type = FILET;
			p2 = new_string(urlstr);
			p3 = strtok(p2,"#");
			ret_url.p.file.filename = new_string(p2);
			p3 = strtok(NULL,"#");
			if (p3)
			{
				ret_url.p.file.anchor_name = new_string(p3);
			}
			else
			{
				ret_url.p.file.anchor_name = NULL;
			}
			free(p2);
			free(urlstr);
			return new_url(&ret_url);
		}

		/* determine hostname */
		p2 = new_string(urlstr);
		
		p = strtok(p2 + 9 , ":/");

		if  (!p)
		{
			ret_url.p.gopher.host = new_string(urlstr + 9);
		}
		else
		{
			ret_url.p.gopher.host = new_string(p);
		}
		free(p2);

		/* determine port number */
		ret_url.p.gopher.port = prottable[i].default_port;
		if (*(urlstr + 9 + strlen(ret_url.p.gopher.host)) == ':')
		{
			p2 =  new_string(urlstr);
			p = strtok(p2 + 10 + strlen(ret_url.p.gopher.host) , "/");

			if (p)
			{
				ret_url.p.gopher.port = _atoi(p);
				if (errno == ERANGE)
				{
					ret_url.p.gopher.port = prottable[i].default_port;
				}
			}
			else
			{
				ret_url.p.gopher.port = 
					_atoi(urlstr + 10 + strlen(ret_url.p.gopher.host));
				if (errno == ERANGE)
				{
					ret_url.p.gopher.port = prottable[i].default_port;
				}
			}
			
			free(p2);
		}
		/* determine type && selector */

		p2 = new_string(urlstr);
		p = strchr(p2 + 9 + strlen(ret_url.p.gopher.host) , '/');
		if (p)
		{
			if (*(p+1))
			{
				ret_url.p.gopher.type = *(p+1);
				ret_url.p.gopher.selector = new_string(p+2);
			}
			else
			{
				ret_url.p.gopher.type = '1';
			}
		}

		free(p2);
		free(urlstr);
		
		lowerstr(ret_url.p.gopher.host);
		if (ret_url.p.gopher.selector == NULL)
			ret_url.p.gopher.selector = new_string("");
		return new_url(&ret_url);

	    }
	    break;
#ifndef USE_SSL
	    case HTTPS:
#endif
	    case SHTTP:
	    case NEWS:
	    case NNTP:
	    case WAIS:
	    case WHOIS:
	    case MAILTO:
	    case PROSPERO:
	    case TELNET:
	    case TN3270:
	    case X500:
	    case RLOGIN:
	    case JAVASCRIPT:
	    case URN:
	    case LDAP:
	    case Z39_50R:
	    case Z39_50S:
	    {
		free(p2);
	    }
	    break;
	    default:
	    {
		ret_url.type = UNKNOWN;
		free(p2);	
	    }
	}

	free(urlstr);
	return new_url(&ret_url);
}

/**** presmerovanie URL na absolutne ****/
char *url_to_absolute_url(base , baset , parent , act)
char *base;
char *baset;
url *parent;
char *act;
{
	char *psp = NULL;
	url *purl = parse_url(act);

	if ((parent->type == GOPHER) &&
	    cfg.gopher_proxy && cfg.gopher_via_http &&
	    !(parent->status & URL_REDIRECT) &&
	    (!strncmp(act, "//", 2)))
	{
		char pom[4096];
		
		sprintf(pom , "gopher:%s" , act);
		psp = new_string(pom);
	}
	else if (purl->type == FILET && 
		(parent->type == FILET))
	{
		char pom[2048];

		if (!(*purl->p.file.filename))
		{
			strcpy(pom,baset);
		}
		else 
		{
			if (*(purl->p.file.filename) != '/')
			{
				strcpy(pom , base);
				strcat(pom , purl->p.file.filename);

				free(purl->p.file.filename);
				purl->p.file.filename = new_string(pom);						
			}
			else
				sprintf(pom , "%s%s" , 
					prottable[purl->type].typestr ,
					purl->p.file.filename);
		}
		psp = new_string(pom);
	}
	else if ((purl->type == FILET) && 
		(parent->type == HTTP ||
		 parent->type == HTTPS ||
		 parent->type == FTP))
	{
		char pom[2048];
		char *ri;
		if (*(purl->p.file.filename) == '/')
		{
			char *idx;

			strcpy(pom , base);
			idx = strfindnchr(pom , '/' , 3);
			if (idx)
				strcpy(idx - 1 , purl->p.file.filename );
			else
				strcat(pom , purl->p.file.filename );
 
			if ( purl->p.file.anchor_name)
			{
				strcat(pom , "#");
				strcat(pom , purl->p.file.anchor_name);
			}
		}
		else if (!(*purl->p.file.filename))
		{
			strcpy(pom,baset);
			if (purl->p.file.anchor_name)
			{
				strcat(pom , "#");
				strcat(pom , purl->p.file.anchor_name);
			}	
		}
		else
		{
			strcpy(pom , base);
			ri = strrchr(pom , '/');
			strcpy(ri + 1 , purl->p.file.filename);

			if ((strlen(purl->p.file.filename) >= strlen(cfg.index_name)) &&
			    !strcmp(cfg.index_name , purl->p.file.filename + strlen(purl->p.file.filename) - strlen(cfg.index_name)))
			{
				*(pom + strlen(pom) - strlen(cfg.index_name)) = '\0';
			}

			if (purl->p.file.anchor_name)
			{
				strcat(pom , "#");
				strcat(pom , purl->p.file.anchor_name);
			}
		}
		psp = new_string(pom);
	}
	else
	{
		psp = new_string(act);	
	}

	free_deep_url(purl);
	_free(purl);

	if (psp)
	{
		purl = parse_url(psp);
		url_path_abs(purl);
		_free(psp);
		psp = url_to_urlstr(purl);
		free_deep_url(purl);
		_free(purl);		
	}

	return psp;
}

/**************************************/
/* zakodovanie znakov v URL , ktore   */
/* nie su bezpecne alebo povolene     */
/**************************************/
char * encode_url(urlstr)
char *urlstr;
{
	unsigned char *res,*p,*r;

	if (urlstr == NULL) return NULL;

	res = _malloc(strlen(urlstr) * 3 + 1);

	for (p = urlstr , r = res ; *p ; p++ , r++)
	{
		if (IS_UNSAFE(*p)) 
		{
			*r = '%';
			r++;
			*r = hexa[*p >> 4];
			r++;
			*r = hexa[*p % 16];
		}
		else
		{
			*r = *p;
		}
	}
	*r = '\0';

	return res;
}

/*********************************************/
/* zakodovanie znakov v search stringu URL   */
/*********************************************/
char * encode_searchstr(urlstr)
char *urlstr;
{
	unsigned char *res,*p,*r;

	if (urlstr == NULL) return NULL;

	res = _malloc(strlen(urlstr) * 3 + 1);

	for (p = urlstr , r = res ; *p ; p++ , r++)
	{
		if (IS_SUNSAFE(*p)) 
		{
			*r = '%';
			r++;
			*r = hexa[*p >> 4];
			r++;
			*r = hexa[*p % 16];
		}
		else
		{
			*r = *p;
		}
	}
	*r = '\0';

	return res;
}

/*****************************************/
/* dekodovanie zakodovanych znakov z URL */
/*****************************************/
char * decode_url(urlstr)
char *urlstr;
{
	char *res,*p,*r;

	if (urlstr == NULL) return NULL;

	res = _malloc(strlen(urlstr) + 1);

	for (p = urlstr , r = res ; *p ; r++ , p++)
	{
		if (*p == '%' && *(p + 1) && *(p + 2) && 
			isxdigit(*(p + 1)) && isxdigit(*(p + 2)))
		{
			*r = HEX2CHAR(p);
			p += 2;
		}
		else
		{
			*r = *p;
		}
	}
	*r = '\0';
	return res;
}

/*************************************/
/* uvolnenie pamate po strukture URL */
/*************************************/
void free_deep_url(urlp)
url *urlp;
{
	switch(urlp->type)
	{
		case FILET:
			_free(urlp->p.file.filename);
			_free(urlp->p.file.anchor_name);
			break;
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			_free(urlp->p.http.host);
			_free(urlp->p.http.document);
			_free(urlp->p.http.searchstr);
			_free(urlp->p.http.anchor_name);
			_free(urlp->p.http.password);
			_free(urlp->p.http.user);
			break;
		case FTP:
			_free(urlp->p.ftp.host);
			_free(urlp->p.ftp.user);
			_free(urlp->p.ftp.password);
			_free(urlp->p.ftp.anchor_name);
			_free(urlp->p.ftp.path);
			if (urlp->extension)
			{
				_free(((ftp_url_extension*)urlp->extension)->slink);
				_free(urlp->extension);
			}
			break;
		case GOPHER:
			_free(urlp->p.gopher.host);
			_free(urlp->p.gopher.selector);
		default:
			break;
	}

	_free(urlp->parent_url);
	
#ifdef WITH_TREE
#ifdef I_FACE
	_free(urlp->tree_nfo);

	if (urlp->prop)
	{
		_free(urlp->prop->type);
		free(urlp->prop);
	}
#endif
#endif
}

void append_url_to_list(urlp)
url *urlp;
{
	if (!prottable[urlp->type].supported)
	{
		xprintf(1 , gettext("unsupported URL type \"%s\"\n") , 
			prottable[urlp->type].urlid ? prottable[urlp->type].urlid :
			gettext("unknown"));
		return;
	}

	urlp->ref_cnt = 1;

	cfg.urlstack =_realloc(cfg.urlstack , 
		(cfg.total_cnt + 2) * sizeof(url *));
	cfg.urlstack[cfg.total_cnt] =  urlp;
	cfg.urlstack[cfg.total_cnt + 1] =  NULL;
	cfg.total_cnt++;


	url_add_to_hash_tab(urlp);

	if (!urlp->parent_url)
	{
		urlp->parent_url = _malloc(sizeof(url *));
		urlp->parent_url[0] = NULL;
	}

#ifdef WITH_TREE
#ifdef I_FACE
	if (cfg.xi_face)
	{
		urlp->tree_nfo = (GUI_TREE_RTYPE*) _malloc(sizeof(GUI_TREE_RTYPE));
		urlp->tree_nfo[0] = iface_make_tree_entry(urlp);
	}
#endif
#endif
}

void link_url_in_list(orig , copy)
url *orig;
url *copy;
{

	if (copy->parent_url && (orig != copy->parent_url[0]))
	{
		int i;
		bool found = FALSE;

		for (i = 0 ; i < orig->ref_cnt ; i++)
			if (orig->parent_url[i] == copy->parent_url[0]) found = TRUE;

		if (!found)
		{
			orig->ref_cnt ++;

			orig->parent_url = (url **)_realloc(orig->parent_url , (orig->ref_cnt + 1) * sizeof(url *));
			orig->parent_url[orig->ref_cnt - 1] = copy->parent_url[0];
			orig->parent_url[orig->ref_cnt] = NULL;

			if ((copy->parent_url[0]) && (orig->status & URL_MOVED) && (orig->status & URL_MOVED))
			{
				url *purl = orig;
				char *fn = NULL;

				while(purl->moved_to) purl = purl->moved_to;

				if (purl->status & URL_DOWNLOADED)
					fn = url_to_filename(purl);

				rewrite_one_parent_links(copy , copy->parent_url[0] , fn);

				_free(fn);
			}
#ifdef WITH_TREE
#ifdef I_FACE
			if (cfg.xi_face)
			{
				orig->tree_nfo = (GUI_TREE_RTYPE*) _realloc(orig->tree_nfo , orig->ref_cnt *sizeof(GUI_TREE_RTYPE));
				orig->tree_nfo[orig->ref_cnt - 1] = iface_make_tree_entry(orig);
			}
#endif
#endif
		}
	}
}

void url_redirect_to(src , dst)
url *src;
url *dst;
{
	url *pomurl,*pomurl2;

	src->status |= URL_MOVED;

	if ((pomurl = url_was_befor(dst)))
	{
		pomurl2 = pomurl;
		while (pomurl2)
		{
			if (src == pomurl2) 
			{
				src->status &= ~URL_MOVED;
				cfg.errcode = ERR_HTTP_CYCLIC;
				return;
			}
			pomurl2 = pomurl2->moved_to;
		}

		pomurl->parent_url = _realloc(pomurl->parent_url , 
				(pomurl->ref_cnt + 2) * sizeof(url *));
		pomurl->parent_url[pomurl->ref_cnt] = src;
		pomurl->parent_url[pomurl->ref_cnt+1] = NULL;
		pomurl->ref_cnt ++;

		src->moved_to = pomurl;
		src->status |= URL_MOVED;

#ifdef WITH_TREE
#ifdef I_FACE
		if (cfg.xi_face)
		{
			pomurl->tree_nfo = (GUI_TREE_RTYPE*) _realloc(pomurl->tree_nfo , 
					(pomurl->ref_cnt) * sizeof(GUI_TREE_RTYPE));
			pomurl->tree_nfo[pomurl->ref_cnt-1] = iface_make_tree_entry(pomurl);
		}
#endif
#endif

		xprintf(1 , gettext("!!!! m o v e d   t o   a l r e a d y   p r o c e s s e d !!!!\n"));

		if ((pomurl->status & URL_MOVED) || (pomurl->status & URL_DOWNLOADED))
		{
			url *purl = pomurl;
			char *fn = NULL;

			if (pomurl->status & URL_MOVED)
			{
				while(purl->moved_to) purl = purl->moved_to;
				fn = url_to_filename(purl);
			}
			else
				fn = url_to_filename(pomurl);

			if (cfg.rewrite_links && (purl->status & URL_DOWNLOADED))
				rewrite_parents_links(src , fn);

			_free(fn);
		}
		
	}
	else
	{
		dst->parent_url = _malloc(2 * sizeof(url *));
		dst->parent_url[0] = src;
		dst->parent_url[1] = NULL;
		src->moved_to = dst;
		src->status |= URL_MOVED;

		url_add_to_hash_tab(dst);

#ifdef WITH_TREE
#ifdef I_FACE
		if (cfg.xi_face)
		{
			dst->tree_nfo = (GUI_TREE_RTYPE*) _malloc(sizeof(GUI_TREE_RTYPE));
			dst->tree_nfo[0] = iface_make_tree_entry(dst);
		}
#endif
#endif
	}

}

void url_add_to_hash_tab(urlp)
url *urlp;
{
	char *p,*us;
	unsigned int hset;

	if (!(p = url_to_filename(urlp))) return;  /*** unsupported protocol ***/

	us = get_abs_file_path(p);
	_free(p);

	hset = hash_func(us , HASH_TBL_SIZE);
	free(us);

	cfg.url_hash_tbl[hset].nr ++;
	cfg.url_hash_tbl[hset].urls = (url **)_realloc(
		(char *) cfg.url_hash_tbl[hset].urls ,
		cfg.url_hash_tbl[hset].nr * sizeof(url *));

	cfg.url_hash_tbl[hset].urls[cfg.url_hash_tbl[hset].nr - 1] = urlp;
}

void url_remove_from_hash_tab(urlp)
url *urlp;
{
	char *p,*us;
	int i;
	bool fnd = FALSE;
	unsigned int hset;

	if (!(p = url_to_filename(urlp))) return;  /*** unsupported protocol ***/
	us = get_abs_file_path(p);
	_free(p);

	hset = hash_func(us , HASH_TBL_SIZE);
	free(us);

	for(i = 0; i < cfg.url_hash_tbl[hset].nr ; i++)
	{
		if (cfg.url_hash_tbl[hset].urls[i] == urlp)
			fnd = TRUE;

		if (fnd && (i + 1) < cfg.url_hash_tbl[hset].nr)
		{
			cfg.url_hash_tbl[hset].urls[i] =
				cfg.url_hash_tbl[hset].urls[i+1];
		}
	}

	if (fnd) cfg.url_hash_tbl[hset].nr--;
}

/**********************************************/
/* kopirovanie obsahu na nove miesto v pamati */
/**********************************************/
url *new_url(urlo)
url *urlo;
{
	url *res = (url *)_malloc(sizeof(url));

	memcpy(res , urlo , sizeof(url));

	return res;
}

/******************************************************/
/* k danemu URL vytvori meno suboru v lokalnom strome */
/******************************************************/
char *url_to_filename(urlp)
url *urlp;
{
	char pom[PATH_MAX];
	char pom2[PATH_MAX];
	char *p1,*p2;
	char *p;

	if (cfg.store_name && cfg.mode == MODE_SINGLE && 
	     !(urlp->status & URL_INLINE_OBJ) &&
	       urlp == cfg.urlstack[0])
		return new_string(cfg.store_name);

	switch(urlp->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			sprintf(pom2 ,"%s/%s_%d/%s",
				prottable[urlp->type].dirname ,
				urlp->p.http.host , urlp->p.http.port ,
				urlp->p.http.document);
			if (urlp->p.http.searchstr)
			{
				strcat(pom2 , "?");
				strcat(pom2 , urlp->p.http.searchstr);
			}
			if (*(pom2 + strlen(pom2)-1) == '/')
				strcat(pom2, cfg.index_name);
			break;

		case FILET:
			strcpy(pom , urlp->p.file.filename);
			pom2[0] = '\0';
			break;
		case FTP:
			if (urlp->p.ftp.dir)
				sprintf(pom2 , "%s/%s_%d/%s/%s", 
					prottable[FTP].dirname ,
					urlp->p.ftp.host , urlp->p.ftp.port ,
					urlp->p.ftp.path , cfg.index_name);
			else
				sprintf(pom2 , "%s/%s_%d/%s",
					prottable[FTP].dirname ,
					urlp->p.ftp.host , urlp->p.ftp.port , 
					urlp->p.ftp.path);
			break;
		case GOPHER:
			sprintf(pom2 , "%s/%s_%d/%c%s" ,
				prottable[GOPHER].dirname ,
				urlp->p.gopher.host , urlp->p.gopher.port ,
				urlp->p.gopher.type , urlp->p.gopher.selector);

			if (urlp->p.gopher.type == '1')
				strcat(pom2 , cfg.index_name);
			break;
		default:
			return NULL;
	}

	if (pom2[0])
	{
		char *trs;

		trs = tr(pom2);
		sprintf(pom , "%s/%s" , cfg.cache_dir , trs);
		_free(trs);
	}

	if ((urlp->type != FILET) && cfg.base_level)
	{
		p = get_abs_file_path(pom);
		strcpy(pom , p);
		_free(p);
		p1 = pom + strlen(cfg.cache_dir) + 
			(*(cfg.cache_dir + strlen(cfg.cache_dir) - 1) != '/');

		if (!(p2 = strfindnchr(p1 , '/' , cfg.base_level)))
		{
			if ((p2 = strrchr(pom , '/'))) p2 ++;
		}

		if (p2)
			memmove(p1 , p2 , strlen(p2) + 1);
	}

	return new_string(pom);
}

/****************************************************************/
/* k danemu URL vytvori meno docasneho suboru v lokalnom strome */
/****************************************************************/
char *url_to_in_filename(urlp)
url *urlp;
{
	char pom[PATH_MAX];
	char pom2[PATH_MAX];
	char *p;
	char *p1,*p2;


	if (cfg.mode == MODE_NOSTORE)
	{
		sprintf(pom , "%s/.in_pavuk_nostore_%d" , cfg.cache_dir ,
			getpid());
		return new_string(pom);
	}

	if (cfg.store_name && cfg.mode == MODE_SINGLE && 
	     !(urlp->status & URL_INLINE_OBJ) &&
	      cfg.actual_url == 0 && urlp == cfg.urlstack[0])
	{
		strcpy(pom , cfg.store_name);
		p = strrchr(pom , '/');
		if (!p) p = pom;
		else p++;
		memmove(p + 4 , p , strlen(p) + 1);
		strncpy(p, ".in_" , 4);
		
		return new_string(pom);
	}

	switch(urlp->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			sprintf(pom2 ,"%s/%s_%d/%s",
				prottable[urlp->type].dirname ,
				urlp->p.http.host , urlp->p.http.port ,
				urlp->p.http.document + 1);
			if (urlp->p.http.searchstr)
			{
				strcat(pom2 , "?");
				strcat(pom2 , urlp->p.http.searchstr);
			}
			if (*(pom2 + strlen(pom2)-1) == '/')
				strcat(pom2, cfg.index_name);
			break;
		case FTP:
			if (urlp->p.ftp.dir)
				sprintf(pom2 , "%s/%s_%d/%s/%s", 
					prottable[FTP].dirname ,
					urlp->p.ftp.host , urlp->p.ftp.port ,
					urlp->p.ftp.path , cfg.index_name);

			else
				sprintf(pom2 , "%s/%s_%d/%s",
					prottable[FTP].dirname ,
					urlp->p.ftp.host , urlp->p.ftp.port , 
					urlp->p.ftp.path);
			break;
		case FILET:
			strcpy(pom , urlp->p.file.filename);
			pom2[0]='\0';
			break;
		case GOPHER:
			sprintf(pom2 , "%s/%s_%d/%c%s" ,
				prottable[GOPHER].dirname ,
				urlp->p.gopher.host , urlp->p.gopher.port ,
				urlp->p.gopher.type , urlp->p.gopher.selector);

			if (urlp->p.gopher.type == '1')
				strcat(pom2 , cfg.index_name);
			break;
		default:
			return NULL;
	}

	if (pom2[0])
	{
		char *trs;

		trs = tr(pom2);
		sprintf(pom , "%s/%s" , cfg.cache_dir , trs);
		_free(trs);
	}

	p = strrchr(pom , '/');
	if (!p) p = pom;
	else p++;
	memmove(p + 4 , p , strlen(p) + 1);
	strncpy(p, ".in_" , 4);

	if ((urlp->type != FILET) && cfg.base_level)
	{
		p = get_abs_file_path(pom);
		strcpy(pom , p);
		_free(p);
		p1 = pom + strlen(cfg.cache_dir) +
			(*(cfg.cache_dir + strlen(cfg.cache_dir) - 1) != '/');

		if (!(p2 = strfindnchr(p1 , '/' , cfg.base_level)))
		{
			if ((p2 = strrchr(pom , '/'))) p2 ++;
		}

		if (p2)
			memmove(p1 , p2 , strlen(p2) + 1);
	}

 	return new_string(pom);
}

/************************************************/
/* zo struktury URL vytvori retazec URL + hesla */
/************************************************/
char *url_to_urlstr(urlp)
url *urlp;
{
	char pom[PATH_MAX];
	char *p;

	switch(urlp->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			sprintf(pom ,"%s%s%s%s%s%s:%d%s%s%s%s%s", prottable[urlp->type].typestr ,
				urlp->p.http.user ? urlp->p.http.user : "" ,
				urlp->p.http.password ? ":" : "" ,
				urlp->p.http.password ? urlp->p.http.password : "" ,					
				(urlp->p.http.password || urlp->p.http.user) ? "@" : "" ,
				urlp->p.http.host , urlp->p.http.port ,
				urlp->p.http.document ,
				urlp->p.http.searchstr ? "?" : "" ,
				urlp->p.http.searchstr ? urlp->p.http.searchstr : "" ,
				urlp->p.http.anchor_name ? "#" : "" ,
				urlp->p.http.anchor_name ? urlp->p.http.anchor_name : "");

			return new_string(pom);
		case FILET:
			p = get_abs_file_path(urlp->p.file.filename);

			if (urlp->p.file.anchor_name)
				sprintf(pom ,"%s%s#%s", prottable[FILET].typestr , p ,
					urlp->p.file.anchor_name);
			else
				sprintf(pom ,"%s%s", prottable[FILET].typestr , p);
			free(p);

			return new_string(pom);
		case FTP:
			sprintf(pom , "%s%s%s%s%s%s:%d%s%s%s", prottable[FTP].typestr ,
				urlp->p.ftp.user ? urlp->p.ftp.user : "" ,
				urlp->p.ftp.password ? ":" : "" ,
				urlp->p.ftp.password ? urlp->p.ftp.password : "" ,					
				(urlp->p.ftp.password || urlp->p.ftp.user) ? "@" : "" ,
				urlp->p.ftp.host , urlp->p.ftp.port , 
				urlp->p.ftp.path ,
				urlp->p.ftp.anchor_name ? "#" : "" ,
				urlp->p.ftp.anchor_name ? urlp->p.ftp.anchor_name : "");

			return new_string(pom);
		case GOPHER:
			sprintf(pom , "%s%s:%d/%c%s" , prottable[GOPHER].typestr ,
				urlp->p.gopher.host , urlp->p.gopher.port ,
				urlp->p.gopher.type , urlp->p.gopher.selector);
	
			return new_string(pom);
		default:
			return NULL;
	}
}


/****************************************************/
/* zo struktury URL vytvori retazec URL - bez hesla */
/****************************************************/
char *url_to_urlstr_woauth(urlp)
url *urlp;
{
	char pom[PATH_MAX];
	char *p,*p2;

	switch(urlp->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			p = encode_url(urlp->p.http.document);
			p2 = urlp->p.http.searchstr ? encode_searchstr(urlp->p.http.searchstr) : NULL;
			sprintf(pom ,"%s%s:%d%s%s%s%s%s", prottable[urlp->type].typestr ,
				urlp->p.http.host , urlp->p.http.port ,
				p ,
				p2 ? "?" : "" ,
				p2 ? p2 : "" ,
				urlp->p.http.anchor_name ? "#" : "" ,
				urlp->p.http.anchor_name ? urlp->p.http.anchor_name : "");
			_free(p);
			return new_string(pom);
		case FILET:
			p = get_abs_file_path(urlp->p.file.filename);

			if (urlp->p.file.anchor_name)
				sprintf(pom ,"%s%s#%s", prottable[FILET].typestr , p ,
					urlp->p.file.anchor_name);
			else
				sprintf(pom ,"%s%s", prottable[FILET].typestr , p);
			free(p);

			return new_string(pom);
		case FTP:
			p = encode_url(urlp->p.ftp.path);
			sprintf(pom , "%s%s:%d%s%s%s", prottable[FTP].typestr ,
				urlp->p.ftp.host , urlp->p.ftp.port , 
				p ,
				urlp->p.ftp.anchor_name ? "#" : "" ,
				urlp->p.ftp.anchor_name ? urlp->p.ftp.anchor_name : "");

			_free(p);

			return new_string(pom);
		case GOPHER:
			sprintf(pom , "%s%s:%d/%c%s" , prottable[GOPHER].typestr ,
				urlp->p.gopher.host , urlp->p.gopher.port ,
				urlp->p.gopher.type , urlp->p.gopher.selector);
	
			return new_string(pom);
		default:
			return NULL;
	}
}


/********************************************************/
/* z URL vrati adresu servera pre dokument		*/
/********************************************************/
char *url_get_site(urlr)
url *urlr;
{
	switch(urlr->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			return urlr->p.http.host;
		case FTP:
			return urlr->p.ftp.host;
		case GOPHER:
			return urlr->p.gopher.host;
		default: return NULL;
	}
}

int url_get_port(urlr)
url *urlr;
{
	switch(urlr->type)
	{
                case HTTP:
#ifdef USE_SSL
                case HTTPS:
#endif
                        return urlr->p.http.port;
                case FTP:
                        return urlr->p.ftp.port;
                case GOPHER:
                        return urlr->p.gopher.port;
                default: return 0;
	}
}

char *url_get_path(urlr)
url *urlr;
{
	switch(urlr->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			return urlr->p.http.document;
		case FTP:
			return urlr->p.ftp.path;
		case GOPHER:
			return urlr->p.gopher.selector;
		case FILET:
			return urlr->p.file.filename;
		default: return NULL;
	}
}

void url_set_path(urlr, path)
url *urlr;
char *path;
{
	switch(urlr->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			urlr->p.http.document = new_string(path);
			break;
		case FTP:
			urlr->p.ftp.path = new_string(path);
			break;
		case GOPHER:
			urlr->p.gopher.selector = new_string(path);
			break;
		case FILET:
			urlr->p.file.filename = new_string(path);
			break;
		default: return;
	}
}

char * url_get_pass(urlr , realm)
url *urlr;
char *realm;
{
	char *pass = NULL;
	authinfo *ai;

	switch(urlr->type)
	{
                case HTTP:
#ifdef USE_SSL
                case HTTPS:
#endif
                        pass = urlr->p.http.password;
			break;
                case FTP:
                        pass = urlr->p.ftp.password;
			break;
                default: return NULL;
	}

	if (!pass)
	{
		ai = authinfo_match_entry(urlr->type , url_get_site(urlr) , 
					  url_get_port(urlr) , url_get_path(urlr) , realm);
		if (ai) pass = ai->pass;
	}

	if (!pass)
	{
		pass = cfg.passwd_auth;
	}

	return pass;
}

char * url_get_user(urlr , realm)
url *urlr;
char *realm;
{
	char *user = NULL;
	authinfo *ai;

	switch(urlr->type)
	{
                case HTTP:
#ifdef USE_SSL
                case HTTPS:
#endif
                        user = urlr->p.http.user;
			break;
                case FTP:
                        user = urlr->p.ftp.user;
                       	break;
                default: return NULL;
	}

	if (!user)
	{
		ai = authinfo_match_entry(urlr->type , url_get_site(urlr) , 
					  url_get_port(urlr) , url_get_path(urlr) , realm);
		if (ai) user = ai->user;
	}

	if (!user)
	{
		user = cfg.name_auth;
	}
	
	return user;
}

int url_get_auth_scheme(urlr , realm)
url *urlr;
char *realm;
{
	authinfo *ai;
	int scheme = cfg.auth_scheme;

	ai = authinfo_match_entry(urlr->type , url_get_site(urlr) , 
				  url_get_port(urlr) , url_get_path(urlr) , realm);
	if (ai) scheme = ai->type;

	return scheme;
}

char * url_get_anchor_name(urlp)
url *urlp;
{
	char *anchor = NULL;

	switch(urlp->type)
	{
		case HTTP:
		case HTTPS:
			anchor = urlp->p.http.anchor_name;
			break;
		case FTP:
			anchor = urlp->p.ftp.anchor_name;
			break;
		case FILET:
			anchor = urlp->p.file.anchor_name;
			break;
		default:
			anchor = NULL;
			break;
	}

	return anchor;
}

int url_is_dir_index(urlp)
url *urlp;
{
	return ((urlp->type == HTTP || urlp->type == HTTPS) &&
		*(urlp->p.http.document + 
			strlen(urlp->p.http.document) - 1) == '/') ||
		(urlp->type == FTP && urlp->p.ftp.dir);

}

/**************************************************/
/* absolutna cesta k dokumentu z lokalneho stromu */
/* ktory je referencovany relativne		  */
/**************************************************/
char *get_redirect_abs_path(rurl , fstr)
url *rurl;
char *fstr;
{
	char *pom , *p , *p1;
	
	pom = url_to_filename(rurl);
	p = strrchr(pom , '/');

	p1 = realloc(pom , strlen(fstr) + (p - pom) + 2);
	strcpy(p1 + (p - pom) + 1 , fstr);

	p = get_abs_file_path(p1);
	free(p1);
	
	return p;
}

void url_path_abs(urlp)
url *urlp;
{
	char *p;

	switch (urlp->type)
	{
		case HTTP:
#ifdef USE_SSL
		case HTTPS:
#endif
			p = get_abs_file_path(urlp->p.http.document);
			free(urlp->p.http.document);
			urlp->p.http.document = p;
			break;
		case FTP:
			p = get_abs_file_path(urlp->p.ftp.path);
			free(urlp->p.ftp.path);
			urlp->p.ftp.path = p;
			break;
		case FILET:
			p = get_abs_file_path(urlp->p.file.filename);
			free(urlp->p.file.filename);
			urlp->p.file.filename = p;
			break;
		default:
			break;
	}
}

url *filename_to_url(ifn)
char *ifn;
{
	int cdln = strlen(cfg.cache_dir);
	if (*ifn != '/') return NULL;

	if (!strncmp(ifn , cfg.cache_dir , cdln))
	{
		char *p;
		int i;
		url *nurl = _malloc(sizeof(url));
		char *fn = new_string(ifn);

		p = fn + cdln;
		p += (*p == '/');
		
		nurl->status &= ~URL_REDIRECT;
		nurl->level = 0;
		nurl->parent_url = NULL;
		nurl->moved_to = NULL;
		nurl->extension = NULL;

#ifdef WITH_TREE
#ifdef I_FACE
		nurl->prop = NULL;
		nurl->tree_nfo = NULL;
#endif
#endif

		for (i = 0 ; i < NUM_ELEM(prottable) ; i++)
		{
			if (prottable[i].dirname && 
			    !strncmp(p , prottable[i].dirname , 
			    	     strlen(prottable[i].dirname)))
			{
				char *p2,*p3;

				nurl->type = prottable[i].id;
				nurl->parent_url = NULL;
				p += strlen(prottable[i].dirname) + 1;

				if (!p)
				{
					free(nurl);
					free(fn);
					return NULL;
				}

				switch (nurl->type)
				{
				    case HTTP:
#ifdef USE_SSL
				    case HTTPS:
#endif
					nurl->p.http.password = NULL;
					nurl->p.http.user = NULL;
					nurl->p.http.anchor_name = NULL;
					nurl->p.http.searchstr = NULL;
					nurl->p.http.port = prottable[i].default_port;
					if ((p2 = strchr(p , '/')))
					{
						int p2_len = strlen(p2);
						int idx_len = strlen(cfg.index_name);

						if (idx_len <= p2_len &&
						    !strcmp((p2 + p2_len - idx_len) , cfg.index_name) && 
						    ((p2_len > idx_len && *(p2 + p2_len - idx_len - 1) == '/') || idx_len == p2_len))
						{
							*(p2 + p2_len - idx_len) = '\0';
						}

						p3 = strchr(p2 , '?');
						if (p3)
						{
							*p3 = '\0';
							nurl->p.http.searchstr = new_string(p3+1);
						}
						nurl->p.http.document = new_string(p2);
						*p2 = '\0';
						p2 = strrchr(p , '_');
						if (p2)
						{
							p2++;
							nurl->p.http.port = _atoi(p2);
							if (errno == ERANGE)
							{
								nurl->p.http.host = 
									new_string(p);
								nurl->p.http.port = prottable[i].default_port;
							}
							else
							{
								nurl->p.http.host = 
									new_n_string(p , p2 - p - 1);
							}
						}
						else nurl->p.http.host = new_string(p);
					}
					else
					{
						free(nurl);
						free(fn);
						return NULL;
					}
					break;
				    case GOPHER:
					nurl->p.gopher.port = prottable[i].default_port;
					if ((p2 = strchr(p , '/')))
					{
						int p2_len = strlen(p2);
						int idx_len = strlen(cfg.index_name);

						p2++;
						nurl->p.gopher.type = *p2;

						if (idx_len <= p2_len &&
						    !strcmp((p2 + p2_len - idx_len) , cfg.index_name) && 
						    ((p2_len > idx_len && *(p2 + p2_len - idx_len - 1) == '1') || idx_len == p2_len))
						{
							*(p2 + p2_len - idx_len) = '\0';
						}
						nurl->p.gopher.selector = 
							new_string(p2 + 1);
						*p2 = '\0';
						p2 = strrchr(p , '_');
						if (p2)
						{
							p2++;
							nurl->p.gopher.port = _atoi(p2);
							if (errno == ERANGE)
							{
								nurl->p.gopher.host = 
									new_string(p);
								nurl->p.gopher.port = prottable[i].default_port;
							}
							else
							{
								nurl->p.gopher.host = 
									new_n_string(p , p2 - p - 1);
							}
						}
						else nurl->p.gopher.host = new_string(p);
					}
					else
					{
						free(nurl);
						free(fn);
						return NULL;
					}
					break;
				    case FTP:
					nurl->p.ftp.port = prottable[i].default_port;
					nurl->p.ftp.password = NULL;
					nurl->p.ftp.user = NULL;
					nurl->p.ftp.dir = FALSE;
					nurl->p.ftp.anchor_name = NULL;
					if ((p2 = strchr(p , '/')))
					{
						int p2_len = strlen(p2);
						int idx_len = strlen(cfg.index_name);

						if (idx_len <= p2_len &&
						    !strcmp((p2 + p2_len - idx_len) , cfg.index_name) && 
						    ((p2_len > idx_len && *(p2 + p2_len - idx_len - 1) == '/') || idx_len == p2_len))
						{
							*(p2 + p2_len - idx_len) = '\0';
							nurl->p.ftp.dir = TRUE;
						}
						nurl->p.ftp.path = new_string(p2);
						*p2 = '\0';
						p2 = strrchr(p , '_');
						if (p2)
						{
							p2++;
							nurl->p.ftp.port = _atoi(p2);
							if (errno == ERANGE)
							{
								nurl->p.ftp.host = 
									new_string(p);
								nurl->p.ftp.port = prottable[i].default_port;
							}
							else
							{
								nurl->p.ftp.host = 
									new_n_string(p , p2 - p - 1);
							}
						}
						else nurl->p.ftp.host = new_string(p);
					}
					else
					{
						free(nurl);
						free(fn);
						return NULL;
					}
					break;
#ifndef USE_SSL
				    case HTTPS:
#endif
				    case BROKEN:
				    case UNKNOWN:
				    case FILET:
				    case SHTTP:
				    case NEWS:
				    case NNTP:
				    case WAIS:
				    case WHOIS:
				    case MAILTO:
				    case PROSPERO:
				    case TELNET:
				    case TN3270:
				    case X500:
				    case RLOGIN:
				    case JAVASCRIPT:
				    case URN:
				    case LDAP:
				    case Z39_50R:
				    case Z39_50S:
				    default:
				        free(nurl);
				        nurl = NULL;
				    	break;
				}
				free(fn);
				return nurl;
			}
		}
		free(nurl);
	}
	return NULL;
}

/****************************************/
/* zisti ci bol dokument referencovany  */ 
/* v predchadzajucich cykloch		*/
/****************************************/
url *url_was_befor(urlp)
url *urlp;
{
	char *pn,*mn,*p;
	unsigned int hset;
	int i;

	p = url_to_filename(urlp);
	mn = get_abs_file_path(p);
	free(p);

	hset = hash_func(mn , HASH_TBL_SIZE);

	i = 0;

	while(i < cfg.url_hash_tbl[hset].nr)
	{
		if (urlp->type != cfg.url_hash_tbl[hset].urls[i]->type &&
		     (urlp->type != FILET || 
		      cfg.url_hash_tbl[hset].urls[i]->type != FILET))
		{
			i++;
			continue;
		}

		if (urlp == cfg.url_hash_tbl[hset].urls[i])
		{
			break;
		}
		
		p = url_to_filename(cfg.url_hash_tbl[hset].urls[i]);
		pn = get_abs_file_path(p);
		free(p);

		if (!strcmp(pn,mn))
		{
			free(pn);
			free(mn);
			return cfg.url_hash_tbl[hset].urls[i];
		}
#ifdef I_FACE
		if (!(i % 1000))
		{
			_Xt_Serve
		}
#endif
		free(pn);
		i++;
	}
	free(mn);
	return NULL;
}

