/***************************************************************************/
/* 		This code is part of WWW graber called pavuk		   */
/*		Copyright (c) 1997,1998,1999 Ondrejicka Stefan		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#ifndef _url_h_
#define _url_h_

#ifdef X_FACE
#include <X11/Intrinsic.h>
#include <Nws/TreeList.h>
#endif

#ifdef GTK_FACE
#include "ginterface.h"
#endif

#include "tools.h"

typedef struct {
	char	*host;		/*** adresa HTTP servera ***/
	unsigned short	port;		/*** port HTTP servera ***/
	char	*document;	/*** referencovany dokument ***/
	char	*searchstr;	/*** request for cgi's ***/
	char	*anchor_name;	/*** identifikator fragmentu dokumentu ***/
	char	*user;
	char	*password;
} httpurl;

typedef struct {
	char	*host;		/*** adresa FTP servera ***/
	unsigned short	port;		/*** cislo portu FTP servera ***/
	char	*user;		/*** meno pouzivatela pre pristup na dany server ***/
	char	*password;	/*** heslo pre pristup na dany server ***/
	char	*path;		/*** cesta k FTP dokumentu ***/
	char	*anchor_name;	/*** fragment HTML dokumentu ***/
	bool	dir;		/*** je dokument FTP adresar ? ***/
} ftpurl;

typedef struct {
	char *filename;		/*** cesta k dokumentu ***/
	char *anchor_name;	/*** fragment HTML dokumentu ***/
} fileurl;

typedef struct {
	char	*host;		/*** adresa GOPHER servera ***/
	unsigned short	port;		/*** cislo portu GOPHER servera ***/
	char	type;		/*** typ dokumentu ***/
	char	*selector;	/*** selektor dokumentu ***/
} gopherurl;

typedef struct {
	char	*host;		/*** adresa NNTP servera ***/ 
	unsigned short	port ;		/*** cislo portu NNTP servera ***/
	char	*news_group ;	/*** diskusna skupina ***/
	char	*article;	/*** identifikator prispevku ***/
} newsurl;

typedef struct {
	char	*mailadr;	/*** e-mailova adresa ***/
} mailurl;

typedef enum { 		/*** id for URL types ***/
	UNKNOWN ,
	BROKEN ,
	HTTP , 
	HTTPS , 
	SHTTP ,
	FTP ,
	FILET , 
	GOPHER , 
	NEWS , 
	NNTP ,
	WAIS , 
	MAILTO , 
	TELNET ,
	RLOGIN ,
	TN3270 ,
	X500 ,
	WHOIS ,
	PROSPERO ,
	JAVASCRIPT ,
	URN ,
	LDAP ,
	Z39_50S ,
	Z39_50R
} protocol;

#define URL_REDIRECT		(unsigned short) (1 << 0)
#define URL_INLINE_OBJ		(unsigned short) (1 << 1)
#define URL_PROCESSED		(unsigned short) (1 << 2)
#define URL_DOWNLOADED		(unsigned short) (1 << 3)
#define URL_ERR_UNREC		(unsigned short) (1 << 4)
#define URL_MOVED		(unsigned short) (1 << 5)
#define URL_REJECTED		(unsigned short) (1 << 6)
#define URL_USER_DISABLED	(unsigned short) (1 << 7)
#define URL_NOT_FOUND		(unsigned short) (1 << 8)
#define URL_TRUNCATED		(unsigned short) (1 << 9)
#define URL_ERR_REC		(unsigned short) (1 << 10)
#define URL_STYLE		(unsigned short) (1 << 11)

typedef struct _protinfo{
	protocol	id;
	char		*dirname;
	char		*urlid;
	char		*typestr;
	int		default_port;
	bool		supported;
}protinfo;

typedef struct {			/*** properties of document ***/
	char		*type;		/*** MIME type of document ***/
	int		size;		/*** size of document ***/
} url_prop;

typedef	union {			/***  unifikacia reprezentacie jednotlivych URL ***/ 
	    httpurl	http;
	    fileurl	file;
	    ftpurl	ftp;
	    gopherurl	gopher;
	    newsurl	news;
	    mailurl	mail;
} abs_url;

typedef struct _url {
	protocol	type;		/*** typ daneho URL ***/
	struct _url	**parent_url;	/*** pointer to array of parent URLs ***/
	struct _url	*moved_to;	/*** pointer to new URL if document was moved ***/
	unsigned short	level;		/*** tree level of document ***/
	unsigned short	ref_cnt;	/*** number of references to this URL structure ***/
	unsigned short	parent_cnt;	/*** number of references to this URL as parent URL ***/
	unsigned short	status;		/*** status flags of URL ***/
#ifdef WITH_TREE
#ifdef I_FACE
	url_prop	*prop;		/*** document properties***/
	GUI_TREE_RTYPE	*tree_nfo;	/*** graphical representation of tree nodes ***/
#endif
#endif /* WITH_TREE */
	abs_url		p;		/*** parsed URL infos ***/
	void		*extension;	/*** url extensions ***/
} url;

#if NeedFunctionPrototypes
extern url *parse_url(char *);
extern char *url_to_absolute_url(char * , char * , url * , char *);
extern char * encode_url(char *);
extern char * encode_searchstr(char *);
extern char * decode_url(char *);
extern url * new_url(url *);
extern char * url_to_filename(url *);
extern char * url_to_in_filename(url *);
extern void free_deep_url(url *);
extern char *get_redirect_abs_path(url * , char *);
extern char *url_to_urlstr(url*);
extern char *url_to_urlstr_woauth(url*);
extern void url_path_abs(url *);
extern url *filename_to_url(char *);
extern void append_url_to_list(url *);
extern char * url_get_site(url *);
extern int url_get_port(url *);
extern char * url_get_pass(url *, char *);
extern char * url_get_user(url *, char *);
extern int url_get_auth_scheme(url *, char *);
extern char * url_get_path(url *);
extern char * url_get_anchor_name(url *);
extern void url_set_path(url *, char *);
extern int url_is_dir_index(url *);
extern void url_add_to_hash_tab(url *);
extern void url_remove_from_hash_tab(url *);
extern url * url_was_befor(url *);
extern void replace_url_in_list(url *, int);
extern void link_url_in_list(url *,url *);
extern void url_redirect_to(url *, url*);
#else
extern url *parse_url();
extern char *url_to_absolute_url();
extern char * encode_url();
extern char * encode_searchstr();
extern char * decode_url();
extern url * new_url();
extern char * url_to_filename();
extern char * url_to_in_filename();
extern void free_deep_url();
extern char *get_redirect_abs_path();
extern char *url_to_urlstr();
extern char *url_to_urlstr_woauth();
extern void url_path_abs();
extern url *filename_to_url();
extern void append_url_to_list();
extern char *url_get_site();
extern int url_get_port();
extern char * url_get_pass();
extern char * url_get_user();
extern int url_get_auth_scheme();
extern char * url_get_path();
extern void url_set_path();
extern void url_add_to_hash_tab();
extern char * url_get_anchor_name();
extern int url_is_dir_index();
extern void url_remove_from_hash_tab();
extern url * url_was_befor();
extern void replace_url_in_list();
extern void link_url_in_list();
extern void url_redirect_to();
#endif

extern protinfo prottable[23];

#endif
